<?php
/**
 * Tabular (tabular).
 */

/**
 * Class Tabular.
 * Extend - and implement TabularInterface - and provide tabular entity
 * type specs.
 * @see TabularInterface
 */
abstract class Tabular {

  /**
   * Our error codes must be within this range (both included).
   *
   * Do not override.
   *
   * @var array
   */
  public static $errorRange = array(1002000, 1002999);

  /**
   * Error codes.
   *
   * Do not override. Or do but include at least the codes listed here.
   *
   * @var array
   */
  public static $errorCodes = array(
    // The usual suspects.
    'unknown' => 1002001,
    'unknown_propagated' => 1002002,
    //  Programmatic errors and wrong use of program.
    'algo' => 1002006,
    'use' => 1002007,
    'not_implemented' => 1002008,
    // Configuration errors.
    'configuration' => 1002010,
    'multiple_identical_groups' => 1002015,
    'non_resolvable_js_function' => 1002018,
    'inapplicable_dropdown' => 1002019,
    // Various.
    'message_entitykind' => 1002021,
    'message_name' => 1002022,
    'primary_key_nonexist' => 1002031,
    'primary_key_empty' => 1002032,
    'data_column_nonexist' => 1002033,
    'widget_name_nonexist' => 1002035,
    'widget_index_nonexist' => 1002036,
    // Frontend initialisation errors.
    'frontend_init_group' => 1002040,
    'non_unique_widget_name' => 1002041,
    'frontend_init_widget' => 1002045,

    // Provider errors.
    // HTTP status 502.
    'provider_bad_gateway' => 1002052,
    // HTTP status 504.
    'provider_gateway_timeout' => 1002054,
    'provider_search_validation_failure' => 1002059,

    'ajax_response_failure' => 1002061,
    'ajax_response_malformatted' => 1002062,
    'frontend_convert_data' => 1002065,

    'widget_table_no_datatable' => 1002091,
  );

  /**
   * Get one or all translated messages for a tabular entity kind (or generic).
   *
   * @param string $entityKind
   *   Empty ~ generic.
   * @param string $name
   *
   * @return array|string
   *   Array on empty $name.
   */
  final public static function message($entityKind = '', $name = '') {
    static $generic;
    if (!$entityKind || $entityKind == 'generic') {
      $entityKind = 'generic';
      $translation = self::translation();
    }
    elseif ($entityKind == static::ENTITY_KIND) {
      $translation = static::translation();
    }
    else {
      return !$name ? array() : '_MESSAGE_NONEXISTENT_ENTITYKIND_';
    }
    if (!$name) {
      return $translation;
    }
    return isset($translation[$name]) ? $translation[$name] :
      '_MESSAGE_NONEXISTENT_NAME_';
  }

  /**
   * Multibyte-safe strpos/stripos checker, for searchable providers.
   *
   * @param string $haystack
   * @param string $needle
   * @param integer $position
   * @param boolean $caseSensitive
   *
   * @return boolean
   */
  final public static function unicodeMatchPosition($haystack, $needle, $position = 0, $caseSensitive = FALSE) {
    static $mb = -1;
    if ($mb < 0) {
      $mb = function_exists('mb_stripos') ? 1 : 0;
    }
    if ($mb) {
      return $caseSensitive ? mb_strpos($haystack, $needle) === $position : mb_stripos($haystack, $needle) === $position;
    }
    if (!$caseSensitive) {
      $haystack = drupal_strtolower($haystack);
      $needle = drupal_strtolower($needle);
    }
    return drupal_substr($haystack, $position, drupal_strlen($needle)) === $needle;
  }


  // Child classes are required to override these attributes and methods. ------

  /**
   * Name of this module.
   *
   * @type string
   */
  const MODULE = 'tabular';

  /**
   * This class declares no entity kind, but it has to set this constant
   * (for message()).
   * And 'generic' is an illegal entity kind name, so no one will use that.
   *
   * @type string
   */
  const ENTITY_KIND = 'generic';

  /**
   * Get all generic (not entity kind specific) translated messages.
   *
   * Child class (implementing TabularInterface) must implement,
   * translating messages of it's entity kind.
   *
   * @see TabularInterface::schema()
   */
  protected static function translation() {
    static $translation;
    if (!$translation) {
      // Prepare reusable.
      $report_error = t('<br>The incident has been logged, but please report this error (@errorcode) to Service Desk.',
        array(), array('context' => 'module:tabular:message:report_error'));

      $translation = array(
        'report_error' => $report_error,
        'fatal_frontend_error' => t('Fatal frontend error.',
            array(), array('context' => 'module:tabular:message:fatal_frontend_error'))
          . $report_error,
      );
    }

    return $translation;
  }


  // Optionally override. ------------------------------------------------------

  /**
   * Characters forbidden in a search string (searchable provider).
   *
   * See also Javacript property populateSearchIllegals, which must be the same.
   *
   * @type string
   */
  const POPULATE_SEARCH_ILLEGALS = '/[\/%\*\!\|\"\[\]\r\n`<>\\\\]/';

  /**
   * Validate a search string (searchable provider).
   *
   * See
   *
   * @param string $str
   * @return boolean
   *   False: on empty $str and on validation failure.
   */
  public static function validatePopulateSearch($str) {
    if (($str || $str !== '') && !preg_match(static::POPULATE_SEARCH_ILLEGALS, $str)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Number of seconds to cache data.
   *
   * An hour.
   *
   * @type integer
   */
  const CACHE_TIME_TO_LIVE = 3600;

  /**
   * Send (no)cache headers, to secure/prevent browser caching.
   *
   * AJAX callbacks better use this method.
   *
   * Browser caching (on/off) is controlled by frontend; configurable via
   * field settings UI.
   *
   * @see Tabular::CACHE_TIME_TO_LIVE
   *
   * @param integer $timeToLive
   *   Default: zero (~ static::CACHE_TIME_TO_LIVE).
   *   Minus one: ignore request specifics, do not cache.
   */
  public static function sendCacheHeaders($timeToLive = 0) {
    if (!$timeToLive) {
      $timeToLive = static::CACHE_TIME_TO_LIVE;
    }
    if ($timeToLive == -1) {
      $noCache = TRUE;
    }
    else {
      // Look for jQuery.ajax()' no-cache GET arg.
      $noCache = !empty($_GET['_']);
    }
    if ($noCache) {
      header('Cache-Control: private, no-store, no-cache, must-revalidate, post-check=0, pre-check=0');
      //header('Expires: Thu, 01 Jan 1970 00:00:01 GMT');
    }
    else {
      header('Cache-Control: public, max-age=' . $timeToLive);
      header('Expires: ' . gmdate('D, d M Y H:i:s', (REQUEST_TIME + $timeToLive)) . ' GMT');
      header('Last-modified: ' . gmdate('D, d M Y H:i:s', REQUEST_TIME) . ' GMT');
    }
  }

}

/**
 * Interface TabularInterface.
 * Implement - and extend Tabular - to provide tabular entity kind specs.
 * Interface design because PHP doesn't static abstract methods.
 * @see Tabular
 */
interface TabularInterface {

  /**
   * Get entity kind schema of this class (entity kind).
   *
   * A list of primary key, view_columns and optionally buttons.
   *
   * NB: the view_columns of schema are just the columns to be displayed in GUI.
   * The actual data rows may contain more columns; which the tabular ignores
   * but keep in values.
   *
   * @see TabularExample::schema()
   *
   * @return array
   */
  static function schema();

  /**
   * Get all translated messages for the tabular entity kind of this class.
   *
   * @see TabularExample::translation()
   *
   * @return array
   */
  static function translation();

}
