/**
 * @file
 * Tabular (tabular) base.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';
  var
    /**
     * Namespace.
     *
     * @type {object}
     */
    Tabular = {
      /**
       * Inherit prototypal properties without instantiating parent(s).
       *
       * Sequence of parent class args should start with most basic class
       * and end with most specialized class, to secure highest degree
       * of specialization.
       *
       * @param {class} child
       * @param {string} childName
       * @param {class} parent
       *   Any number of args.
       */
      extend: function(child, childName, parent) {
        var proto = child.prototype, protoOrig = {}, superProto, i, le = arguments.length, k;
        // Record names of own original prototypal properties to prevent super
        // from overriding this sub.
        for (k in proto) {
          if (proto.hasOwnProperty(k)) {
            protoOrig[k] = null;
          }
        }
        // For every parent.
        for (i = 2; i < le; i++) {
          // Clear ref.
          superProto = null;
          superProto = arguments[i].prototype;
          for (k in superProto) {
            if (superProto.hasOwnProperty(k) && !protoOrig.hasOwnProperty(k)) {
              proto[k] = superProto[k];
            }
          }
        }
        proto.constructor = child;
        // For .typeOf() method.
        proto.type_of = childName;
      }
    };

  // Make globally available.
  Drupal.Tabular = Tabular;

  // Work on native Error, TypeError and RangeError.
  if (!Error.prototype.hasOwnProperty('customCode')) {
    /**
     * Get/set custom error code.
     * @param {number} [code]
     *   Get code if no arg at all.
     * @returns {Error|number}
     *   Error self if arg code.
     *   Self.custom_code if no args.
     */
    Error.prototype.customCode = function(code) {
      if (arguments.length) {
        this.custom_code = code || 0;
        return this;
      }
      return this.custom_code || 0;
    };
    if (!TypeError.prototype.hasOwnProperty('customCode')) {
      TypeError.prototype.customCode = Error.prototype.customCode;
    }
    if (!RangeError.prototype.hasOwnProperty('customCode')) {
      RangeError.prototype.customCode = Error.prototype.customCode;
    }
  }

  /**
   * Abstract base class providing prototypal properties for other classes to copy.
   */
  Tabular.base = function() {
    throw new Error('Abstract class ' + this.type_of + ' is not to be instantiated.');
  };
  Tabular.base.prototype = {
    constructor: Tabular.base,
    type_of: 'Tabular.base',
    /**
     * For inspect + Judy.typeOf().
     * @returns {string}
     */
    typeOf: function() {
      return this.type_of;
    },
    /**
     * Deletes all instance properties after inviting them to detroy().
     */
    destroy: function() {
      var k;
      for (k in this) {
        if (this.hasOwnProperty(k)) {
          if (typeof this[k].destroy === 'function') {
            this[k].destroy();
          }
          delete this[k];
        }
      }
    }
  };

  /**
   * List of 'fields'.
   * See tabular.base.js.
   * @type {object}
   */
  Tabular.widgetGroups = {};

  /**
   * @param {string} groupId
   * @returns {Tabular.widgetGroup|undefined}
   */
  Tabular.getGroup = function(groupId) {
    if (Tabular.widgetGroups.hasOwnProperty(groupId)) {
      return Tabular.widgetGroups[groupId];
    }
  };

  /**
   * @param {string} groupId
   * @param {string} widgetName
   * @returns {Tabular.widget.base|undefined}
   */
  Tabular.getWidget = function(groupId, widgetName) {
    if (Tabular.widgetGroups.hasOwnProperty(groupId)) {
      return Tabular.widgetGroups[groupId].getWidget(widgetName);
    }
  };

  /**
   * For internal use only - don't mess with it.
   * @param {string} groupId
   * @param {string} widgetName
   * @returns {Tabular.widget.base|undefined}
   */
  Tabular.setLastUserActionWidget = function(groupId, widgetName) {
    if (Tabular.widgetGroups.hasOwnProperty(groupId)) {
      return Tabular.widgetGroups[groupId].setLastUserActionWidget(widgetName);
    }
  };

  /**
   * @param {string} name
   * @returns {number}
   *   Integer.
   */
  Tabular.errorCode = function(name) {
    var codes = Drupal.settings.tabular.errorCodes;
    if (codes.hasOwnProperty(name)) {
      return codes[name];
    }
    return codes.unknown || 0;
  };
  /**
   * @param {number} code
   * @returns {string}
   *   Empty if no such error code.
   */
  Tabular.errorName = function(code) {
    var codes = Drupal.settings.tabular.errorCodes, nm;
    for (nm in codes) {
      if (codes.hasOwnProperty(nm) && codes[nm] === code) {
        return nm;
      }
    }
    return '';
  };
  Tabular.severityToString = [
    'emergency',
    'alert',
    'critical',
    'error',
    'warning',
    'notice',
    'info',
    'debug'
  ];
  /**
   * Convert severity as number|string to number or string, securing validity.
   *
   * Defaults to critical(2).
   *
   * @param {number|string} severity
   * @param {boolean} toString
   */
  Tabular.severity = function(severity, toString) {
    var svrt = severity;
    // Convert to integer, or confirm existence of such severity.
    switch (typeof svrt) {
      case 'number':
        if (Tabular.severityToString[svrt] === undefined) {
          svrt = 2;
        }
        break;
      case 'string':
        if (!svrt || (svrt = $.inArray(svrt)) === -1) {
          svrt = 2;
        }
        break;
      default:
        svrt = 2;
    }
    return !toString ? svrt : Tabular.severityToString[svrt];
  };
  /**
   * Get one or all translated messages for a entity kind (or 'generic').
   *
   * @param {string} [entityKind]
   *   Empty ~ generic.
   * @param {string} [name]
   * @param {object} [replacers]
   *   List of replacers by (string) needles.
   * @returns {Array|string}
   *   Array on empty entityKind + name.
   */
  Tabular.message = function(entityKind, name, replacers) {
    var msgs = Drupal.settings.tabular.message, eKind, msg = '', k;
    if (!entityKind && !name) {
      return msgs;
    }
    eKind = entityKind || 'generic';
    try {
      if (!msgs.hasOwnProperty(eKind)) {
        msg = '_MESSAGE_NONEXISTENT_ENTITYKIND_';
        throw new RangeError('No messages for that entityKind[' + eKind + '], name[' + name + '].').
          customCode(Tabular.errorCode('message_entitykind'));
      }
      if (!name) {
        return msgs[eKind];
      }
      if (!msgs[eKind].hasOwnProperty(name)) {
        msg = '_MESSAGE_NONEXISTENT_NAME_';
        throw new RangeError('No message of that name[' + name + '], for entityKind[' + eKind + '].').
          customCode(Tabular.errorCode('message_name'));
      }
      msg = msgs[eKind][name];
      if (replacers) {
        for (k in replacers) {
          if (replacers.hasOwnProperty(k)) {
            msg = msg.replace(new RegExp(k.replace(/\!/, '\\!', 'g')), replacers[k]);
          }
        }
      }
      return msg;
    }
    catch (er) {
      inspect.traceLog(er, {
        type: 'tabular',
        message: er.message,
        code: (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown'),
        severity: 'error'
      });
    }
    return msg;
  };

  /**
   * Get setting(s) or set setting.
   * @param {string} [name]
   * @param {*|undefined} [set]
   * @returns {*}
   *   No args: returns all settings.
   *   One arg: returns setting by that name, or undefined.
   *   Two args: sets setting [name] with value [set], and returns undefined.
   */
  Tabular.settings = function(name, set) {
    var o = Drupal.settings.tabular, nArgs = arguments.length;
    if (!o || !nArgs) {
      return o;
    }
    if (nArgs === 1) {
      return o[name];
    }
    o[name] = set;
  };

  /**
   * Blocking alert()-like dialogue box.
   *
   * @param {string} msg
   * @param {string} [severity]
   */
  Tabular.alertDialog = function(msg, severity) {
    alert(msg.replace(/(<br\/?>)/g, "\n"));
  };

  /**
   * Blocking confirm()-like dialogue box.
   *
   * @param {string} msg
   * @param {object} obj
   *   If falsy, the function will be apply'ed to window.
   * @param {function} func
   * @param {Array} [args]
   *   Arguments for obj.func. Ignored if no arg obj.
   */
  Tabular.confirmDialog = function(msg, obj, func, args) {
    var errorCode;
    if (confirm(msg)) {
      try {
        func.apply(obj || window, args || []);
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
        inspect.traceLog(
          er,
          {
            type: self.module,
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        Tabular.alertDialog(
          Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode}),
          'error'
        );
      }
    }
  };

  /**
   * Get element without instantiating a jQuery object unless necessary.
   * @param {string|HTMLElement} selector
   * @returns {HTMLElement|undefined}
   */
  Tabular.element = function(selector) {
    if (selector) {
      if (typeof selector === 'string' && selector.charAt(0) === '#') {
        // Prevent getElementById() null for none.
        return document.getElementById(selector.substr(1)) || undefined;
      }
      return $(selector).get(0);
    }
  };

  if (typeof $.fn.showByHidden === 'undefined') {
    /**
     * jQuery method to show an element by removing HTML 'hidden' attribute.
     * @function
     * @name jquery#showByHidden
     * @returns {jquery}
     */
    $.fn.showByHidden = function() {
      var $jq = this, elm;
      if ($jq.length) {
        elm = $jq.get(0);
        elm.removeAttribute('aria-hidden');
        elm.removeAttribute('hidden');
      }
      return $jq;
    };
  }
  if (typeof $.fn.hideByHidden === 'undefined') {
    /**
     * jQuery method to hide an element by adding HTML 'hidden' attribute.
     * @function
     * @name jquery#hideByHidden
     * @returns {jquery}
     */
    $.fn.hideByHidden = function() {
      var $jq = this, elm;
      if ($jq.length) {
        elm = $jq.get(0);
        elm.setAttribute('aria-hidden', 'true');
        elm.setAttribute('hidden', '1');
      }
      return $jq;
    };
  }

  /**
   * Set 'visible-first' and 'visible-last' HTML element attribute
   * on a collection of elements.
   * @param {string} selector
   *
  Tabular.visibleFirstLast = function(selector) {
    var all = $(selector).get(), le = all.length, i, first = -1, last = -1, u, elm;
    if (le) {
      for (i = 0; i < le; i++) {
        if (!(u = all[i].getAttribute('hidden')) || u === '0' || u === 'false') {
          if (first === -1) {
            first = i;
          }
          last = i;
        }
      }
      for (i = 0; i < le; i++) {
        elm = all[i];
        if (first > -1 && i === first) {
          elm.setAttribute('visible-first', '1');
        }
        else {
          elm.removeAttribute('visible-first');
        }
        if (last > -1 && i === last) {
          elm.setAttribute('visible-last', '1');
        }
        else {
          elm.removeAttribute('visible-last');
        }
      }
    }
  };*/

  /**
   * Finds of matches of first sub pattern () of a regular expression.
   * Resets arg rgx' index so rgx is reusable.
   * @param {RegExp} rgx
   * @param {string} haystack
   * @returns {Array}
   */
  Tabular.matchSubPattern = function(rgx, haystack) {
    var mtchs = [], fnd, limit = 100, i = 0;
    while ((++i) <= limit && (fnd = rgx.exec(haystack)) !== null) {
      if (fnd.length === 2) {
        mtchs.push(fnd[1]);
      }
    }
    // Reset; otherwise the RegExp is useless hereafter.
    rgx.lastIndex = 0;
    return mtchs;
  };

  /**
   * @param {string} str
   * @param {object} dataRow
   * @returns {string}
   */
  Tabular.replaceColumnPatterns = function(str, dataRow) {
    var mtchs, le, i;
    if (str.indexOf('!') !== -1) {
      mtchs = Tabular.matchSubPattern(/\!([a-zA-Z\d_\-]+)/g, str);
      if ((le = mtchs.length)) {
        for (i = 0; i < le; i++) {
          if (dataRow.hasOwnProperty(mtchs[i])) {
            str = str.replace(
              '!' + mtchs[i],
              dataRow[mtchs[i]]
            );
          }
        }
      }
    }
    return str;
  };

  /**
   * Find parent field wrapper.
   *
   * @param {string|HTMLElement} selector
   * @returns {jquery}
   */
  Tabular.findFieldWrapper = function(selector) {
    return $(selector).closest('div.form-wrapper[id^="edit-field-"]');
  };

  /**
   * Find ancestor or parent field collection wrapper.
   *
   * @param {string|HTMLElement} selector
   * @param {string} [collectionClass]
   *   Require that the collection has that CSS class.
   * @returns {jquery}
   */
  Tabular.findCollectionWrapper = function(selector, collectionClass) {
    // That field collection CSS class may not be correct in all cases :-(
    var cls = 'field-type-field-collection-form', $jq = $(selector), me = $(selector).get(0);
    if (me) {
      // If arg selector is field collection in itself: go one element upwards.
      if (me.id && me.id.indexOf('edit-field-') === 0 && $jq.hasClass(cls)) {
        $jq = $(me.parentNode);
      }
      $jq = $jq.closest('div.form-wrapper[id^="edit-field-"].' + cls);
      if (!collectionClass || ($jq.length && $jq.hasClass(collectionClass))) {
        return $jq;
      }
    }
    return $();
  };

  /**
   * @param {string} str
   * @param {boolean} [first]
   *   Truthy: first char to uppercase.
   *   Default: false.
   * @returns {string}
   */
  Tabular.toCamelCase = function(str, first) {
    var a = str.split(/[_\- ]/), le = a.length, i;
    for (i = 0; i < le; i++) {
      if (i || first) {
        a[i] = a[i].charAt(0).toUpperCase() + a[i].substr(1);
      }
    }
    return a.join('');
  };

  /**
   * Converts string 'some.function' to reference to some.function if possible.
   * @param {string} evalName
   *   'some.function'.
   * @returns {Function|undefined}
   */
  Tabular.evalFunction = function(evalName) {
    var func = window, nm = '' + evalName, prts, le, i, u;
    if (nm) {
      if (nm.indexOf('window.') === 0) {
        nm = nm.substr(7);
      }
      if (nm) {
        prts = evalName.split('.');
        le = prts.length;
        if (le) {
          for (i = 0; i < le; i++) {
            // Clear ref.
            u = null;
            u = func[prts[i]];
            if (!u) {
              return;
            }
            switch (typeof u) {
              case 'object':
                if (i == le -1) {
                  // Last part must be function.
                  return;
                }
                func = u;
                break;
              case 'function':
                func = u;
                break;
              default:
                return;
            }
          }
          // IDE complains that func is window, but it can't be.
          return func;
        }
      }
    }
  };

  /**
   * Converts string 'someObject.someMethod'
   * to [(object)someObject, (string) 'someMethod']
   * or [(function)someObject, (string) 'someMethod']
   * if possible.
   * If there's only two parts and the first is window
   * [null, (string) 'someMethod'] because we don't want to refer window.
   * @param {string} evalName
   *   'some.function'.
   * @returns {Array|undefined}
   */
  Tabular.evalMethod = function(evalName) {
    var obj = window, func, nm = '' + evalName, prts, le, i, u;
    if (nm) {
      if (nm.indexOf('window.') === 0) {
        nm = nm.substr(7);
      }
      if (nm) {
        prts = evalName.split('.');
        le = prts.length;
        if (le) {
          // Iterate all but last part.
          for (i = 0; i < le - 1; i++) {
            // Clear ref.
            u = null;
            u = obj[prts[i]];
            if (!u) {
              return;
            }
            switch (typeof u) {
              case 'object':
              case 'function':
                obj = u;
                break;
              default:
                return;
            }
          }
          func = prts[prts.length - 1];
          if (typeof obj[func] === 'function') {
            return obj === window ? [null, func] : [obj, func];
          }
        }
      }
    }
  };

  /**
   * Convert raw data to common data format or from common data format to local.
   *
   * NB: first instantiated Tabular.widgetGroup will instantiate
   * to Drupal.Tabular.convert.
   * @constructor
   * @see Tabular.widgetGroup
   */
  Tabular.converter = function() {
  };
  Tabular.converter.prototype = {
    constructor: Tabular.converter,
    /**
     * @param {object} [format]
     *   Default: format of current language.
     * @returns {Tabular.converter}
     */
    init: function(format) {
      var self = this, frmt = format || Tabular.settings('converter_format');
      // Override (mask) prototypal format; unless missing format.
      if (frmt) {
        // 'money' has 'no_fraction'; 'number' hasn't.
        frmt.number.no_fraction = '';
        // 'money' must have same decimal mark as 'number'.
        frmt.money.decimal_mark = frmt.number.decimal_mark;
        self.format = frmt;
      }
      // Else, use prototypal (en-us) format.

      // List for derivative regexes that can only be created run-time.
      self.rgxRuntime = {};

      return self;
    },
    /**
     * Gets overridden (masked) by init() unless missing settings by language.
     */
    format: {
      number: {
        thousand_sep: ',',
        decimal_mark: '.',
        // Number of decimals; float only.
        n_decimals: 3,
        no_fraction: ''
      },
      money: {
        thousand_sep: ',',
        // 'money' must have same decimal mark as 'number'.
        decimal_mark: '.',
        n_decimals: 2,
        no_fraction: '/-'
      },
      time: {
        timestamp: 'C',
        // U.S. English; otherwise d/m/Y.
        datetime: 'm/d/Y H:i',
        date: 'm/d/Y',
        time: 'H:i'
      }
    },
    rgxScp: [
      /[.*+?^${}()|[\]\\]/g,
      '\\$&'
    ],
    rgx: {
      integer: /^\d+$/,
      // ISO-8601: At least year and month.
      iso8601Minimum: /^\d{4}\-\d{2}/,
      // ISO-8601: Match until no-zone(~local)|zone|Z.
      iso8601: /^(\d{4})\-?(\d{2})?\-?(\d{2})?[T ]?(\d{2})?:?(\d{2})?:?(\d{2})?\.?(\d{3})?/,
      // ISO-8601: Grab zone.
      iso8601Zone: /[\+\-]\d{2}:\d{2}$/,
      // Unix Epoch: integer, possibly negative.
      unixEpoch: /^\-?\d+$/
    },
    /**
     * Escape string for RegExp use.
     * @param {string|*} str
     * @returns {string}
     */
    regexEscape: function(str) {
      return ('' + str).replace(this.rgxScp[0], this.rgxScp[1], 'g');
    },
    emptySymbol: '&nbsp; &ndash; &nbsp;',
    /**
     * Symbol if null or empty string.
     * @param {string|*} any
     * @return {string}
     */
    emptyToSymbol: function(any) {
      return !any && (any === null || any === '') ? this.emptySymbol : ('' + any);
    },
    /**
     * @param {string|*} str
     * @return {string}
     */
    toLowerCase: function(str) {
      return ('' + str).toLowerCase();
    },
    /**
     * @param {string|*} str
     * @return {string}
     */
    toUpperCase: function(str) {
      return ('' + str).toUpperCase();
    },
    /**
     * @param {string|*} str
     * @return {string}
     */
    toUpperCaseFirst: function(str) {
      var s = '' + str, le = s.length;
      return !le ? '' : (s.charAt(0).toUpperCase() + (le === 1 ? '' : s.substr(1)));
    },
    /**
     * Convert number string in locale format (number or money) to number.
     * @param {number|string} num
     * @returns {number}
     */
    fromLocaleNumber: function(num) {
      var frmt, n = num, nf, nfx, sp, spx, dm;
      switch (typeof n) {
        case 'number':
          break;
        case 'string':
          n = $.trim(n);
          frmt = this.format;
          // Remove no fraction mark.
          nf = frmt.money.no_fraction;
          if (nf && n.indexOf(nf) > -1) {
            if (!(nfx = self.rgxRuntime.money_no_fraction)) {
              self.rgxRuntime.money_no_fraction = nfx = new RegExp(this.regexEscape(nf) + '$');
            }
            n = n.replace(nfx, '');
          }
          // Remove thousand separators.
          sp = frmt.money.thousand_sep;
          if (sp && n.indexOf(sp) > -1) {
            if (!(spx = self.rgxRuntime.money_thousand_sep)) {
              self.rgxRuntime.money_thousand_sep = spx = new RegExp(this.regexEscape(sp), 'g');
            }
            n = n.replace(spx, '');
          }
          sp = frmt.number.thousand_sep;
          if (sp && n.indexOf(sp) > -1) {
            if (!(spx = self.rgxRuntime.number_thousand_sep)) {
              self.rgxRuntime.number_thousand_sep = spx = new RegExp(this.regexEscape(sp), 'g');
            }
            n = n.replace(spx, '');
          }
          // Replace decimal mark.
          dm = frmt.number.decimal_mark;
          if (dm !== '.') {
            n = n.replace(dm, '.');
          }
          n = parseFloat(n);
          if (!isFinite(n)) {
            return this.parseFailed('number', {
              num: num,
              n: n
            }, 'Tabular.converter.fromLocaleNumber()');
          }
          break;
        default:
          return 0;
      }
      return isFinite(n) ? n : 0;
    },
    /**
     * @private
     * @param {number} num
     * @param {string} type
     * @return {string}
     */
    toLclNmbr: function(num, type) {
      var format = this.format[type] || this.format.number,
      // Work on stringed number instead of (parsed) float, because JS float
      // fraction isnt literal: 0.1 may be interpreted as 0.000000000007.
        n = num && typeof num === 'number' ? ('' + num) : num,
        fract = format.no_fraction, decs = format.n_decimals,
        sign = '', s = '', pos, u;
      if (!n) {
        return '0' + fract;
      }
      // Fraction.
      if ((pos = n.indexOf('.')) > -1) {
        if (type !== 'integer') {
          // Append traling zeros.
          u = (n.substr(pos + 1) + (!decs ? '' : new Array(decs).join('0'))).substr(0, decs);
          // Not all zeros: use decimal and fraction (instead of no_fraction).
          if (!/^0+$/.test(u)) {
            fract = format.decimal_mark + u;
          }
        }
        n = n.substr(0, pos);
      }
      // Sign.
      if (n.charAt(0) === '-') {
        sign = '-';
        n = n.substr(1);
      }
      n = parseInt(n, 10);
      if (!isFinite(n)) {
        return this.parseFailed(type, {
          num: num,
          n: n
        }, 'Tabular.converter.toLclNmbr()');
      }
      // Thousand separator.
      if (format.thousand_sep) {
        while (n > 999) {
          u = n % 1000;
          s = format.thousand_sep + (u > 99 ? '' : (u < 10 ? '00' : '0')) + u + s;
          n = (n - u) / 1000;
        }
      }
      return sign + n + s + fract;
    },
    /**
     * Convert number or stringed number into locale formatted integer string.
     * @param {number|string} num
     * @return {string}
     */
    toLocaleInteger: function(num) {
      return this.toLclNmbr(num, 'integer');
    },
    /**
     * Convert number or stringed number into locale formatted float string.
     * @param {number|string} num
     * @return {string}
     */
    toLocaleFloat: function(num) {
      return this.toLclNmbr(num, 'float');
    },
    /**
     * Convert number or stringed number into locale formatted money string.
     * @param {number|string} num
     * @return {string}
     */
    toLocaleMoney: function(num) {
      return this.toLclNmbr(num, 'money');
    },
    /**
     * Convenience method, relays to toLocaleFloat().
     * @param {number|string} num
     * @return {string}
     */
    toLocaleNumber: function(num) {
      return this.toLocaleFloat(num);
    },
    /**
     * Unix time format characters.
     */
    timeToChars: {
      // RFC 2822: Wdy Mon DD YYYY HH:ii:ss timezone.
      r: 'toString',
      // ISO-8601: YYYY-MM-DDTHH:ii:ss.uuu+ZZ:ZZ.
      // NB: custom algo, because JS Date only supports UTC no-zone ISO-8601.
      c: 'toISOString',
      // ISO-8601: YYYY-MM-DDTHH:ii:ss.uuuZ.
      // NB: PHP date() knows no 'C' char.
      C: 'toISOString',
      // YYYY.
      Y: 'getFullYear',
      // YY.
      y: 'getYear',
      // MM.
      m: 'getMonth',
      // M.
      n: 'getMonth',
      // DD.
      d: 'getDate',
      // D.
      j: 'getDate',
      // HH (24).
      H: 'getHours',
      // H (24).
      h: 'getHours',
      // HH (12).
      G: 'getHours',
      // H (12).
      g: 'getHours',
      // II.
      i: 'getMinutes',
      // SS
      s: 'getSeconds',
      // UUU.
      u: 'getMilliseconds',
      // Seconds since Unix Epoch (1970-01-01).
      U: 'getTime'
    },
    /**
     * Unix time format characters.
     */
    timeFromChars: {
      // RFC 2822: Wdy Mon DD YYYY HH:ii:ss timezone.
      r: 'constructor',
      // ISO-8601: YYYY-MM-DDTHH:ii:ss.uuu+ZZ:ZZ.
      // NB: JS Date only supports UTC no-zone ISO-8601.
      c: 'custom',
      // ISO-8601: YYYY-MM-DDTHH:ii:ss.uuuZ.
      C: 'custom',
      // YYYY.
      Y: 'setFullYear',
      // MM.
      m: 'setMonth',
      // DD.
      d: 'setDate',
      // HH (24).
      H: 'setHours',
      // II.
      i: 'setMinutes',
      // SS
      s: 'setSeconds',
      // UUU.
      u: 'setMilliseconds',
      // Seconds since Unix Epoch (1970-01-01).
      U: 'custom'
    },
    /**
     * ISO-8601 time part methods.
     */
    timeFromIso: [
      'setFullYear',
      'setMonth',
      'setDate',
      'setHours',
      'setMinutes',
      'setSeconds',
      'setMilliseconds'
    ],
    /**
     *
     * @param {string} [type]
     *   Default: 'number'
     * @param {object} [vars]
     *   Log to console, using inspect().
     * @param {string} [msg]
     * @returns {Date|number}
     */
    parseFailed: function(type, vars, msg) {
      var d;
      if (vars) {
        inspect(vars, msg || '');
      }
      else if (msg) {
        inspect.console(msg);
      }
      if (type === 'date') {
        d = new Date(1, 0, 1, 1, 1, 1, 1);
        d.setFullYear(1);
        return d;
      }
      return 0;
    },
    /**
     * @private
     * @param {string} dateStr
     * @param {string} [frmt]
     *   Unix date format.
     *   Default: c (ISO-8601), U (Unix Epoch) or r (RFC 2822).
     * @return {Date}
     */
    fromLclTm: function(dateStr, frmt) {
      var s = dateStr, f = frmt, sLe, fLe, fChrs = this.timeFromChars,
        c, m, l, prt, z, iLe, i, d = new Date();
      if (!f) {
        if (this.rgx.iso8601Minimum.test(s)) {
          // ISO-8601.
          f = 'c';
        }
        else if (this.rgx.unixEpoch.test(s)) {
          // Unix Epoch.
          f = 'U';
        }
        else {
          // RFC 2822.
          f = 'r';
        }
      }
      while ((sLe = s.length) && (fLe = f.length)) {
        c = f.charAt(0);
        prt = '';
        // Supported; with Date method?
        if (fChrs[c] && fChrs.hasOwnProperty(c)) {
          m = fChrs[c];
          switch (c) {
            case 'r':
              return new Date(s);
            // ISO-8601 with timezone (not Z).
            case 'c':
            // ISO-8601 Z (UTC) or no timezone (local).
            case 'C':
              if (!(this.rgx.iso8601.test(s))) {
                return this.parseFailed('date', {
                  dateStr: dateStr,
                  frmt: frmt
                }, 'Tabular.converter.fromLclTm()');
              }
              if (s.charAt(sLe - 1) === 'Z') {
                return new Date(s);
              }
              if (s.indexOf('T') > -1) {
                // Test for +/-NN:NN; JS Date doesn't support timezone.
                if (!(this.rgx.iso8601Zone.test(s))) {
                  return new Date(s);
                }
                d = new Date(s.substr(0, sLe - 6) + 'Z');
                z = s.substr(sLe - 6);
                z = (z.charAt(0) === '+' ? -1000 : 1000) * ((parseInt(z.substr(1, 2)) * 60) + parseInt(z.substr(4, 2)));
                return new Date(d.getTime() + z);
              }
              // In local timezone.
              d = new Date();
              s = s.replace(this.rgx.iso8601, '$1,$2,$3,$4,$5,$6,$7').split(',');
              for (i = 0; i < 7; i++) {
                if ((prt = s[i]) === '') {
                  break;
                }
                prt = parseInt(prt, 10);
                // Month.
                if (i === 1) {
                  --prt;
                }
                d[this.timeFromIso[i]](prt);
              }
              return d;
            // Unix Epoch seconds.
            case 'U':
              s = parseInt(s, 10);
              if (!isFinite(s)) {
                return this.parseFailed('date', {
                  dateStr: dateStr,
                  frmt: frmt
                }, 'Tabular.converter.fromLclTm()');
              }
              return new Date(s * 1000);
            case 'Y':
              l = 4;
              break;
            case 'u':
              l = 3;
              break;
            default:
              l = 2;
          }
          if (sLe >= l) {
            prt = s.substr(0, l);
            if (this.rgx.integer.test(prt)) {
              d[m](parseInt(prt, 10));
              s = s.substr(l);
              f = f.substr(1);
              continue;
            }
          }
          // Failed.
          return this.parseFailed('date', {
            dateStr: dateStr,
            frmt: frmt,
            s: s,
            f: f
          }, 'Tabular.converter.fromLclTm()');
        }
        else {
          // Remove separator.
          s = s.substr(1);
          f = f.substr(1);
        }
      }
      return d;
    },
    /**
     * @param {string} timestampStr
     *   Must be formatted according to converter_format.time.timestamp.
     * @return {string}
     */
    fromLocaleTimestamp: function(timestampStr) {
      return this.fromLclTm(timestampStr, this.format.time.timestamp);
    },
    /**
     * @param {string} datetimeStr
     *   Must be formatted according to converter_format.time.datetime.
     * @return {string}
     */
    fromLocaleDatetime: function(datetimeStr) {
      return this.fromLclTm(datetimeStr, this.format.time.datetime);
    },
    /**
     * @param {string} dateStr
     *   Must be formatted according to converter_format.time.date.
     * @return {string}
     */
    fromLocaleDate: function(dateStr) {
      return this.fromLclTm(dateStr, this.format.time.date);
    },
    /**
     * @param {string} timeStr
     *   Must be formatted according to converter_format.time.time.
     * @return {string}
     */
    fromLocaleTime: function(timeStr) {
      return this.fromLclTm(timeStr, this.format.time.time);
    },
    /**
     * @private
     * @param {Date|string} date
     *   String: must be ISO-8601 (supposedly in local time if incomplete),
     *   Unix Epoch or RFC 2822.
     * @param {string} frmt
     *   Unix date format. Ignored if arg date is string.
     * @return {string}
     */
    toLclTm: function(date, frmt) {
      var d = date, f = frmt.split(''), le = f.length, fChrs = this.timeToChars,
        i, c, l, u, z, zh, s = '';
      if (typeof d === 'string') {
        // Must be ISO-8601, Unix Epoch or RFC 2822.
        d = this.fromLclTm(d);
      }
      for (i = 0; i < le; i++) {
        c = f[i];
        // Supported; with Date method?
        if (fChrs[c] && fChrs.hasOwnProperty(c)) {
          l = 0;
          u = d[fChrs[c]]();
          switch (c) {
            // ISO-8601 with timezone (not Z).
            case 'C':
              z = d.getTimezoneOffset();
              u = new Date(d.getTime() - (z * 60 * 1000)).toISOString();
              u = u.substr(0, u.length - 1);
              if (z > 0) {
                u += '-';
              }
              else {
                u += '+';
                z *= -1;
              }
              zh = Math.floor(z / 60);
              u += ('0' + zh).substr(-2) + ':' + ('0' + (z - (zh * 60))).substr(-2);
              break;
            // Month.
            case 'm':
              u += 1;
              l = 2;
              break;
            case 'n':
              u += 1;
              break;
            case 'y':
            case 'd':
            case 'H':
            case 'G':
            case 'i':
            case 's':
              l = 2;
              break;
            // Milliseconds.
            case 'u':
              l = 3;
              break;
            case 'U':
              u = Math.round(u / 1000);
              break;
          }
          // Prefix leading zeros.
          if (l) {
            u = (new Array(l).join('0') + u).substr(-l);
          }
          s += u;
        }
        else {
          s += c;
        }
      }
      return s;
    },
    /**
     * @param {Date|string} date
     *   String: must be ISO-8601 (supposedly in local time if incomplete),
     *   Unix Epoch or RFC 2822.
     * @return {string}
     */
    toLocaleTimestamp: function(date) {
      return this.toLclTm(date, this.format.time.timestamp);
    },
    /**
     * @param {Date|string} date
     *   String: must be ISO-8601 (supposedly in local time if incomplete),
     *   Unix Epoch or RFC 2822.
     * @return {string}
     */
    toLocaleDatetime: function(date) {
      return this.toLclTm(date, this.format.time.datetime);
    },
    /**
     * @param {Date|string} date
     *   String: must be ISO-8601 (supposedly in local time if incomplete),
     *   Unix Epoch or RFC 2822.
     * @return {string}
     */
    toLocaleDate: function(date) {
      return this.toLclTm(date, this.format.time.date);
    },
    /**
     * @param {Date|string} date
     *   String: must be ISO-8601 (supposedly in local time if incomplete),
     *   Unix Epoch or RFC 2822.
     * @return {string}
     */
    toLocaleTime: function(date) {
      return this.toLclTm(date, this.format.time.time);
    }
  };
  Tabular.extend(Tabular.converter, 'Tabular.converter', Tabular.base);

}(jQuery));
