/**
 * @file
 * Tabular (tabular) Drupal behavior subscriptions.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Creates JS representation of a solitary field or a group of sources plus
   * a target field.
   *
   * For non-solitary groups initialisation gets aborted and postponed if
   * some tabular_group_source element(s) isn't available in DOM yet.
   *
   * Behaviour callback.
   *
   * @returns {boolean|undefined}
   */
  Tabular.initField = function() {
    var groupElm = this,
      groupId, group,
      discoveryMode, nSourcesConfigured, $parent,
      sources, le, i, j, nSources, weights = [], weightedSources = [];

    // This function MUST NOT alter DOM because it may abort before creating
    // a field.
    // Well, a target with discovery mode 'sibling' must actually.

    // Solitary tabular.
    if (groupElm.tagName.toLowerCase() === 'tabular_group_solitary') {
      // A solitary widget has no group ID in it's backend widget settings.
      // And shouldn't have - because that would break things if multiple
      // instances of the field appears in the DOM.
      // But it still needs a group ID so that we can access it's controller
      // via the Tabular.widgetGroups list.
      groupId = Judy.randName();
      group = new Tabular.widgetGroup(groupId, [groupElm]);
      // The widgetGroup has no groupId property if something fails during
      // instantiation.
      if (group.groupId) {
        Tabular.widgetGroups[groupId] = group;
        // Setup interaction between the group and it's widgets.
        group.initInteraction();
        // Log group instance if example.
        if (group.entityKind === 'example') {
          inspect.log(Tabular.widgetGroups[groupId], 'Tabular example solitary group');
        }
      }
    }
    // Target in group.
    else {
      // Grouped widgets have a hardcoded group ID in their backend settings.
      groupId = groupElm.getAttribute('tabular-group-id');
      discoveryMode = groupElm.getAttribute('tabular-discovery-mode');
      // Discover in-group source widgets by group ID.
      if (discoveryMode === 'group_id') {
        sources = $('tabular_group_source[tabular-group-id="' + groupId + '"]').get();
        nSourcesConfigured = parseInt(groupElm.getAttribute('tabular-group-nsources'), 10);
        le = sources.length;
        // Get out and wait for next Drupal.behaviors trigger.
        if (le < nSourcesConfigured) {
          return;
        }
        else if (le > nSourcesConfigured) {
          inspect.log(
            {
              'field wrapper': Tabular.findFieldWrapper(groupElm).get(0),
              configured: nSourcesConfigured,
              actual: le
            },
            {
              type: groupElm.getAttribute('tabular-module') || 'tabular',
              message: 'Target widget has configured less sources than actually exist',
              severity: 'error'
            }
          );
          return;
        }
      }
      // Discover in-group source widgets as siblings under same parent.
      else {
        $parent = Tabular.findCollectionWrapper(groupElm);
        if (!$parent.length) {
          $parent = $(self.selectors.field).closest('form');
        }
        sources = $parent.find('tabular_group_source').get();
        le = sources.length;
        // If no sources found within parent, all sources might be 'occupied'
        // by other targets. Thus it may not be an error.
        if (!le) {
          // Don't try again.
          $(groupElm).remove();
          return;
        }
        // Group ID may be empty in discoveryMode:sibling.
        if (groupId === '') {
          groupId = Judy.randName();
          groupElm.setAttribute('tabular-group-id', groupId);
        }
      }
      // Order by weight.
      for (i = 0; i < le; i++) {
        weights.push(sources[i].getAttribute('tabular-group-weight'));
      }
      weights.sort();
      nSources = le;
      for (i = 0; i < le; i++) {
        for (j = 0; j < nSources; j++) {
          if (sources[j].getAttribute('tabular-group-weight') === weights[i]) {
            weightedSources.push(sources.splice(j, 1)[0]);
            --nSources;
            break;
          }
        }
      }
      // Pass to new Tabular widget group controller.
      // Dupe group ID?
      if ((group = Tabular.widgetGroups[groupId])) {
        if (!Tabular.settings('allow_group_overwrite')) {
          inspect.log(
            groupId,
            {
              type: groupElm.getAttribute('tabular-module') || 'tabular',
              message: 'Multiple instances of same widget group, has already created a field by/for that group ID',
              code: Tabular.errorCode('multiple_identical_groups'),
              severity: 'error'
            }
          );
          Tabular.alertDialog(
            Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode}),
            'error'
          );
          // Break out of jQuery.each() loop
          // in Drupal.behaviors.tabularInit
          return false;
        }
        else {
          group.destroy();
        }
      }
      group = new Tabular.widgetGroup(groupId, weightedSources, groupElm);
      // The widgetGroup has no groupId property if something fails during
      // instantiation.
      if (group.groupId) {
        Tabular.widgetGroups[groupId] = group;
        // Setup interaction between the group and it's widgets.
        group.initInteraction();
      }
      // Log group instance if example.
      if (group.entityKind === 'example') {
        inspect.log(Tabular.widgetGroups[groupId], 'Tabular example widget group');
      }
    }
  };


  Drupal.behaviors.tabularInit = {
    attach: function() {

      // Init solitary tabular widgets.
      // Tabular.widget.base sets tabular-group-prepared attribute
      // on all group flag elements.
      $('tabular_group_solitary').not('[tabular-group-prepared]').
        each(Tabular.initField);

      // Init groups of tabular source(s) and target.
      $('tabular_group_target').not('[tabular-group-prepared]').
        each(Tabular.initField);
    }
  };
}(jQuery));
