/**
 * @file
 * Tabular (tabular) status list.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Fades and eventually hides or removes an element.
   *
   *  Options and their defaults:
   *  - (float) opacity: initial opacity (100)
   *  - (bool) fast: a fast combination of delay+pause+factor (false)
   *  - (bool) slow: default delay+pause+factor (true)
   *  - (int) delay: in millisecs. (1000)
   *  - (int) pause: in millisecs. (100)
   *  - (float) factor: amount of change for every iteration (1.2)
   *  - (bool) remove: remove the element instaed of hiding it (false)
   *
   * @example var fader = new Drupal.Tabular.fader('#whatever').start();
   * setTimeout(function(){fader.stop()}, 5000);
   * setTimeout(function(){fader.reset()}, 8000);
   *
   * @param {string} selector
   * @param {object|string} [options]
   *   String: fast|remove.
   */
  Tabular.fader = function(selector, options) {
    var k, o = options, useFast, remove,
      defaults = {
        // Initial opacity.
        opacity: 100,
        delay: 1000,
        pause: 100,
        // Amount of change for every interval.
        factor: 1.2,
        remove: false
      },
      fast = {
        delay: 50,
        pause: 50,
        factor: 1.5
      };
    if (o) {
      if (o === 'fast') {
        o = fast;
      }
      else if (o === 'remove') {
        remove = true;
        o = null;
      }
      else if (typeof o === 'object') {
        if (o.fast && o.hasOwnProperty('fast')) {
          useFast = true;
        }
        delete o.fast;
        delete o.slow;
      }
      else {
        o = null;
      }
    }
    for (k in defaults) {
      if (defaults.hasOwnProperty(k)) {
        this[k] = o && o.hasOwnProperty(k) ? options[k] : (useFast && fast.hasOwnProperty(k) ? fast[k] : defaults[k]);
      }
    }
    if (remove) {
      this.remove = true;
    }

    this.selector = selector;
    this.subtractor = 1;
    this.initialOpacity = this.opacity;
    this.started = this.stopped = false;
  };
  /**
   * Convenience method (no 'new' contructor) which also starts the fading.
   * @param {string} selector
   * @param {object} [options]
   * @returns {Drupal.Tabular.fader}
   */
  Tabular.fader.make = function(selector, options) {
    return new Tabular.fader(selector, options).start();
  };
  Tabular.fader.prototype = {
    constructor: Tabular.fader,
    /**
     * @param {number} [delay]
     *   Integer
     * @returns {Drupal.Tabular.fader}
     *   Self.
     */
    start: function(delay) {
      var self = this;
      if (!self.started) {
        self.stopped = false;
        self.timeout = setTimeout(function() {
          try {
            self.interval = setInterval(function() {
              try {
                self.fade();
              }
              catch (ignore) {
              }
            }, self.pause);
          }
          catch (ignore) {
          }
        }, delay || self.delay);
        self.started = true;
      }
      return self;
    },
    /**
     * @returns {Drupal.Tabular.fader}
     *   Self.
     */
    stop: function() {
      var self = this;
      if (self.started && !self.stopped) {
        self.stopped = true;
        if (self.timeout) {
          clearTimeout(self.timeout);
        }
        if (self.interval) {
          clearInterval(self.interval);
        }
      }
      return self;
    },
    /**
     * @returns {Drupal.Tabular.fader}
     *   Self.
     */
    reset: function() {
      var self = this, $jq;
      if (self.started) {
        self.stop();
        if (($jq = $(self.selector)).length) {
          if (self.opacity < self.initialOpacity) {
            $jq.css('opacity', self.initialOpacity / 100).show();
          }
        }
        else {

        }
        self.started = false;
      }
      return self;
    },
    /**
     * @private
     */
    fade: function() {
      var self = this, $jq;
      if (!self.stopped && ($jq = $(self.selector)).length) {
        if ((self.opacity -= (self.subtractor *= self.factor)) > 0) {
          $jq.css('opacity', self.opacity / 100);
        }
        else {
          self.stop();
          if (!self.remove) {
            $jq.hide();
          }
          else {
            $jq.remove();
            self.destroy();
          }
        }
      }
    }
  };
  Tabular.extend(Tabular.fader, 'Tabular.fader', Tabular.base);

  /**
   * Display messages in a list of small popups.
   *
   * @param {string} listSelector
   */
  Tabular.statusList = function(listSelector) {
    this.listSelector = listSelector;
    this.index = -1;
    this.last = '';
    this.faders = {};
  };
  Tabular.statusList.prototype = {
    constructor: Tabular.statusList,
    itemTemplate: '<div id="__ID__" class="status-__TYPE__">' +
      '<div><span>__CONTENT__</span></div><div title="' + Drupal.t('Close') + '">__CLOSE__</div></div>',
    /**
     * Add message to the top of the list.
     *
     *  Options:
     *  - (bool) noFade
     *  - (int) fadeDelay
     *  - (bool) closeButton
     *  - (func) closeHandler
     *  - (bool) modal
     *
     *  Defaults by type:
     *  - status: fade after 1 second, no close button
     *  - info: fade after 2 seconds, no close button
     *  - notice: fade after 5 seconds, with close button
     *  - warning|error: no fading, with close button
     *
     * If the type and message is the same as the previous message,
     * that previous message will be hidden.
     *
     * @example var statusList = new Tabular.statusList('div.my-status-list);
     * statusList.add('A-OK');
     * statusList.add('Emph', 'notice');
     * statusList.add('Oops', 'error');
     *
     * @param {string} msg
     * @param {string} [type]
     *   Default: status.
     *   Values: status|info|notice|warning|error.
     * @param {object|undefined} [options]
     * @returns {Object}
     */
    add: function(msg, type, options) {
      var self = this, tp = type || 'status', selector, list, current, $item, $close, fader, k, fadeDelay = 1000, o = {
        noFade: true,
        fadeDelay: fadeDelay,
        closeButton: true,
        closeHandler: undefined,
        modal: false
      };
      switch (tp) {
        case 'status':
          o.noFade = o.closeButton = false;
          break;
        case 'info':
          o.noFade = o.closeButton = false;
          o.fadeDelay = fadeDelay * 2;
          break;
        case 'notice':
          o.noFade = false;
          o.fadeDelay = fadeDelay * 5;
          break;
      }
      if (o.modal) {
        o.closeButton = true;
        o.noFade = true;
      }
      if ((list = Tabular.element(self.listSelector))) {
        if (options) {
          for(k in o) {
            if(o.hasOwnProperty(k) && options.hasOwnProperty(k)) {
              o[k] = options[k];
            }
          }
        }
        // Remove previous from DOM if same as current.
        current = type + msg;
        if (current === self.last) {
          $('#' + list.id + '-item-' + self.index).remove();
        }
        self.last = current;
        ++self.index;
        // Add message to DOM.
        selector = list.id + '-item-' + self.index;
        $item = $(
          self.itemTemplate.
            replace('"__ID__"', '"' + selector + '"').
            replace('__NO__"', self.index + '"').
            replace('__TYPE__"', tp + '"').
            replace('__CONTENT__', msg.replace(/\n/g, '<br/>')).
            replace('__CLOSE__', !o.closeButton ? '' : 'x')
        ).prependTo(list);
        $item.id = selector;
        selector = '#' + selector;

        // Close button click handlers; even if hidden.
        $close = $item.find('> div:last-child').click(function() {
          $(this.parentNode).hide();
        });
        if (o.modal) {
          $close.click(Judy.overlay);
          // Opaque, no hover title.
          Judy.overlay(1, true);
        }
        if (o.closeHandler) {
          $close.click(o.closeHandler);
        }

        if (!o.noFade) {
          self.faders['' + self.index] = fader = new Tabular.fader(selector).start(o.fadeDelay);
          // Make click on message content unfade the message.
          $item.find('> div:first-child').click(function() {
            fader.reset();
            // Display close button.
            $(this).next().html('x');
          });
        }
        $item.show();
      }
      return self;
    },
    /**
     * Show all list items.
     */
    showAll: function() {
      var self = this, list, selector, last, i, fader;
      if ((list = Tabular.element(self.listSelector))) {
        selector = list.id + '-item-';
        // In reverse order; oldest first.
        last = self.index;
        for (i = last; i > -1; i--) {
          if ((fader = self.faders['' + i])) {
            fader.reset();
          }
          else {
            $(selector + i).show();
          }
        }
      }
    }
  };
  Tabular.extend(Tabular.statusList, 'Tabular.statusList', Tabular.base);

}(jQuery));
