/**
 * @file
 * Tabular (tabular) widget.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Widget builder classes.
   *
   * @type {object}
   */
  Tabular.widget = {};

  /**
   * Abstract base class for all widget classes.
   * @constructor
   */
  Tabular.widget.base = function() {
    throw new Error('Abstract class ' + this.type_of + ' is not to be instantiated.');
  };
  Tabular.widget.base.prototype = {
    constructor: Tabular.widget.base,
    /**
     *  Values:
     *  0: initializing
     *  1: init error
     *  2: fatal runtime error
     *  3: self-destroying
     *  4: sound and healthy
     *
     * @type {number}
     */
    state: 0,
    /**
     * Set up what's generic across solitary|source|target and widget type.
     * @throws {Error}
     * @param {string} groupId
     * @param {HTMLElement} groupElm
     * @returns {object}
     *   Self.
     */
    initBase: function(groupId, groupElm) {
      var self = this, errorCode, msg, $field, fieldId, nmSp,
        $root, elm, o, u, col, fNm;
      if (self.state) {
        return self;
      }
      try {
        /** @see Tabular.widget.base#blockExecution */
        self.blockExecutions = {};

        // Make ignored by Drupal.behaviors.tabularInit.
        groupElm.setAttribute('tabular-group-prepared', '1');

        self.groupId = groupId;
        self.widgetName = groupElm.getAttribute('tabular-widget-name');
        self.module = groupElm.getAttribute('tabular-module');
        self.entityKind = groupElm.getAttribute('tabular-entity-kind');
        self.conf = JSON.parse(groupElm.getAttribute('tabular-widget-ctlr-conf'));

        // Set group ID on field wrapper.
        $field = Tabular.findFieldWrapper(groupElm);
        $field.attr('tabular-group-id', groupId);
        fieldId = $field.get(0).id;
        // Remove group element entirely; we're done with it.
        $(groupElm).remove();

        // Declare selector namespace: field wrapper ID, with 'edit-' replaced
        // with 'nmspc-'.
        self.selectors = {
          namespace: (nmSp = 'nmspc-' + fieldId.substr(5))
        };
        // Memorize field Wrapper selector.
        self.selectors.field = '#' + fieldId;

        // Find field label.
        elm = $field.find('label.tabular-field-label').first().get(0);
        self.selectors.label = '#' + (elm.id = (nmSp + '--tabular-label'));

        // Declare inline event expression for setting (this widget as)
        // last useraction widget.
        self.inlineEventSetLastActionWidget =
          'Drupal.Tabular.setLastUserActionWidget(\'' + self.groupId + '\',\'' + self.widgetName + '\');';

        // Find root element.
        elm = ($root = $field.find('div.tabular-root').first()).get(0);
        // Set last user action inline event handlers on root.
        elm.setAttribute('onclick', self.inlineEventSetLastActionWidget);
        elm.setAttribute('onkeydown', self.inlineEventSetLastActionWidget);
        // Set HTML ID on root.
        self.selectors.root = '#' + (elm.id = (nmSp + '--tabular-root'));

        // Error message container.
        elm = $field.find('div.tabular-messages-error').first().get(0);
        self.selectors.errorMsgCntnr = '#' + (elm.id = (nmSp + '--tabular-messages-error'));

        // Find status dialog list, within root, and set HTML ID on it.
        elm = $root.find('div.tabular-status-list').first().get(0);
        self.selectors.statusList = '#' + (elm.id = nmSp + '--tabular-status-list');
        self.statusList = new Tabular.statusList(self.selectors.statusList);

        // Find view container, within root, and set HTML ID on it.
        elm = $root.find('div.tabular-view-container').first().get(0);
        self.selectors.viewContainer = '#' + (elm.id = nmSp + '--tabular-view-container');

        // Find field description, if any.
        elm = $root.find('div.description.tabular-field-description').first().get(0);
        if (elm) {
          self.selectors.description = '#' + (elm.id = (nmSp + '--tabular-description'));
        }

        // Set some buckets in viewColumns, buttonsBefore and buttonsAfter,
        // to avoid iterative hasOwnProperty() checks.
        o = self.conf.schema.viewColumns;
        for (col in o) {
          // Resolve schema view_columns convert functions.
          if (o.hasOwnProperty(col)) {
            if (o[col].hasOwnProperty('convert')) {
              // Clear ref.
              u = null;
              u = Tabular.evalMethod(fNm = o[col].convert);
              if (!u) {
                throw new Error('viewColumn[' + col + '] convert[' + fNm + '] won\'t eval to function.').
                customCode(Tabular.errorCode('non_resolvable_js_function'));
              }
              o[col].convert = u;
            }
            else {
              o[col].convert = undefined;
            }
            // Pure define.
            if (!o[col].hasOwnProperty('hover_title')) {
              o[col].hover_title = undefined;
            }
            if (!o[col].hasOwnProperty('sort')) {
              o[col].sort = undefined;
            }
          }
        }
        // Clear ref.
        o = null;
        if ((o = self.conf.schema.buttonsBefore)) {
          for (col in o) {
            if (o.hasOwnProperty(col)) {
              // Pure define.
              if (!o[col].hasOwnProperty('hover_title')) {
                o[col].hover_title = undefined;
              }
              if (!o[col].hasOwnProperty('click')) {
                o[col].click = undefined;
              }
            }
          }
        }
        // Clear ref.
        o = null;
        if ((o = self.conf.schema.buttonsAfter)) {
          for (col in o) {
            if (o.hasOwnProperty(col)) {
              // Pure define.
              if (!o[col].hasOwnProperty('hover_title')) {
                o[col].hover_title = undefined;
              }
              if (!o[col].hasOwnProperty('click')) {
                o[col].click = undefined;
              }
            }
          }
        }

        return self;
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_widget');
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // Display error to user, unless other previous initialization error.
        if (!self.state) {
          // Flag initialization error.
          self.state = 1;
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
        // Re-throw the error but tell caller to abort; don't log (again) etc.
        er.abortSilently = true;
        throw er;
      }
    },
    /**
     * Secures that data is object (not array) and that every row contains
     * a primary key bucket.
     *
     * If a row doesn't contain primary key, the row's key in the data list
     * will be set as primary key within the row.
     *
     * Prefixes data list key with underscore to prevent arbitrary sorting
     * of stringed numeric keys.
     *
     * @see Tabular.widget.value#initValue
     * @see Tabular.widget.source#populateProcede
     *
     * @throws {TypeError}
     *
     * @param {object|Array} data
     * @returns {object}
     */
    normalizeData: function(data) {
      var self = this, primaryKey = self.conf.schema.primaryKey, nrml = {}, le, i, k;
      if (data) {
        if (typeof data !== 'object') {
          throw new TypeError('Data type[' + inspect.typeOf(data) + '] is not object or array.').
            customCode(Tabular.errorCode('use'))
        }
        if ($.isArray(data)) {
          le = data.length;
          for (i = 0; i < le; i++) {
            // Row has primary key bucket: use that value as data list key.
            if (data[i].hasOwnProperty(primaryKey)) {
              // Prefix data list key with underscore, to prevent arbitrary
              // sorting of stringed numeric keys.
              nrml['_' + data[i][primaryKey]] = data[i];
            }
            // Primary key value must be the data list key:
            // set that key as primary key in the row.
            else {
              data[i][primaryKey] = i;
              nrml['_' + i] = data[i];
            }
          }
        }
        else {
          for (k in data) {
            if (data.hasOwnProperty(k)) {
              if (data[k].hasOwnProperty(primaryKey)) {
                nrml['_' + data[k][primaryKey]] = data[k];
              }
              else {
                data[k][primaryKey] = k;
                nrml['_' + k] = data[k];
              }
            }
          }
        }
      }
      return nrml;
    },
    eventHandlerRegexes : {
      entityKind: /_ENTITY_KIND_/g,
      groupId: /_GROUP_ID_/g,
      widgetName: /_WIDGET_NAME_/g,
      primaryKey: /_PRIMARY_KEY_/g
    },
    /**
     * Creates a represention of a data set row
     * according to schema view columns and buttons.
     *
     * Extending widget[view type] needn't override.
     *
     * @throws {RangeError}
     *
     * @param {string} key
     * @param {object} dataRow
     * @returns {object}
     *   {click: string, view: array}
     */
    viewRow: function(key, dataRow) {
      var self = this, errorCode,
        rowClick = self.conf.rowClick,
        schm = self.conf.schema, primaryKey = schm.primaryKey,
        viewColumns = schm.viewColumns, buttonsBefore = schm.buttonsBefore, buttonsAfter = schm.buttonsAfter,
        viewRow = [], values = {}, col,
        nColsnButs, iColsnButs, str, rgx = /\!([a-zA-Z\d_\-]+)/g, mtchs, le, i;

      if (buttonsBefore) {
        for (col in buttonsBefore) {
          if (buttonsBefore.hasOwnProperty(col)) {
            // (on)click attribute supports simple token replacers:
            // _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
            if ((str = buttonsBefore[col].click)) {
              str = self.replaceInlineEventHandlerExpressions(str).
                replace(self.eventHandlerRegexes.entityKind, self.entityKind).
                replace(self.eventHandlerRegexes.groupId, self.groupId).
                replace(self.eventHandlerRegexes.widgetName, self.widgetName).
                replace(self.eventHandlerRegexes.primaryKey, dataRow[primaryKey]);
            }
            viewRow.push({
              isButton: true,
              hover_title: buttonsBefore[col].hover_title,
              click: str,
              label: buttonsBefore[col].label,
              name: col
            });
          }
        }
      }
      for (col in viewColumns) {
        if (viewColumns.hasOwnProperty(col)) {
          // Skip the data row if it misses a column specified by schema.
          if (!dataRow.hasOwnProperty(col)) {
            errorCode = Tabular.errorCode('data_column_nonexist');
            inspect.log(
              {
                'entity kind': self.entityKind,
                'dataSet key': key,
                'row column': col,
                row: dataRow
              },
              {
                type: self.module,
                message: 'Data row misses column specified by schema.viewColumns.',
                code: errorCode,
                severity: 'error'
              }
            );
            throw new RangeError('Data row misses column specified by schema.viewColumns.').
              customCode(errorCode);
          }
          viewRow.push({
            hover_title: viewColumns[col].hover_title,
            raw: values[col] = dataRow[col],
            /**
             * @see Tabular.evalMethod
             */
            converted: !viewColumns[col].convert ? undefined :
              // .convert[0] if object (or falsy), .convert[1] is function.
              (values[col] = (viewColumns[col]['convert'][0] || window)[viewColumns[col]['convert'][1]](dataRow[col], dataRow)),
            name: col
          });
        }
      }
      if (buttonsAfter) {
        for (col in buttonsAfter) {
          if (buttonsAfter.hasOwnProperty(col)) {
            // (on)click attribute supports simple token replacers:
            // _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
            if ((str = buttonsAfter[col].click)) {
              str = self.replaceInlineEventHandlerExpressions(str).
                replace(self.eventHandlerRegexes.entityKind, self.entityKind).
                replace(self.eventHandlerRegexes.groupId, self.groupId).
                replace(self.eventHandlerRegexes.widgetName, self.widgetName).
                replace(self.eventHandlerRegexes.primaryKey, dataRow[primaryKey]);
            }
            viewRow.push({
              isButton: true,
              hover_title: buttonsAfter[col].hover_title,
              click: str,
              label: buttonsAfter[col].label,
              name: col
            });
          }
        }
      }
      // Fix hover title replacements; cannot be done until established
      // all column values.
      nColsnButs = viewRow.length;
      for (iColsnButs = 0; iColsnButs < nColsnButs; iColsnButs++) {
        if ((str = viewRow[iColsnButs].hover_title)) {
          // View columns only.
          if (str === '_VALUE_') {
            str = viewRow[iColsnButs].converted !== undefined ? viewRow[iColsnButs].converted : viewRow[iColsnButs].raw;
          }
          else if (str === '_RAW_') {
            str = viewRow[iColsnButs].raw;
          }
          // View columns and buttons.
          else if (str.indexOf('!') !== -1) {
            mtchs = Tabular.matchSubPattern(rgx, str);
            if ((le = mtchs.length)) {
              for (i = 0; i < le; i++) {
                str = str.replace('!' + mtchs[i], values.hasOwnProperty(mtchs[i]) ? values[mtchs[i]] : dataRow[mtchs[i]]);
              }
            }
          }
          viewRow[iColsnButs].hover_title = str;
        }
      }
      // Metadata.
      if (rowClick) {
        rowClick = self.replaceInlineEventHandlerExpressions(rowClick).
          replace(self.eventHandlerRegexes.entityKind, self.entityKind).
          replace(self.eventHandlerRegexes.groupId, self.groupId).
          replace(self.eventHandlerRegexes.widgetName, self.widgetName).
          replace(self.eventHandlerRegexes.primaryKey, dataRow[primaryKey]);
      }
      return {
        click: rowClick,
        view: viewRow
      };
    },
    /**
     * Replaces magic inline expressions with their 'real' counterparts.
     * @see Tabular.widget.source#inlineEventHandlerExpressions
     * @see Tabular.widget.value#inlineEventHandlerExpressions
     * @param {string} str
     * @returns {string}
     */
    replaceInlineEventHandlerExpressions: function(str) {
      var rplcrs = this.inlineEventHandlerExpressions, k, s = str;
      for (k in rplcrs) {
        if (rplcrs.hasOwnProperty(k)) {
          s = s.replace(k, rplcrs[k]);
        }
      }
      return s;
    },
    /**
     * Empty the widget's view container.
     * NB: only this method is allowed to empty the view container, because it
     * removes all (non-inline) event handlers.
     * @return {jquery}
     *   The view container; for convenience.
     */
    clearView: function() {
      var self = this, $viewCntnr = $(self.selectors.viewContainer), viewWrpr;
      if ($viewCntnr.length) {
        // Remove event listeners.
        // Non-searchable dropdown collapse when view loses focus.
        /**
         * @see Tabular.widget.base#viewEffectsApply
         */
        if (self.isSource && self.conf.dropdown && !self.conf.provider.searchable) {
          viewWrpr = $viewCntnr.children(self.viewTypeSelectors.wrapper).first().get(0);
          if (viewWrpr) {
            viewWrpr.get(0).removeEventListener('blur', self.dropdownViewCollapse, true);
          }
        }

        $viewCntnr.empty();
        return $viewCntnr;
      }
      return $();
    },
    /**
     * Inline event handler expression for stopping event propagation (bubbling)
     * @type {string}
     */
    inlineEventStopPropagation: 'var evt=arguments[0]||window.event;evt.cancelBubble=true;if(evt.stopPropagation){evt.stopPropagation();}',
    /**
     * Inline event handler expression for setting last user action widget.
     * Gets set on widget initialization.
     * @type {string}
     */
    inlineEventSetLastActionWidget: '',
    /**
     * Must be overridden, by view type specializing class (like table).
     * @type {string}
     */
    viewType: 'none',
    viewTypeSelectors: {
      wrapper: 'div.tabular-view-wrapper',
      view: 'div[tabular-widget-view="1"]',
      row: 'div'
    },
    /**
     * Must be overridden, by view type specializing class (like table).
     * Tag starts must not terminate (with >).
     * @type {string}
     */
    viewTypeHtml: {
      viewEmptyStart: '<div tabular-widget-view="1" tabular-widget-view-empty="1"',
      viewEmptyEnd: '</div>',
      // tabindex to be focusable.
      viewStart: '<div class="tabular-view-wrapper" tabindex="0"><div tabular-widget-view="1" tabindex="0"',
      viewEnd: '</div></div>',
      headStart: '<div',
      headEnd: '</div>',
      headRowStart: '<div',
      headRowEnd: '</div>',
      headColStart: '<span',
      headColEnd: '</span>',
      bodyStart: '<div',
      bodyEnd: '</div>',
      bodyRowStart: '<div tabindex="0"',
      bodyRowEnd: '</div>',
      bodyColStart: '<span',
      bodyColEnd: '</span>'
    },
    /**
     * Target only displays empty if conf displayEmpty.
     * @returns {string}
     */
    renderEmpty: function() {
      var self = this, html = self.viewTypeHtml;
      if (self.isSource) {
        // Source shan't render empty if searchable.
        if (!self.conf.provider.searchable) {
          return html.viewEmptyStart + '>' + Tabular.message(self.entityKind, 'provider_no_results') + html.viewEmptyEnd;
        }
      }
      else if (self.conf.displayEmpty) {
        return html.viewEmptyStart + '>' + Tabular.message(self.entityKind, 'none_selected') + html.viewEmptyEnd;
      }
      return '';
    },
    /**
     * Must be overridden, by view type specializing class (like table).
     * @see Tabular.widget.base#renderEmpty
     * @see Tabular.widget.table#renderView
     * @param {object} dataSet
     * @param {object} firstViewRow
     *   Reference needed in callers scope.
     * @returns {string}
     */
    renderView: function(dataSet, firstViewRow) {
      throw new Error('Abstract method of ' + this.type_of + '.');
    },
    /**
     * Must be overridden, by view type specializing class (like table).
     * @see Tabular.widget.table#renderLegend
     * @param {Array} viewRow
     *   First view row of a data set.
     * @returns {string}
     */
    renderLegend: function(viewRow) {
      throw new Error('Abstract method of ' + this.type_of + '.');
    },
    /**
     * Must be overridden, by view type specializing class (like table).
     * @see Tabular.widget.table#renderRow
     * @param {Array} viewRow
     * @returns {string}
     */
    renderRow: function(viewRow) {
      throw new Error('Abstract method of ' + this.type_of + '.');
    },
    /**
     * Apply special effects on the view when added to DOM.
     * @see Tabular.widget.table#viewEffectsApply
     * @param {object} firstViewRow
     */
    viewEffectsApply: function(firstViewRow) {
      var self = this, $viewWrpr;
      // Make non-searchable dropdown collapse when view loses focus.
      if (self.isSource && self.conf.dropdown && !self.conf.provider.searchable) {
        $viewWrpr = $(self.selectors.viewContainer + ' > ' + self.viewTypeSelectors.wrapper);
        if ($viewWrpr.length) {
          // Use event capture because Firefox doesn't support the (bubbling)
          // focusout event. jQuery doesn't support event capturing.
          $viewWrpr.get(0).addEventListener('blur', self.dropdownViewCollapse, true);
        }
      }
    },
    /**
     * Empty view container and display progress spinner.
     */
    displayProgress: function() {
      var self = this;
      self.clearView().append(
        '<div class="ajax-progress">' + Drupal.t('Loading...') + '<div class="throbber"></div></div>'
      );
    },
    /**
     * Display message in Drupal message-like container placed within
     * the widgets field wrapper.
     * Filters repeated message away.
     * @param {string} msg
     * @return {boolean}
     *   False if can't find message container or it cannot be displayed.
     */
    displayErrorMessage: function(msg) {
      var self = this, slctrs = self.selectors, $cntnr, $list, cntnt;
      if (slctrs && slctrs.root && slctrs.errorMsgCntnr &&
        $(slctrs.root).showByHidden().show().filter(':visible').first().length
      ) {
        $cntnr = $(slctrs.errorMsgCntnr).showByHidden().show().filter(':visible').first();
        if ($cntnr.length) {
          if (self.errorMessages) {
            if ($.inArray(msg, self.errorMessages) === -1) {
              self.errorMessages = [msg];
            }
            else {
              return true;
            }
          }
          else {
            self.errorMessages = [msg];
          }
          $list = $cntnr.find('ul').first();
          if ($list.length) {
            $list.append('<li>' + msg + '</li>');
          }
          else {
            cntnt = $.trim($cntnr.html());
            if (cntnt) {
              $cntnr.html('<ul><li>' + cntnt + '</li><li>' + msg + '</li></ul>');
            }
            else {
              $cntnr.append(msg);
            }
          }
          return true;
        }
      }
      return false;
    },
    /**
     * Adds common attributes to custom option/value change events.
     * @private
     * @param {object} event
     *   Synthetic event object.
     * @param {Array} objFuncArgs
     *   [object or falsy, function, array or falsy].
     */
    dispatchEvent: function(event, objFuncArgs) {
      var args = objFuncArgs[2] || [], frst = args[0];
      if (frst === undefined || frst === null) {
        event.tabular_group_id = this.groupId;
        event.tabular_widget_name = this.widgetName;
        args[0] = event;
      }
      objFuncArgs[1].apply(objFuncArgs[0] || window, args);
    },
    /**
     * Event handler for collapsing dropdown view upon (bubbling) blur.
     * @see Tabular.widget.base#viewEffectsApply
     * @private
     * @param {Event} event
     *   Native event, not jQuery event.
     */
    dropdownViewCollapse: function(event) {
      // _this_ is the view wrapper.
      var elm = this;
      // Only if currently visible.
      if (!elm.parentNode.hasAttribute('tabular-view-dropdown-collapsed')) {
        // Do it later, to allow toggling click event on dropdown value display
        // text field to take effect.
        // But if this happens too earlier, it only means that user can't hide
        // view by clicking the dropdown display; user can still hide by
        // clicking on the document.
        setTimeout(function() {
          // view-container > this.
          elm.parentNode.setAttribute('tabular-view-dropdown-collapsed', '1');
        }, 100);
      }
    },
    /**
     * Simple on/off listing usable when attempting to reduce number of answers
     * to frequently triggered events.
     * @param {string} name
     * @param {boolean} [set]
     * @return {boolean|undefined}
     */
    blockExecution: function(name, set) {
      var self = this, o = self.blockExecutions;
      if (set === undefined) {
        return o.hasOwnProperty(name) ? o[name] : undefined;
      }
      o[name] = set;
      return set;
    }
  };

}(jQuery));
