/**
 * @file
 * Tabular (tabular) widget.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Abstract base class for the value part of solitary and target classes.
   * @constructor
   */
  Tabular.widget.value = function() {
    throw new Error('Abstract class ' + this.type_of + ' is not to be instantiated.');
  };
  Tabular.widget.value.prototype = {
    constructor: Tabular.widget.value,
    /**
     * Type flag; to check if current instance extends this class.
     */
    isValue: true,
    /**
     * Set up stuff related to the value element.
     * @throws {Error}
     * @returns {object}
     *   Self.
     */
    initValue: function() {
      var self = this, errorCode, msg, fieldElm, elm, data;
      if (self.state) {
        return self;
      }
      self.valueChangeSubstriptions = {};
      try {
        // Find the value element and set id on it.
        fieldElm = Tabular.element(self.selectors.field);
        elm = $('input[type="hidden"][tabular-entity-kind]', fieldElm).get(0);
        self.selectors.value = '#' + (elm.id = this.selectors.namespace + '--tabular-source-value');
        // Set group ID on value element too, for debugging.
        elm.setAttribute('tabular-group-id', self.groupId);

        // Normalize value stored in the value element,
        // in case it was set by something else than this module.
        if (self.anyValue()) {
          if ((data = self.getValue())) {
            data = self.normalizeData(data);
          }
          if (data && Object.keys(data).length) {
            elm.value = JSON.stringify(data);
          }
          else {
            elm.value = '';
          }
        }
        return self;
      }
      catch (er) {
        errorCode = Tabular.errorCode('frontend_init_widget');
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // Display error to user, unless other previous initialization error.
        if (!self.state) {
          // Flag initialization error.
          self.state = 1;
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
        // Re-throw the error but tell caller to abort; don't log (again) etc.
        er.abortSilently = true;
        throw er;
      }
    },
    /**
     * @see Tabular.widgetGroup#setValueChangeSubscription
     * @private
     * @param {string} name
     * @param {object|null|undefined} obj
     *   If falsy, the function will be apply'ed to window.
     * @param {function} func
     * @param {Array} [args]
     */
    setValueChangeSubscription: function(name, obj, func, args) {
      this.valueChangeSubstriptions[name] = [obj !== window ? obj : null, func, args];
    },
    /**
     * @see Tabular.widgetGroup#removeValueChangeSubscription
     * @private
     * @param {string} name
     * @returns {Array|undefined}
     *   Array: [object, function, args].
     */
    removeValueChangeSubscription: function(name) {
      var scrptns = this.valueChangeSubstriptions, objFuncArgs;
      if (scrptns.hasOwnProperty(name)) {
        objFuncArgs = scrptns[name];
        delete scrptns[name];
        return objFuncArgs;
      }
    },
    /**
     * @see Tabular.widgetGroup#anyValue
     * @private
     * @returns {boolean}
     */
    anyValue: function() {
      var self = this, v = Tabular.element(self.selectors.value).value;
      return v !== '' && v !== '{}';
    },
    /**
     * Set row into value element's JSON.
     * Triggers change event on the value element on success.
     * @see Tabular.widgetGroup#setValue
     * @private
     * @param {object} row
     * @returns {boolean}
     */
    setValue: function(row) {
      var self = this, errorCode, msg, err, cnf = self.conf, primaryKey = cnf.schema.primaryKey, primaryKeyValue,
        maxRows = cnf.mockCardinality,
        elm = Tabular.element(self.selectors.value), all, allow, grp, srcActive;
      try {
        if (elm && !elm.readonly && !elm.disabled) {
          if (!row.hasOwnProperty(primaryKey)) {
            throw new Error('Arg row contains no primary key column[' + primaryKey + ']').
              customCode(Tabular.errorCode('primary_key_nonexist'));
          }
          primaryKeyValue = row[primaryKey];
          switch (typeof primaryKeyValue) {
            case 'string':
              if (!primaryKeyValue) {
                err = true;
              }
              break;
            case 'number':
              if (!isFinite) {
                err = true;
              }
              break;
            default:
              err = true;
          }
          if (err) {
            throw new Error('Arg row primary key column[' + primaryKey + '] value[' + primaryKeyValue + '] type[' +
              inspect.typeOf(primaryKeyValue) + '] is not finite number or non-empty string.').
              customCode(Tabular.errorCode('use'));
          }

          all = elm.value;
          // No rows yet.
          if (all === '' || all === '{}') {
            all = {};
            allow = true;
          }
          else {
            all = JSON.parse(all);
            // Already set?
            if (all.hasOwnProperty('_' + primaryKeyValue)) {
              // Nothing to do, no change.
              // But do dispatch change 'none' event, for searchable dropdown.
              self.dispatchValueChangeEvent('none', row);
              return false;
            }
            // Only one row allowed: remove existing before setting.
            if (maxRows === 1) {
              // But don't dispatch remove or truncate event, would be overkill.
              all = {};
              allow = true;
            }
            // Append because not exceeding mock cardinality.
            else if (maxRows < 1 || Object.keys(all).length < maxRows) {
              allow = true;
            }
          }
          if (allow) {
            all['_' + primaryKeyValue] = row;
            elm.value = JSON.stringify(all);
            // Dispatch _internal_ low-level value change event.
            $(elm).trigger('change');
            // Dispatch value change event.
            self.dispatchValueChangeEvent('set', row);
            // Show value change message in status list of the widget where
            // the change was instigated, if that widget wishes so.
            self.showValueChange('set', row);
            return true;
          }
          // No change.
          self.dispatchValueChangeEvent('none', row);
          // Deny, and tell user in active source widget's status list.
          if ((grp = Tabular.getGroup(self.groupId)) && (srcActive = grp.getActiveSourceWidget())) {
            srcActive.statusList.add(
              Tabular.message(self.entityKind, !self.conf.displaySelected ? 'max_rows' : 'max_rows_removable', {
                '!max': maxRows
              }),
              'notice'
            );
          }
          return false;
        }
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('use');
        inspect.traceLog(
          er,
          {
            type: self.module,
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // No message and no state change if already unhealthy.
        if (!self.state || self.state === 4) {
          // Init > init error, or healthy > fatal runtime error.
          self.state = !self.state ? 1 : 2;
          // Display error message to user, if active and healty up to now.
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
      }
      return false;
    },
    /**
     * @see Tabular.widgetGroup#getPrimaryKeys
     * @private
     * @param {boolean} [asString]
     *   Primary keys may be numbers, but the HTML attribute data-primary-key
     *   is string.
     * @returns {Array|undefined}
     */
    getPrimaryKeys: function(asString) {
      var self = this, errorCode, msg, primaryKey,
        elm = Tabular.element(self.selectors.value), all, keys = [], key;
      try {
        if (elm) {
          all = elm.value;
          if (all === '' || all === '{}') {
            return undefined;
          }
          all = JSON.parse(all);
          if (!all || !Object.keys(all).length) {
            return undefined;
          }
        }
        primaryKey = self.conf.schema.primaryKey;
        for (key in all) {
          if (all.hasOwnProperty(key)) {
            !asString ? keys.push(all[key][primaryKey]) : keys.push('' + all[key][primaryKey]);
          }
        }
        return keys;
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
        inspect.traceLog(
          er,
          {
            type: self.module,
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // No message and no state change if already unhealthy.
        if (!self.state || self.state === 4) {
          // Init > init error, or healthy > fatal runtime error.
          self.state = !self.state ? 1 : 2;
          // Display error message to user, if active and healty up to now.
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
      }
    },
    /**
     * Get row or all rows from the value element's JSON.
     * @see Tabular.widgetGroup#getValue
     * @private
     * @param {string|number|undefined} [primaryKeyValue]
     *   Undefined (no arg at all): get all values.
     *   Must not be underscore-prefixed key in a data list; has to be
     *   the actual value of the primary key column.
     * @returns {object|undefined}
     */
    getValue: function(primaryKeyValue) {
      var self = this, errorCode, msg, err, elm = Tabular.element(self.selectors.value), all;
      try {
        if (!elm) {
          return undefined;
        }
        all = elm.value;
        if (all === '' || all === '{}') {
          return undefined;
        }
        all = JSON.parse(all);

        if (!all || !Object.keys(all).length) {
          return undefined;
        }
        if (!arguments.length) {
          return all;
        }
        switch (typeof primaryKeyValue) {
          case 'string':
            if (!primaryKeyValue) {
              err = true;
            }
            break;
          case 'number':
            if (!isFinite) {
              err = true;
            }
            break;
          default:
            err = true;
        }
        if (err) {
          throw new Error('Arg primaryKeyValue[' + primaryKeyValue + '] type[' + inspect.typeOf(primaryKeyValue) +
            '] is not finite number or non-empty string.').
            customCode(Tabular.errorCode('use'));
        }
        if (all.hasOwnProperty('_' + primaryKeyValue)) {
          return all['_' + primaryKeyValue];
        }
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
        inspect.traceLog(
          er,
          {
            type: self.module,
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // No message and no state change if already unhealthy.
        if (!self.state || self.state === 4) {
          // Init > init error, or healthy > fatal runtime error.
          self.state = !self.state ? 1 : 2;
          // Display error message to user, if active and healty up to now.
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
      }
    },
    /**
     * @see Tabular.widgetGroup#removeValue
     * @private
     * @param {string|number} primaryKeyValue
     *   Must not be underscore-prefixed key in a data list; has to be
     *   the actual value of the primary key column.
     * @returns {object|undefined}
     */
    removeValue: function(primaryKeyValue) {
      var self = this,
        elm = Tabular.element(self.selectors.value), all, row;
      if (elm) {
        row = self.getValue(primaryKeyValue);
        if (row) {
          elm = Tabular.element(self.selectors.value);
          all = JSON.parse(elm.value);
          delete all['_' + primaryKeyValue];
          elm.value = JSON.stringify(all);
          // Dispatch _internal_ low-level value change event.
          $(elm).trigger('change');
          // Dispatch value change event.
          self.dispatchValueChangeEvent('remove', row);
          // Show value change message in status list of the widget where
          // the change was instigated, if that widget wishes so.
          self.showValueChange('remove', row);

          return row;
        }
        // @todo: perhaps dispatch 'none' value change event.
      }
    },
    /**
     * @see Tabular.widgetGroup#truncateValue
     * @private
     * @returns {object|undefined}
     */
    truncateValue: function() {
      var self = this,
        elm = Tabular.element(self.selectors.value), all;
      if (elm) {
        all = self.getValue();
        if (all) {
          elm = Tabular.element(self.selectors.value);
          elm.value = '';
          // Dispatch _internal_ low-level value change event.
          $(elm).trigger('change');
          // Dispatch value change event.
          self.dispatchValueChangeEvent('truncate', all);

          // Show value change message in status list of the widget where
          // the change was instigated, if that widget wishes so.
          self.showValueChange(
            'truncate',
            self.conf.mockCardinality === 1 ? all[Object.keys(all)[0]] : {}
          );

          return all;
        }
        // @todo: perhaps dispatch 'none' value change event.
      }
    },
    /**
     * @see Tabular.widgetGroup#confirmRemoveValue
     * @see Tabular.widgetGroup#removeValue
     * @private
     *
     * @throws {RangeError}
     *
     * @param {string|number} primaryKeyValue
     *   Must not be underscore-prefixed key in a data list; has to be
     *   the actual value of the primary key column.
     */
    confirmRemoveValue: function(primaryKeyValue) {
      var self = this, msg = self.conf.removeConfirm, row, mtchs, le, i;
      row = self.getValue(primaryKeyValue);
      if (!row) {
        throw new RangeError('Row having primary key value[' + primaryKeyValue + '] type[' + inspect.typeOf(primaryKeyValue) +
          '] cannot be removed because it doesnt exists, or bad arg.').
          customCode(Tabular.errorCode('use'));
      }
      // Replace !column. But don't use view columns convert.
      if (msg.indexOf('!') !== -1) {
        msg = Tabular.replaceColumnPatterns(msg, row);
      }
      Tabular.confirmDialog(msg, self, self.removeValue, [primaryKeyValue]);
    },
    /**
     * @private
     * @see Tabular.widgetGroup#setValueChangeSubscription
     * @see Tabular.dispatchEvent
     * @param {string} type
     *   Values: set|remove|truncate|none.
     * @param {object} rowOrAll
     */
    dispatchValueChangeEvent: function(type, rowOrAll) {
      var self = this, scrptns = self.valueChangeSubstriptions, nm, evt;
      evt = {
        type: 'tabular_value_change_' + type,
        tabular_value: rowOrAll
      };
      if (self.state === 4) {
        for (nm in scrptns) {
          if (scrptns.hasOwnProperty(nm)) {
            self.dispatchEvent(evt, scrptns[nm]);
          }
        }
      }
    },
    /**
     * Sets message in status list of last-user-action widget, if that widget
     * desires that.
     * @see Tabular.widget.value#setValue
     * @see Tabular.widget.value#removeValue
     * @see Tabular.widget.value#truncateValue
     * @private
     * @param {string} type
     *   Values: set|remove|truncate.
     * @param {object} [valueRow]
     */
    showValueChange: function(type, valueRow) {
      var self = this;
      // Wait a sec for last-user-action event handlers to kick in.
      setTimeout(function() {
        var grp, wdgt, msg;
        try {
          if (self.isSolitary) {
            wdgt = self;
          }
          else if ((grp = Tabular.getGroup(self.groupId)) && !(wdgt = grp.getLastUserActionWidget())) {
            // last-user-action event apparantly not kicked in.
            return;
          }
          // showValueChange 0: no | 1: on remove | 2: on set | 3: both.
          if (type === 'set') {
            if (wdgt.conf.showValueChange < 2) {
              return;
            }
          }
          else if (!wdgt.conf.showValueChange || wdgt.conf.showValueChange === 2) {
            return;
          }
          msg = Tabular.message(
            self.entityKind,
            type === 'set' ? 'value_set' : (type === 'remove' || wdgt.conf.mockCardinality === 1 ? 'value_removed' : 'value_truncated')
          );
          if (valueRow && msg.indexOf('!') !== -1 && Object.keys(valueRow)) {
            msg = Tabular.replaceColumnPatterns(msg, valueRow);
          }
          wdgt.statusList.add(msg);
        }
        catch (ignore) {
          // Try-catch because setTimeout().
        }
      }, 50);
    },
    /**
     * List of prefab expressions for inline event handlers on rows and buttons.
     * Not to be confused with tokens:
     * _ENTITY_KIND_, _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
     * @type {object}
     */
    inlineEventHandlerExpressions: {
      // Deliberately not semi-colon ended because must be wrappable.
      _GET_: 'Drupal.Tabular.getGroup(\'_GROUP_ID_\').getValue(\'_PRIMARY_KEY_\')',
      _REMOVE_CONFIRMED_: 'Drupal.Tabular.getGroup(\'_GROUP_ID_\').confirmRemoveValue(\'_PRIMARY_KEY_\')',
      _REMOVE_: 'Drupal.Tabular.getGroup(\'_GROUP_ID_\').removeValue(\'_PRIMARY_KEY_\');',
      _INSPECT_: 'inspect(Drupal.Tabular.getGroup(\'_GROUP_ID_\').getValue(\'_PRIMARY_KEY_\'))'
    }
  };

}(jQuery));
