/**
 * @file
 * Tabular (tabular) widget.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Abstract base class for all source widget classes, including solitary.
   * @see Tabular.base
   * @see Tabular.widget.base
   * @constructor
   */
  Tabular.widget.source = function() {
    throw new Error('Abstract class ' + this.type_of + ' is not to be instantiated.');
  };
  Tabular.widget.source.prototype = {
    constructor: Tabular.widget.source,
    /**
     * Type flag; to check if current instance extends this class.
     * @type {boolean}
     */
    isSource: true,
    /**
     * State: ever populated.
     * @type {boolean}
     */
    everPopulated: false,
    /**
     * State: currently populated.
     * @type {boolean}
     */
    isPopulated: false,
    /**
     * State: currently retrieving options or populating view.
     * @type {number}
     */
    isPopulating: 0,
    /**
     * State: last population returned no results.
     * @type {boolean}
     */
    populationNone: false,
    /**
     * @throws {Error}
     * @returns {object}
     *   Self.
     */
    initSource: function() {
      var self = this, errorCode, msg, widget = this, slctrs = self.selectors, root, elm, $tab, tab,
        $srchCntnr, $srchFld, _populateSearchField, $srchBut, $dropdownDisplay,
        o, u, fNm;
      if (self.state) {
        return self;
      }
      self.optionChangeSubstriptions = {};
      try {
        // Resolve provider populateAjaxPathArgs and populateSearchChange.
        o = self.conf.provider;
        if (o.populateAjaxPathArgs && o.hasOwnProperty('populateAjaxPathArgs')) {
          // Clear ref.
          u = null;
          u = Tabular.evalFunction(fNm = o.populateAjaxPathArgs);
          if (!u) {
            errorCode = Tabular.errorCode('non_resolvable_js_function');
            throw new Error('provider.populateAjaxPathArgs[' + fNm + '] won\'t eval to function.');
          }
          // Mask Tabular.widget.source's prototypal ditto
          self.populateAjaxPathArgs = u;
          delete o.populateAjaxPathArgs;
        }
        root = Tabular.element(slctrs.root);
        root.setAttribute('tabular-provider-name', o.name);

        if (o.titulate && o.hasOwnProperty('titulate')) {
          // Clear ref.
          u = null;
          u = Tabular.evalFunction(fNm = o.titulate);
          if (!u) {
            errorCode = Tabular.errorCode('non_resolvable_js_function');
            throw new Error('provider.titulate[' + fNm + '] won\'t eval to function.');
          }
          // Mask Tabular.widget.source's prototypal ditto
          self.titulate = u;
          delete o.titulate;
        }

        if (o.searchable) {
          root.setAttribute('tabular-provider-searchable', '1');

          if (o.populateSearchMin && o.hasOwnProperty('populateSearchMin')) {
            // Mask Tabular.widget.source's prototypal ditto.
            self.populateSearchMin = o.populateSearchMin;
            delete o.populateSearchMin;
          }
          if (o.populateSearchChange && o.hasOwnProperty('populateSearchChange')) {
            // Clear ref.
            u = null;
            u = Tabular.evalFunction(fNm = o.populateSearchChange);
            if (!u) {
              errorCode = Tabular.errorCode('non_resolvable_js_function');
              throw new Error('provider.populateSearchChange[' + fNm + '] won\'t eval to function.');
            }
            // Mask Tabular.widget.source's prototypal ditto
            self.populateSearchChange = u;
            delete o.populateSearchChange;
          }
        }
        // Clear ref.
        o = null;

        // Find tab.
        $tab = $(slctrs.field).find('div.tabular-tab').first();
        tab = $tab.get(0);
        slctrs.tab = '#' + (tab.id = slctrs.namespace + '--tabular-tab');
        tab.setAttribute('tabular-group-id', self.groupId);
        tab.setAttribute('tabular-widget-name', self.widgetName);
        // And set group ID on parent .tabular-tab-container.
        tab.parentNode.setAttribute('tabular-group-id', self.groupId);

        // Set up populate_search (provider) behaviour.
        if (self.conf.provider.searchable) {
          // Get - and set ID on - populate_search container.
          $srchCntnr = $('div.tabular-populate-search-container', root);
          if ($srchCntnr.length) {
            elm = $srchCntnr.get(0);
            slctrs.populateSearchContainer = '#' + (elm.id = slctrs.namespace + '--tabular-populate-search-container');

            // Get - and set ID on - populate_search field.
            $srchFld = $('input[type="text"]', elm);
            if ($srchFld.length) {
              slctrs.populateSearchField = _populateSearchField =
                '#' + ($srchFld.get(0).id = slctrs.namespace + '--tabular-populate-search-field');

              // Focus on search field upon click on tab.
              $tab.click(function() {
                Judy.focus(_populateSearchField);
              });

              // Clear button inside text field.
              if (self.conf.textClearButton) {
                $srchFld.addClass('tabular-text-clear-button');
                $('<span class="tabular-text-clear-button"><span>&nbsp;</span></span>').insertAfter($srchFld).
                  click(function() {
                    // Emptying must trigger keyup on search field, even if
                    // 'button' mode. Otherwise listening to keyup will be
                    // in complete.
                    var $srchFld = $(self.selectors.populateSearchField), chng = $srchFld.val() !== '';
                    $srchFld.val('').trigger('keyup');
                    // Emptying the search field must also trigger plain change;
                    // for dropdown.clearTruncate to work.
                    if (chng) {
                      $srchFld.trigger('change');
                    }
                  });
              }

              // populate_search mode:button.
              if (elm.getAttribute('tabular-populate-search-mode') === 'button') {

                // Get - and set ID and click listener on - populate_search button.
                $srchBut = $('input[type="button"]', elm);
                if ($srchBut.length) {
                  slctrs.populateSearchButton = '#' + ($srchBut.get(0).id = slctrs.namespace + '--tabular-populate-search-button');

                  // populate_search button click handler.
                  $srchBut.click(function() {
                    var srchFld = Tabular.element(widget.selectors.populateSearchField), pathArgs;
                    if (srchFld) {
                      if (!widget.isPopulating) {
                        pathArgs = widget.populateSearchChange(srchFld);
                        if (pathArgs.length) {
                          widget.populateCommence(pathArgs);
                        }
                        else {
                          // No results view when no match (or empty srch fld).
                          widget.clearView();
                          self.isPopulated = 0;
                          widget.dispatchOptionChangeEvent('truncate');
                        }
                      }
                    }
                  });
                  // Make 'enter' key in search field click search button,
                  // not submit the form.
                  Judy.keydown($srchFld, 'enter', function(event) {
                    event.preventDefault();
                    event.stopPropagation();
                    $(widget.selectors.populateSearchButton).trigger('click');
                  });
                }
              }
              // populate_search mode:auto(complete).
              else {
                // populate_search field keyup handler.
                $srchFld.keyup(function() {
                  var srchFld = this, pathArgs;
                  if (!widget.isPopulating) {
                    pathArgs = widget.populateSearchChange(srchFld);
                    if (pathArgs.length) {
                      widget.populateCommence(pathArgs);
                    }
                    else {
                      // No results view when no match (or empty search field).
                      widget.clearView();
                      self.isPopulated = 0;
                      widget.dispatchOptionChangeEvent('truncate');
                    }
                  }
                });
                // 'enter' key in search field shan't submit the form.
                Judy.keydown($srchFld, 'enter', function(event) {
                  event.preventDefault();
                  event.stopPropagation();
                });
              }
            }
          }
        }
        // Not searchable.
        else if (self.conf.dropdown) {
          $dropdownDisplay = $('div.tabular-dropdown-display-container > div > input[type="text"]', root);
          if ($dropdownDisplay.length) {
            elm = $dropdownDisplay.get(0);
            slctrs.dropdownDisplay = '#' + (elm.id = slctrs.namespace + '--tabular-dropdown-display');
            // Display/hide view upon focus on the dropdown value display.
            /**
             * @see Tabular.widget.base#viewEffectsApply
             * @see Tabular.widget.base#dropdownViewCollapse
             */
            $dropdownDisplay.bind('focus', function(event) {
              var viewCntnr, attr = 'tabular-view-dropdown-collapsed', blckd = 'dropdown-display_toggle_view', viewWrpr, root;
              event.preventDefault();
              // Ignore focus/click if view isn't populated yet, or if previous
              // event is being processed.
              if (!self.isPopulated || self.blockExecution(blckd) === true) {
                // Blur the dropdown display textfield, because otherwise it
                // won't be focusable (like by click) later.
                root = $(self.selectors.root).get(0);
                if (root) {
                  try {
                    root.focus();
                  }
                  catch (ignore) {
                    // Try-catch because focus().
                  }
                }
                return;
              }
              self.blockExecution('dropdown-display_toggle_view', true);
              viewCntnr = Tabular.element(self.selectors.viewContainer);
              if (viewCntnr) {
                if (viewCntnr.hasAttribute(attr)) {
                  // Show view container.
                  // Put focus on the view (table).
                  // Do it later, to allow view container to be displayed.
                  setTimeout(function() {
                    viewCntnr.removeAttribute(attr);
                    viewWrpr = $(viewCntnr).children(self.viewTypeSelectors.wrapper).first().get(0);
                    if (viewWrpr) {
                      try {
                        viewWrpr.focus();
                        self.blockExecution(blckd, false);
                      }
                      catch (ignore) {
                        // Try-catch because focus().
                        self.blockExecution(blckd, false);
                      }
                    }
                  }, 50);
                }
                else {
                  // Set focus on something else than the view wrapper.
                  // The view wrapper will then hide because blurred.
                  setTimeout(function() {
                    if (!viewCntnr.hasAttribute(attr)) {
                      root = $(self.selectors.root).get(0);
                      if (root) {
                        try {
                          root.focus();
                          self.blockExecution(blckd, false);
                        }
                        catch (ignore) {
                          // Try-catch because focus().
                          self.blockExecution(blckd, false);
                        }
                      }
                    }
                  }, 50);
                }
              }
            });
          }
        }
        return self;
      }
      catch (er) {
        if (!errorCode) {
          errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_widget');
        }
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // Display error to user, unless other previous initialization error.
        if (!self.state) {
          // Initialization error.
          self.state = 1;
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
        // Re-throw the error but tell caller to abort; don't log (again) etc.
        er.abortSilently = true;
        throw er;
      }
    },
    /**
     * Init interaction between the widgets (and the group).
     * Like stuff that requires that the groups can be referred.
     * @see Tabular.widgetGroup#initInteraction
     */
    interactSource: function() {
      var self = this, errorCode, msg, group, grpIndex, value, keys;
      if (self.state) {
        return;
      }
      try {
        group = Tabular.widgetGroups[self.groupId];

        // Fix div.form-item (CSS inherited) expansion;
        // only first source widget with tab must expand upwards,
        // and only last source widget with tab must expand downwards,
        // otherwise bad spacing between tab container and controls.
        if (!self.isSolitary && $(self.selectors.tab + ' > label').html() !== '') {
          if ((grpIndex = group.nameToIndex[self.widgetName]) === 0) {
            $(self.selectors.field).attr('tabular-group-source-first', '1');
          }
          else if (grpIndex === group.nSources - 1) {
            $(self.selectors.field).attr('tabular-group-source-last', '1');
          }
          else {
            $(self.selectors.field).attr('tabular-group-source-middle', '1');
          }
        }

        // Subscribe to value changes, to mark options like values selected.
        group.setValueChangeSubscription(
          'group:' + self.groupId + '.widget:' + self.widgetName + '.updateOptionsSelected',
          self,
          self.updateOptionsSelected
        );

        // For show_value_change (message in status list) behaviour,
        // see the widget.value class.

        // Dropdown pass title and (searchable) collapse effects.
        if (self.conf.dropdown) {
          // First do sanity checks which cant be performed earlier because
          // grouped source mock_cardinality isn't evident before now.
          if (self.conf.mockCardinality !== 1) {
            if (!self.conf.provider.searchable) {
              errorCode = Tabular.errorCode('inapplicable_dropdown');
              throw new Error('Dropdown inapplicable when non-searchable provider and mock cardinality not 1');
            }
            else {
              // We don't pass selection's title_column to search field when
              // mock_cardinality != 1. Override setting, don't err.
              self.conf.dropdown.passTitle = false;
              self.conf.dropdown.clearTruncate = false;
            }
          }
          // Pass title_column to dropdown display or populate search field.
          // updateOptionsSelected() does the same, except initially.
          if (!self.conf.provider.searchable || self.conf.dropdown.passTitle) {
            if ((value = group.getValue())) {
              keys = Object.keys(value);
              if (keys.length === 1) {
                if (!self.conf.provider.searchable) {
                  $(self.selectors.dropdownDisplay).val(self.titulate(value[keys[0]]));
                }
                else {
                  $(self.selectors.populateSearchField).val(self.titulate(value[keys[0]]));
                }
              }
            }
          }

          if (self.conf.provider.searchable) {
            // Fade searchable dropdown view away on value change
            // (initiated by this source).
            group.setValueChangeSubscription(
              'group:' + self.groupId + '.widget:' + self.widgetName + '.dropdown-searchable-fade-away',
              null,
              function(evt) {
                if (evt.tabular_group_id === self.groupId &&
                  (evt.type === 'tabular_value_change_set' || evt.type === 'tabular_value_change_none')
                ) {
                  if (self.isSolitary) {
                    self.dropdownSearchableFadeAway();
                  }
                  else {
                    // Wait a sec for last-user-action event handlers to kick in.
                    setTimeout(function() {
                      var grp, wdgt;
                      try {
                        if ((grp = Tabular.getGroup(self.groupId)) &&
                          (wdgt = grp.getLastUserActionWidget()) && wdgt.widgetName === self.widgetName
                        ) {
                          self.dropdownSearchableFadeAway();
                        }
                      }
                      catch (ignore) {
                        // Try-catch because setTimeout().
                      }
                    }, 50);
                  }
                }
              }
            );
            // Truncate upon user clearing the search field.
            if (self.conf.dropdown.clearTruncate) {
              $(self.selectors.populateSearchField).change(function() {
                var elm = this, grp;
                if ((elm.value === '' || $.trim(elm.value) === '') &&
                  (grp = Tabular.getGroup(self.groupId))
                ) {
                  grp.truncateValue();
                }
              });
            }
          }
        }
        // Initialized, and sound and healthy.
        self.state = 4;
      }
      catch (er) {
        if (!errorCode) {
          errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_widget');
        }
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // Display error to user, unless other previous initialization error.
        if (!self.state) {
          // Initialization error.
          self.state = 1;
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
        // Re-throw the error but tell caller to abort; don't log (again) etc.
        er.abortSilently = true;
        throw er;
      }
    },
    /**
     * Show tab and tabular root element, and (perhaps) commence populating.
     */
    activate: function() {
      var self = this;
      // autoPopulate values:
      // 0: no.
      // 1: First time activated.
      // 2: On page load
      if (!self.everPopulated && !self.isPopulating && self.conf.autoPopulate) {
        // Never display progress spinner if any kind of dropdown.
        if (!self.conf.dropdown) {
          self.displayProgress();
        }
        self.populateCommence();
      }
      $(self.selectors.tab).attr('tabular-active', '1');
      $(self.selectors.root).attr('tabular-active', '1');
      self.active = true;
    },
    /**
     * Hide tab and tabular root element.
     */
    deActivate: function() {
      var self = this;
      $(self.selectors.root).removeAttr('tabular-active');
      $(self.selectors.tab).removeAttr('tabular-active');
      self.active = false;
    },
    /**
     * @see Tabular.widgetGroup#setOptionChangeSubscription
     * @private
     * @param {string} name
     * @param {object|null|undefined} obj
     *   If falsy, the function will be apply'ed to window.
     * @param {function} func
     * @param {Array} [args]
     */
    setOptionChangeSubscription: function(name, obj, func, args) {
      this.optionChangeSubstriptions[name] = [obj !== window ? obj : null, func, args];
    },
    /**
     * @see Tabular.widgetGroup#removeOptionChangeSubscription
     * @private
     * @param {string} name
     * @returns {Array|undefined}
     *   Array: [object, function, args].
     */
    removeOptionChangeSubscription: function(name) {
      var scrptns = this.optionChangeSubstriptions, objFuncArgs;
      if (scrptns.hasOwnProperty(name)) {
        objFuncArgs = scrptns[name];
        delete scrptns[name];
        return objFuncArgs;
      }
    },
    /**
     * May be masked by instance method specified by (PHP) implemention
     * of hook_tabular_entity_kind_provider_info():
     * PHP array[entity kind][provider].js_populate_ajax_path_args.
     *
     * @private
     * @param {Array} [pathArgs]
     *   Strings.
     * @returns {Array}
     */
    populateAjaxPathArgs: function(pathArgs) {
      return pathArgs || [];
    },
    /**
     * May be masked by instance attribute specified by (PHP) implemention
     * of hook_tabular_entity_kind_provider_info():
     * PHP array[entity kind][provider].js_populate_search_min.
     *
     * @type {number}
     *   Integer.
     */
    populateSearchMin: 3,
    /**
     * Any populateSearchReplace function _must_ use this
     * to sanitize search phrase.
     * @type {RegExp}
     */
    populateSearchIllegals: /[\/%\*\!\|\"\[\]\r\n`<>\\]/g,
    /**
     * Search change callback for search phrases that may contain space.
     *
     * May be masked by instance method specified by (PHP) implemention
     * of hook_tabular_entity_kind_provider_info():
     * PHP array[entity kind][provider].js_populate_search_change.
     *
     * @private
     * @param {HTMLElement} searchElm
     *   input[type="text"].
     * @returns {Array}
     *   Empty if search shan't be executed (like: phrase too short).
     */
    populateSearchChange: function(searchElm) {
      var minLength = this.populateSearchMin, v = searchElm.value, s = v, le, trailSpc = '';
      if ((le = s.length) >= minLength) {
        // If trailing space
        if (s.charAt(le - 1) === ' ') {
          trailSpc = ' ';
          // Trailing space shan't count as length.
          if (le - 1 < minLength) {
            return [];
          }
        }
        s = $.trim(s).replace(/[ ]{2,}/, ' ').
          // Any populateSearchChange function must sanitize.
          replace(this.populateSearchIllegals, '');
        if (s.length >= minLength) {
          // Pass back trimmed value.
          if (s + trailSpc !== v) {
            searchElm.value = s + trailSpc;
          }
          return [s];
        }
        // Pass back trimmed value.
        if (s + trailSpc !== v) {
          searchElm.value = s + trailSpc;
        }
      }
      return [];
    },
    /**
     * Commence populating widget
     * @private
     * @param {Array} [pathArgs]
     * @returns {boolean}
     *   False if already populating.
     */
    populateCommence: function(pathArgs) {
      var self = this;
      if (self.state === 4 && !self.isPopulating) {
        self.isPopulating = 1;
        // Populate search field active.
        if (self.conf.provider.searchable) {
          $(self.selectors.populateSearchField).addClass('tabular-populate-search-commenced');
        }
        // Non-searchable dropdown retrieving.
        else if (self.conf.dropdown) {
          $(self.selectors.dropdownDisplay).addClass('tabular-dropdown-display-retrieving');
        }
        self.retrieveOptions(pathArgs);
        return true;
      }
      return false;
    },
    /**
     * Procede populating widget.
     * @private
     * @param {object|*} ajaxResponse
     */
    populateProcede: function(ajaxResponse) {
      var self = this, errorCode, msg, dataSet, firstViewRow, $viewCntnr, respError, severity;
      self.isPopulating = 2;
      if (self.state !== 4 || !ajaxResponse || typeof ajaxResponse !== 'object') {
        self.isPopulating = 0;
        return;
      }
      if (self.conf.provider.successDataMessageResponse) {
        if (!ajaxResponse || typeof ajaxResponse !== 'object' || !ajaxResponse.hasOwnProperty('success')) {
          errorCode = Tabular.errorCode('ajax_response_malformatted');
          inspect.log(
            {
              'entity kind': self.entityKind,
              provider: self.conf.provider,
              response: ajaxResponse
            },
            {
              type: self.module,
              message: 'Provider promises success_data_message_response but response has no success bucket',
              code: errorCode,
              severity: 'error'
            }
          );
          // Display error to user, unless previous fatal runtime error.
          if (self.state === 4) {
            self.state = 2;
            msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
            if (!self.displayErrorMessage(msg)) {
              Tabular.alertDialog(msg, 'error');
            }
          }
          self.isPopulating = 0;
          return;
        }
        if (!ajaxResponse.success) {
          // If active and sound and healthy; otherwise ignore.
          if (self.state === 4) {
            // Tell user - via status list display - about request failure.
            msg = ajaxResponse.message;
            // No result(s) and falsy response success flag.
            // NB: status 204 or 404 would normally not count as an error,
            // but falsy success flag may be used as a means of triggering
            // display of a status message.
            if (ajaxResponse.status === 204 || ajaxResponse.status === 404) {
              // The response may suggest a severity by integer or string);
              // otherwise use 'warning'.
              severity = ajaxResponse.hasOwnProperty('severity') ? Tabular.severity(ajaxResponse.severity, true) : 'warning';
              // Fall back on standard no-result message, if response doesn't
              // provide a message.
              if (!msg) {
                msg = Tabular.message(self.entityKind, 'provider_no_results');
              }
            }
            // Warning if service timeout or provider search validation failure;
            // reported either by status or by error code.
            else if (ajaxResponse.status === 504 ||
              ((respError = ajaxResponse.error_code || ajaxResponse.code || ajaxResponse.errorCode) &&
              (respError = Tabular.errorName(respError)) &&
              (respError === 'provider_gateway_timeout' || respError === 'provider_search_validation_failure'))
            ) {
              severity = 'warning';
            }
            else {
              severity = 'error';
            }
            self.statusList.add(msg, severity);
          }
          self.isPopulating = 0;
          return;
        }
        dataSet = ajaxResponse.data;
      }
      else {
        dataSet = ajaxResponse;
      }

      try {
        // Normalise data set - secure primary keys.
        dataSet = self.normalizeData(dataSet);
        // Show data set.
        firstViewRow = {};
        $viewCntnr = self.clearView().append(self.renderView(dataSet, firstViewRow));
        self.viewEffectsApply(firstViewRow);
        $viewCntnr.showByHidden();
        // And now that we're done.
        self.isPopulating = 0;
        self.everPopulated = self.isPopulated = true;
        // @todo: optionally hide a (this) tab if no results(?).
        self.populationNone = !!Object.keys(dataSet).length;
        // Populate search field passive.
        if (self.conf.provider.searchable) {
          $(self.selectors.populateSearchField).removeClass('tabular-populate-search-commenced');
        }
        // Non-searchable dropdown ready.
        else if (self.conf.dropdown) {
          $(self.selectors.dropdownDisplay).removeClass('tabular-dropdown-display-retrieving');
        }
        self.dispatchOptionChangeEvent('set');
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
        inspect.traceLog(er, {
          type: self.module,
          message: er.message,
          code: errorCode,
          severity: 'error'
        });
        // If active and sound and healthy.
        if (self.state === 4) {
          // Fatal runtime error.
          self.state = 2;
          // Tell user.
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
      }
    },
    /**
     * Requests data from provider via AJAX.
     * @private
     * @param {Array|undefined} [pathArgs]
     */
    retrieveOptions: function(pathArgs) {
      var self = this, aPthArgs = self.populateAjaxPathArgs(pathArgs), rgxSlsh, le, i,
        // Fixed path args given by provider (empty string if none).
        sPthArgs = self.conf.provider.fixed_args;
      if (sPthArgs !== '') {
        sPthArgs = '/' + sPthArgs;
      }
      // Remove slashes from path args; would ruin request path.
      if ((le = aPthArgs.length)) {
        rgxSlsh = /\//g;
        for (i = 0; i < le; i++) {
          sPthArgs += '/' + aPthArgs[i].replace(rgxSlsh, '');
        }
      }
      $.ajax({
        url: Drupal.settings.basePath + self.conf.provider.route +
          // Form build ID or wildcard.
          '/' + (self.conf.provider.caching === 'any' ? '*' :
          $(self.selectors.field).closest('form').find('input[type="hidden"][name="form_build_id"]').first().val()) +
          sPthArgs,
        type: 'GET',
        dataType: 'json',
        cache: self.conf.provider.caching !== 'none',
        success: function(response, textStatus, jqXHR) {
          if (self.state === 4) {
            self.populateProcede(response);
          }
        },
        error: function(jqXHR, textStatus, errorThrown) {
          var errorCode;
          // If active and sound and healthy.
          if (self.state === 4) {
            errorCode = Tabular.errorCode('ajax_response_failure');
            inspect.log(
              {
                'textStatus': textStatus,
                'responseText': jqXHR.responseText
              },
              {
                type: self.module,
                message: 'AJAX response failure in widget.source.retrieveOptions(), error code(' + errorCode + ')',
                code: errorCode,
                severity: 'error'
              }
            )
          }
        }
      });
    },
    /**
     * Marks every option row equivalent to a value row selected.
     * A source widget may not - if grouped - contain all options
     * equivalent to the target widget's values.
     *
     * This is not called initially; instead widget.table.renderView() sets rows
     * selected when building them.
     * @private
     * @see Tabular.widgetGroup#setValueChangeSubscription
     * @param {object} [evt]
     */
    updateOptionsSelected: function(evt) {
      var self = this, keys, rows, le, i, primaryKey, group, drpdwnVal = '';
      if (self.state === 4 &&
        !self.isPopulating && self.isPopulated &&
        // Get the option rows this source widget exposes.
        // self.viewTypeSelectors is a viewType class list.
        (rows = $(self.selectors.viewContainer + ' ' + self.viewTypeSelectors.view).find(self.viewTypeSelectors.row).get()) &&
        (group = Tabular.getGroup(self.groupId))
      ) {
        if (evt.type !== 'tabular_value_change_truncate') {
          keys = group.getPrimaryKeys(true);
        }
        if (evt.type !== 'tabular_value_change_none') {
          le = rows.length;
          if (!keys || !keys.length) {
            for (i = 0; i < le; i++) {
              rows[i].removeAttribute('tabular-row-selected');
            }
          }
          else {
            for (i = 0; i < le; i++) {
              if ((primaryKey = rows[i].getAttribute('data-primary-key')) && $.inArray(primaryKey, keys) !== -1) {
                rows[i].setAttribute('tabular-row-selected', '1');
              }
              else {
                rows[i].removeAttribute('tabular-row-selected');
              }
            }
          }
        }
        // Dropdown pass title_column to dropdown display/provider search field.
        if (self.conf.dropdown && (!self.conf.provider.searchable || self.conf.dropdown.passTitle)) {
          if (keys && keys.length) {
            try {
              drpdwnVal = self.titulate(group.getValue(keys[0]));
            }
            catch (ignore) {
              drpdwnVal = '';
            }
          }
          if (!self.conf.provider.searchable) {
            if (self.conf.dropdown.noneOption || evt.type !== 'tabular_value_change_none') {
              $(self.selectors.dropdownDisplay).val(drpdwnVal);
            }
          }
          else {
            // Searchable dropdown must have the search field filled, if case
            // user changed that and clicked on the selected option afterwards.
            $(self.selectors.populateSearchField).val(drpdwnVal);
          }
        }
      }
    },
    /**
     * Title representation of a value row.
     *
     * May be masked by instance method specified by (PHP) implemention
     * of hook_tabular_entity_kind_provider_info():
     * PHP array[entity kind][provider].js_titulate.
     *
     * @private
     * @param {object} valueRow
     * @returns {string}
     */
    titulate: function(valueRow) {
      // This default implementation simply uses the title_column; converted
      // if that column's view column defineds a converter.
      var self = this, schm = self.conf.schema, col = schm.viewColumns[schm.titleColumn], v = valueRow[schm.titleColumn];
      if (col.convert) {
        v = (col.convert[0] || window)(col.convert[1]);
      }
      return v;
    },
    /**
     * @private
     * @see Tabular.widgetGroup#setOptionChangeSubscription
     * @see Tabular.dispatchEvent
     * @param {string} type
     *   Values: set|remove|truncate.
     */
    dispatchOptionChangeEvent: function(type) {
      var self = this, scrptns = self.optionChangeSubstriptions, nm, evt;
      evt = {
        type: 'tabular_option_change_' + type
      };
      if (self.state === 4) {
        for (nm in scrptns) {
          if (scrptns.hasOwnProperty(nm)) {
            self.dispatchEvent(evt, scrptns[nm]);
          }
        }
      }
    },
    /**
     * Fades provider-searchable view container content away,
     * and dispatches 'truncate' option change event.
     * @see Tabular.widget.source#interactSource
     * @private
     */
    dropdownSearchableFadeAway: function() {
      var self = this, srchFld;
      // This method is instead of:
      // $(self.selectors.viewContainer).empty();
      Tabular.fader.make(self.selectors.viewContainer + ' > div:first-child', {
        fast: true,
        remove: true
      });
      // Wait another sec before setting focus back on the
      // provider populate search field.
      // We don't pass the title_column to the search field here;
      // updateOptionsSelected() handles that.
      try {
        setTimeout(function() {
          if ((srchFld = Tabular.element(self.selectors.populateSearchField))) {
            srchFld.focus();
          }
        }, 50);
      }
      catch (ignore) {
        // Try-catch because setTimeout() and focus().
      }
      // Tell the world that this source tabular now is empty.
      self.dispatchOptionChangeEvent('truncate');
    },
    /**
     * List of prefab expressions for inline event handlers on rows and buttons.
     * Not to be confused with tokens:
     * _ENTITY_KIND_, _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
     * @see Tabular.widget.source.getOptionRowValue
     * @type {object}
     */
    inlineEventHandlerExpressions: {
      // Deliberately not semi-colon ended because must be wrappable.
      _GET_: 'Drupal.Tabular.widget.source.getOptionRowValue(this, \'_ENTITY_KIND_\')',
      _SET_: 'Drupal.Tabular.getGroup(\'_GROUP_ID_\').setValue(Drupal.Tabular.widget.source.getOptionRowValue(this, \'_ENTITY_KIND_\'))',
      _TRUNCATE_: 'Drupal.Tabular.getGroup(\'_GROUP_ID_\').truncateValue()',
      _INSPECT_: 'inspect(Drupal.Tabular.widget.source.getOptionRowValue(this, \'_ENTITY_KIND_\'))'
    }
  };
  // base + widget + source.
  Tabular.extend(Tabular.widget.source, 'Tabular.widget.source', Tabular.base,
    Tabular.widget.base);

  /**
   * Static helper for source inlineEventHandlerExpressions.
   * Makes the expressions usable on option row buttons.
   * The value of an option row is set on the row, not the row's buttons.
   * @see Tabular.widget.source#inlineEventHandlerExpressions
   * @param {HTMLElement} elm
   * @param {string} entityKind
   * @returns {object|undefined}
   */
  Tabular.widget.source.getOptionRowValue = function(elm, entityKind) {
    var s = elm.getAttribute('data-' + entityKind);
    if (!s) {
      // Arg elm is a button - or at least not the row element,
      // so we have to find the ancestor row element.
      s = $(elm).parents('[data-' + entityKind + ']').first().attr('data-' + entityKind);
    }
    if (s) {
      return JSON.parse(s);
    }
  };

}(jQuery));
