/**
 * @file
 * Tabular (tabular) widget.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Abstract base class for all target widget classes.
   * @see Tabular.base
   * @see Tabular.widget.base
   * @see Tabular.widget.value
   * @constructor
   */
  Tabular.widget.target = function() {
    throw new Error('Abstract class ' + this.type_of + ' is not to be instantiated.');
  };
  Tabular.widget.target.prototype = {
    constructor: Tabular.widget.target,
    /**
     * Type flag; to check if current instance extends this class.
     * @type {boolean}
     */
    isTarget: true,
    /**
     * @throws {Error}
     * @returns {object}
     *   Self.
     */
    initTarget: function() {
      var self = this, errorCode, msg;
      if (self.state) {
        return self;
      }
      try {
        // On page load: display selection saved previously.
        // Solitary has no displaySelected conf var.
        if (self.conf.displaySelected) {
          if (self.anyValue()) {
            // Show field label if configured to be displayed when value(s).
            if (self.conf.displayLabelWhenValue) {
              $(self.selectors.label).showByHidden();
            }
            self.displaySelection();
          }
          else if (self.conf.displayEmpty) {
            self.displaySelection(true);
          }
        }

        // On value change: display selection.
        $(self.selectors.value).change(function() {
          var anyValue = self.anyValue();
          if (self.conf.displaySelected) {
            self.displaySelection();
            if (self.conf.displayLabelWhenValue) {
              $(self.selectors.label)[anyValue ? 'showByHidden' : 'hideByHidden']();
            }
          }
        });
        return self;
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_widget');
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // Display error to user, unless other previous initialization error.
        if (!self.state) {
          // Initialization error.
          self.state = 1;
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
        // Re-throw the error but tell caller to abort; don't log (again) etc.
        er.abortSilently = true;
        throw er;
      }
    },
    /**
     * Init interaction between the widgets (and the group).
     * Like stuff that requires that the groups can be referred.
     * @see Tabular.widgetGroup#initInteraction
     */
    interactTarget: function() {
      var self = this;
      if (self.state) {
        return;
      }

      // For show_value_change (message in status list) behaviour,
      // see the widget.value class.

      // Initialized, and sound and healthy.
      self.state = 4;
    },
    /**
     * Should be called even when there's no value(s).
     */
    displaySelection: function(none) {
      var self = this, errorCode, msg, dataSet, firstViewRow = {}, $viewCntnr;
      try {
        // getValue() return undefined in no value.
        if (!none && (dataSet = self.getValue())) {
          // Show data set.
          $viewCntnr = self.clearView().append(self.renderView(dataSet, firstViewRow));
          self.viewEffectsApply(firstViewRow);
        }
        else {
          $viewCntnr = self.clearView();
          if (self.conf.displayEmpty) {
            $viewCntnr.append(self.renderEmpty());
          }
        }
        $viewCntnr.showByHidden();
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // No message and no state change if already unhealthy.
        if (self.state === 4) {
          // Active and healthy > fatal runtime error.
          self.state = 2;
          // Display error message to user, if active and healty up to now.
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
      }
    }
  };
  // base + widget + value + target.
  Tabular.extend(Tabular.widget.target, 'Tabular.widget.target', Tabular.base,
    Tabular.widget.base, Tabular.widget.value);

}(jQuery));
