/**
 * @file
 * Tabular (tabular) widget.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Abstract base class for all widget classes that use an HTML table as view.
   * @constructor
   */
  Tabular.widget.table = function() {
    throw new Error('Abstract class ' + this.type_of + ' is not to be instantiated.');
  };
  Tabular.widget.table.prototype = {
    constructor: Tabular.widget.table,
    viewType: 'table',
    /**
     * @throws {Error}
     * @returns {object}
     *   Self.
     */
    initView: function() {
      var self = this, errorCode, msg, cnf = self.conf, cntnr, viewTypeSelectors = self.viewTypeSelectors;
      if (self.state) {
        return self;
      }
      try {
        cntnr = Tabular.element(self.selectors.viewContainer);
        if (!cnf.viewLegend) {
          cntnr.setAttribute('tabular-view-legend-none', '1');
        }
        if (self.isSource || self.conf.displaySelected) {
          // Check that jQuery.dataTable is available if needed.
          if (cnf.viewTypeSpecs.dynamicTable && typeof $().dataTable !== 'function') {
            inspect.log(
              {
                groupId: self.groupId,
                widgetName: self.widgetName,
                entityKind: self.entityKind,
                fieldId: self.selectors.field,
                conf: self.conf
              },
              {
                type: self.module || 'tabular',
                message: 'Tabular widget set to use dynamic table, but jQuery.dataTable library is not available',
                code: Tabular.errorCode('widget_table_no_datatable'),
                severity: 'warning'
              }
            );
            // Turn dynamic table off, and log warning.
            cnf.viewTypeSpecs.dynamicTable = null;
          }
          // Make view container vertical scrollable.
          if (cnf.visibleRows > -1 && !cnf.viewTypeSpecs.dynamicTable) {
            Tabular.element(self.selectors.viewContainer).setAttribute('tabular-visible-rows', '' + cnf.visibleRows);
          }
        }

        return self;
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_widget');
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        // Display error to user, unless other previous initialization error.
        if (!self.state) {
          // Initialization error.
          self.state = 1;
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
        // Re-throw the error but tell caller to abort; don't log (again) etc.
        er.abortSilently = true;
        throw er;
      }
    },
    viewTypeSelectors: {
      wrapper: 'div.tabular-view-wrapper',
      view: 'table[tabular-widget-view="1"]',
      row: 'tbody > tr'
    },
    viewTypeHtml: {
      viewEmptyStart: '<div tabular-widget-view="1" tabular-widget-view-empty="1"',
      viewEmptyEnd: '</div>',
      // tabindex to be focusable.
      viewStart: '<div class="tabular-view-wrapper" tabindex="0"><table tabular-widget-view="1" tabindex="0"',
      viewEnd: '</table></div>',
      headStart: '<thead',
      headEnd: '</thead>',
      headRowStart: '<tr',
      headRowEnd: '</tr>',
      headColStart: '<th',
      headColEnd: '</th>',
      bodyStart: '<tbody',
      bodyEnd: '</tbody>',
      bodyRowStart: '<tr tabindex="0"',
      bodyRowEnd: '</tr>',
      bodyColStart: '<td',
      bodyColEnd: '</td>'
    },
    /**
     * @see Tabular.widget.base#viewRow
     * @see Tabular.widget.base#renderView
     * @param {object} dataSet
     * @param {object} firstViewRow
     *   Reference needed in callers scope.
     * @returns {string}
     *   Empty on error.
     */
    renderView: function(dataSet, firstViewRow) {
      var self = this, errorCode, msg, cnf = self.conf,
        primaryKey = cnf.schema.primaryKey, viewTypeSpecs = cnf.viewTypeSpecs,
        // jQuery.dataTable adds odd/even CSS class erroneously if already set.
        dynTab = !!viewTypeSpecs.dynamicTable,
        html = self.viewTypeHtml,
        key, nRows = 0, viewRow, metadata, legend = '', sRows = '',
        selectedKeys;
      try {
        if (self.isSource) {
          selectedKeys = Tabular.widgetGroups[self.groupId].getPrimaryKeys();
        }
        for (key in dataSet) {
          if (dataSet.hasOwnProperty(key)) {
            // NB: widget.base#viewRow() throws RangeError.
            // viewRow returns an object listing some row metadata + the actual
            // view row.
            metadata = self.viewRow(key, dataSet[key]);
            // Make viewRow refer viewRow().view only (dirty but efficient).
            viewRow = metadata.view;
            delete metadata.view;

            if (!nRows) {
              // Record first viewRow for viewEffectsApply().
              firstViewRow.view = viewRow;
              // Base legend on column generic properties of first row.
              if (self.conf.viewLegend || dynTab) {
                legend = html.headStart +
                  // jQuery.dataTable fails if no legend.
                  (self.conf.viewLegend ? '' : ' hidden="1" aria-hidden="1"') +
                  '>' + html.headRowStart + ' class="tabular-row">' +
                  self.renderLegend(viewRow) +
                  html.headRowEnd + html.headEnd;
              }
              // Dropdown (non-searchable) 'none' option.
              if (self.isSource && self.conf.dropdown && self.conf.dropdown.noneOption) {
                sRows += html.bodyRowStart +
                  ' tabular-option-none="1" onclick="' +
                  self.inlineEventHandlerExpressions._TRUNCATE_.replace('_GROUP_ID_', self.groupId) +
                  '" class="tabular-row' + (dynTab ? '' : (nRows % 2 ? ' even' : ' odd')) + '">' +
                  html.bodyColStart + ' class="tabular-column" colspan="' + viewRow.length +
                  '"><div>' + Tabular.message(self.entityKind, 'option_none') + '</div>' + html.bodyColEnd +
                  html.bodyRowEnd;
                ++nRows;
              }
            }
            // Set primary key and JSON-encoded row on visual row.
            sRows += html.bodyRowStart +
              // Flag selected if source view and row in values.
              (!selectedKeys || $.inArray(dataSet[key][primaryKey], selectedKeys) === -1 ? '' : ' tabular-row-selected') +
              // (On)click row handler (row_click widget setting).
              (!metadata.click ? '' : (' onclick="' + metadata.click + '"')) +
              // The row's primary key value.
              ' data-primary-key="' + dataSet[key][primaryKey] +
              // The data JSON encoded; single-quotes escaped because rendered
              // within single-quotes.
              '" data-' + this.entityKind + "='" + JSON.stringify(dataSet[key]).replace(/'/g, '&#39;') + "'" +
              ' class="tabular-row' + (dynTab ? '' : (nRows % 2 ? ' even' : ' odd')) + '">' +
              self.renderRow(viewRow) +
              html.bodyRowEnd;
            ++nRows;
          }
        }
        if (nRows) {
          return html.viewStart + ' tabular-entity-kind="' + self.entityKind + '">' +
            legend +
            html.bodyStart + '>' +
            sRows +
            html.bodyEnd + html.viewEnd;
        }
        return self.renderEmpty();
      }
      catch (er) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
        inspect.traceLog(
          er,
          {
            type: self.module || 'tabular',
            message: er.message,
            code: errorCode,
            // noRepeatCode.
            severity: 'error'
          }
        );
        // No message and no state change if already unhealthy.
        if (self.state === 4) {
          // Active and healthy > fatal runtime error.
          self.state = 2;
          // Display error message to user, if active and healty up to now.
          msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
          if (!self.displayErrorMessage(msg)) {
            Tabular.alertDialog(msg, 'error');
          }
        }
      }
      return '';
    },
    /**
     *
     * @see Tabular.widget.base#renderLegend
     * @param {Array} viewRow
     *   First view row of a data set.
     * @returns {string}
     */
    renderLegend: function(viewRow) {
      var self = this, html = self.viewTypeHtml,
        nCols = viewRow.length, iCol,
        sRow = '', lbl,
        dynTab = !!self.conf.viewTypeSpecs.dynamicTable;
      for (iCol = 0; iCol < nCols; iCol++) {
        sRow += html.headColStart + ' class="tabular-legend';
        if (viewRow[iCol].isButton) {
          sRow += ' tabular-button button-' + viewRow[iCol].name + '">&nbsp;';
        }
        else {
          lbl = self.conf.schema.viewColumns[viewRow[iCol].name].label;
          sRow += ' column-' + viewRow[iCol].name + '"><div class="column-' + viewRow[iCol].name + '"' +
            (!dynTab ? '' : (' title="' + Tabular.message(self.entityKind, 'sort_by_column', {'!label': lbl}) + '"')) +
            '>' + lbl + '</div>';
        }
        sRow += html.headColEnd;
      }
      return sRow;
    },
    /**
     *
     * @see Tabular.widget.base#renderRow
     * @param {Array} viewRow
     * @returns {string}
     */
    renderRow: function(viewRow) {
      var self = this, html = self.viewTypeHtml,
        nCols = viewRow.length, iCol, sRow = '';
      for (iCol = 0; iCol < nCols; iCol++) {
        sRow += html.bodyColStart;
        if (viewRow[iCol].isButton) {
          sRow += ' class="tabular-button button-' + viewRow[iCol].name + '">' +
            '<div class="button-' + viewRow[iCol].name + '"' +
            (!viewRow[iCol].hover_title ? '' : ' title="' + viewRow[iCol].hover_title + '"') +
            // Prevent button click event from bubbling up to row.
            // And set last user action widget.
            // We set these on the buttons container in case someone sets
            // other event handlers on the button (forgetting to stop bubbling).
            ' onclick="' + self.inlineEventStopPropagation + self.inlineEventSetLastActionWidget + '"' +
            '><input type="button"' +
            (!viewRow[iCol].click ? '' : ' onclick="' + viewRow[iCol].click + '"') +
            ' value="' + viewRow[iCol].label + '"' +
            '/></div>';
        }
        else {
          sRow += ' class="tabular-column column-' + viewRow[iCol].name + '"' +
            (!viewRow[iCol].hover_title ? '' : ' title="' + viewRow[iCol].hover_title + '"') +
            '><div class="column-' + viewRow[iCol].name + '">' +
            // Make columns sortable (like date) by rendering raw value
            // in span before the converted value.
            (viewRow[iCol].converted === undefined ? viewRow[iCol].raw :
              ('<span hidden aria-hidden="true">' + viewRow[iCol].raw + '</span>' + viewRow[iCol].converted)) +
            '</div>';
        }
        sRow += html.bodyColEnd;
      }
      return sRow;
    },
    /**
     * Apply special effects on the view when added to DOM.
     * @see Tabular.widget.base#viewRow
     * @param {object} firstViewRow
     */
    viewEffectsApply: function(firstViewRow) {
      var self = this, errorCode, msg, spcs = self.conf.viewTypeSpecs,
        $viewCntnr, viewWrpr, $view, viewColumns, viewRow = firstViewRow.view,
        dynTbl = spcs.dynamicTable, opts,
        sortBy, u, le, i, $resSrchCntnr, $resSrchFld;
      // Get out if no results.
      if (!viewRow) {
        return;
      }

      if (!dynTbl) {
        // Make non-searchable dropdown collapse when view loses focus.
        if (self.isSource && self.conf.dropdown && !self.conf.provider.searchable) {
          viewWrpr = $(self.selectors.viewContainer + ' > ' + self.viewTypeSelectors.wrapper).get(0);
          if (viewWrpr) {
            // Use event capture because Firefox doesn't support the (bubbling)
            // focusout event. jQuery doesn't support event capturing.
            viewWrpr.addEventListener('blur', self.dropdownViewCollapse, true);
          }
        }
      }
      else {
        try {
          $viewCntnr = $(self.selectors.viewContainer);
          // Remove the view container's immediate child, because otherwise
          // we can't style away borders and stuff.
          $view = $('> div.tabular-view-wrapper >' + self.viewTypeSelectors.view, $viewCntnr).
            unwrap();
          // Re-use or create jQuery.dataTable options.
          opts = self.dataTableOptions;
          if (!opts) {
            opts = {
              // We don't use jQuery.dataTable's event system; instead all
              // behaviours are attached via inline event listeners.
              // Inline events may seem oldschool, but they are far safer, simpler
              // and performance-wise lighter than jQuery events.
              //fnRowCallback: function (nRow, aData, iDisplayIndex, iDisplayIndexFull) {
              //},
              // Yes, do paginate.
              bPaginate: true,
              // Number of visible rows is a widget setting.
              iDisplayLength: !self.conf.visibleRows ? -1 : self.conf.visibleRows,
              // Disable user option to change number of rows displayed.
              bLengthChange: false,
              // Certainly don't use sScrollY (scrollable table), which detaches
              // legend row and renders it a separate table - and thus seriously
              // f.... up everything.
              //sScrollY: '250px',
              // Don't set column widths; ruins any attempt to control widths.
              bAutoWidth: false,
              // Labels.
              oLanguage: {
                // 'Showing _START_ to _END_ of _TOTAL_ results'.
                sInfo: Tabular.message(self.entityKind, 'displayed_range'),
                // ' - filtering from _MAX_ results'; better left empty.
                sInfoFiltered: '',
                // 'No match'.
                sInfoEmpty: Tabular.message(self.entityKind, 'range_no_match'),
                oPaginate: {
                  sPrevious: Tabular.message(self.entityKind, 'paginate_previous'),
                  sNext: Tabular.message(self.entityKind, 'paginate_next')
                }
              },
              // Search within retrieved data set.
              bFilter: false
            };
            // Search within retrieved result set.
            if (dynTbl.result_search) {
              opts.bFilter = true;
              opts.oLanguage.sSearch = (u = dynTbl.result_search_label) === '' ? '' : ('<span>' + u + '</span>');
              // 'No matching examples.'. Gets displayed in table body.
              opts.oLanguage.sZeroRecords = Tabular.message(self.entityKind, 'result_search_no_match');
            }
            // All columns are - out of the dataTable box - sortable by user;
            // if we desired otherwise, aoColumns is the option to use.
            // Sort by (initially) is something we must handle; option aaSorting.

            // Sort initially.
            // Iterate a viewRow() list to get column indexes right even if some
            // columns are buttons.
            viewColumns = self.conf.schema.viewColumns;
            sortBy = [];
            le = viewRow.length;
            for (i = 0; i < le; i++) {
              if (!viewRow[i].isButton && (u = viewColumns[viewRow[i].name].sort)) {
                sortBy.push([i, u]);
              }
            }
            if (sortBy.length) {
              opts.aaSorting = sortBy;
            }
          }
          self.dataTableOptions = opts;
          // Apply jQuery.dataTable.
          if (typeof $view.dataTable !== 'function') {
            // Highly unlikely, because initView() checks too, but nevertheless.
            errorCode = Tabular.errorCode('widget_table_no_datatable');
            throw new Error('Missing jQuery.dataTable library');
          }
          $view.dataTable(opts);

          // Button columns shan't be sortable; remove all event listeners et al.
          $view.find('th.tabular-button').each(function() {
            $(this).unbind().removeAttr('role').removeAttr('aria-controls').removeAttr('aria-label');
          });

          // Make the result search field look like a Drupal field.
          if (dynTbl.result_search) {
            $resSrchCntnr = $viewCntnr.find('div.dataTables_filter').first().addClass('form-item form-type-textfield');
            // The input text field is placed inside the field label (sic!).
            // Move it out of the label - if not for any other reason then
            // to avoid problems in Chromium with position of the clear-button
            // inside text field.
            $resSrchFld = $resSrchCntnr.find('input[type="text"]').first().addClass('form-text').
              detach();
            $resSrchCntnr.append($resSrchFld);

            // Set placeholder attribute.
            if (dynTbl.result_search_placeholder !== '') {
              $resSrchFld.attr('placeholder', dynTbl.result_search_placeholder);
            }

            // Clear-button inside text field.
            if (self.conf.textClearButton) {
              $('<span class="tabular-text-clear-button"><span>&nbsp;</span></span>').insertAfter($resSrchFld).click(function() {
                // jQuery.dataTable seems to listen to keyup for this.
                $(self.selectors.viewContainer).find('div.dataTables_filter input[type="text"]').first().val('').
                  trigger('keyup');
              });
              $resSrchFld.addClass('tabular-text-clear-button');
            }
          }
        }
        catch (er) {
          if (!errorCode) {
            errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('unknown');
          }
          inspect.traceLog(
            er,
            {
              type: self.module || 'tabular',
              message: er.message,
              code: errorCode,
              severity: 'error'
            }
          );
          // No message and no state change if already unhealthy.
          if (self.state === 4) {
            // Active and healthy > fatal runtime error.
            self.state = 2;
            // Display error message to user, if active and healty up to now.
            msg = Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode});
            if (!self.displayErrorMessage(msg)) {
              Tabular.alertDialog(msg, 'error');
            }
          }
        }
      }
    }
  };

  /**
   * @constructor
   * @see Tabular.widget.solitary
   * @see Tabular.widget.table
   */
  Tabular.widget.solitaryTable = function() {
  };
  Tabular.widget.solitaryTable.prototype = {
    constructor: Tabular.widget.solitaryTable
  };
  // base + widget + value + source + solitary + table
  Tabular.extend(Tabular.widget.solitaryTable, 'Tabular.widget.solitaryTable',
    Tabular.widget.solitary, Tabular.widget.table);

  /**
   * @constructor
   * @see Tabular.widget.source
   * @see Tabular.widget.table
   */
  Tabular.widget.sourceTable = function() {
  };
  Tabular.widget.sourceTable.prototype = {
    constructor: Tabular.widget.sourceTable
  };
  // base + widget + source + table
  Tabular.extend(Tabular.widget.sourceTable, 'Tabular.widget.sourceTable',
    Tabular.widget.source, Tabular.widget.table);

  /**
   * @constructor
   * @see Tabular.widget.target
   * @see Tabular.widget.table
   */
  Tabular.widget.targetTable = function() {
  };
  Tabular.widget.targetTable.prototype = {
    constructor: Tabular.widget.targetTable
  };
  // base + widget + value + target + table
  Tabular.extend(Tabular.widget.targetTable, 'Tabular.widget.targetTable',
    Tabular.widget.target, Tabular.widget.table);

}(jQuery));
