/**
 * @file
 * Tabular (tabular) field.
 */

/*jslint browser: true, continue: true, indent: 2, newcap: true, nomen: true, plusplus: true, regexp: true, white: true, ass: true*/
/*global alert: false, confirm: false, console: false*/
/*global jQuery: false, Drupal: false, Judy: false*/

(function($) {
  'use strict';

  // Import.
  var Tabular = Drupal.Tabular;

  /**
   * Group controller.
   *
   * A Tabular group instance is a group of widgets, or a solitary widget.
   *
   * @constructor
   * @param {string} groupId
   * @param {Array} sourceElms
   *   List of tabular_group_solitary or tabular_group_source
   *   elements.
   * @param {HTMLElement} [targetElm]
   *   tabular_group_target element; if grouped.
   */
  Tabular.widgetGroup = function(groupId, sourceElms, targetElm) {
    var self = this,
      errorCode, widgets = [], nameToIndex = {}, nSources = sourceElms.length, i, widgetClass, elm, $tab, $tabContainer, u;

    try {
      // General.
      // Instantiate .converter to .convert unless already done.
      if (!Tabular.convert) {
        Tabular.convert = new Tabular.converter().init();
      }
      // Instantiate widgets.
      // Solitary.
      if (!targetElm) {
        widgetClass = 'solitary' + Tabular.toCamelCase(sourceElms[0].getAttribute('tabular-view-type'), true);
        u = new Tabular.widget[widgetClass]().
          initBase(groupId, sourceElms[0]).
          initValue().
          initSource().
          initSolitary().
          initView();
        nameToIndex[u.widgetName] = 0;
        widgets.push(u);
      }
      // Group.
      else {
        // Sources.
        for (i = 0; i < nSources; i++) {
          /**
           * @see Tabular.widget.sourceTable
           */
          widgetClass = 'source' + Tabular.toCamelCase(sourceElms[i].getAttribute('tabular-view-type'), true);
          // Clear ref.
          u = null;
          u = new Tabular.widget[widgetClass]().
            initBase(groupId, sourceElms[i]).
            initSource().
            initView();
          // Make sure widget names are unique across the group.
          if (nameToIndex.hasOwnProperty(u.widgetName)) {
            errorCode = Tabular.errorCode('non_unique_widget_name');
            throw new Error('Widget name[' + u.widgetName + '] is not unique across the group');
          }
          nameToIndex[u.widgetName] = i;
          widgets.push(u);
        }
        // Target.
        /**
         * @see Tabular.widget.targetTable
         */
        widgetClass = 'target' + Tabular.toCamelCase(targetElm.getAttribute('tabular-view-type'), true);
        u = new Tabular.widget[widgetClass]().
          initBase(groupId, targetElm).
          initValue().
          initTarget().
          initView();
        nameToIndex[u.widgetName] = i;
        widgets.push(u);
      }

      // Fix tabs.
      if (targetElm) {
        for (i = 0; i < nSources; i++) {
          elm = Tabular.element(widgets[i].selectors.tab);
          $tab = $(elm);
          if (!i) {
            // Find first tab's container.
            $tabContainer = $(elm.parentNode);
          }
          else if ($tabContainer) {
            // Move later widgets' tabs into first widget's tab container.
            $tabContainer.append(elm);
          }
          $tab.click(function() {
            try {
              Tabular.widgetGroups[this.parentNode.getAttribute('tabular-group-id')].
                activateWidget(this.getAttribute('tabular-widget-name'));
            }
            catch (er) {
              errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0);
              inspect.traceLog(
                er,
                {
                  type: self.module,
                  message: er.message,
                  code: errorCode,
                  severity: 'error'
                }
              );
            }
          });
        }
        // Display the (first) tab container. If the tabs are hidden then
        // the container will be too.
        $tabContainer.showByHidden();
      }
      // Solitary.
      else {
        $(Tabular.element(widgets[0].selectors.tab).parentNode).showByHidden();
      }
    }
    catch (er) {
      if (er.abortSilently && er.hasOwnProperty('abortSilently')) {
        return;
      }
      if (!errorCode) {
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_group');
      }
      inspect.traceLog(
        er,
        {
          type: self.module,
          message: er.message,
          code: errorCode,
          severity: 'error'
        }
      );
      Tabular.alertDialog(
        Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode}),
        'error'
      );
      return;
    }

    // Public properties.
    this.groupId = groupId;
    // Module defining entity kind.
    this.module = widgets[0].module;
    this.entityKind = widgets[0].entityKind;
    this.isSolitary = !targetElm;
    this.nSources = nSources;
    this.widgets = widgets;
    // Refer group target - even when solitary.
    this.groupTarget = widgets[widgets.length - 1];
    this.nameToIndex = nameToIndex;
    this.mockCardinality = this.groupTarget.conf.mockCardinality;
  };
  Tabular.widgetGroup.prototype = {
    constructor: Tabular.widgetGroup,
    /**
     * Init interaction between the widgets (and the group).
     * Like stuff that requires that this groups can be referred by the widgets.
     */
    initInteraction: function() {
      var self = this, errorCode, widgets = self.widgets, le, i;
      try {
        // Check that no initialization error has occurred.
        le = widgets.length;
        for (i = 0; i < le; i++) {
          if (self.widgets[i].state) {
            // Get out.
            return;
          }
        }
        // Call source and target interaction init methods.
        le = self.nSources;
        for (i = 0; i < le; i++) {
          if (!self.isSolitary) {
            self.widgets[i].conf.mockCardinality = self.groupTarget.conf.mockCardinality;
          }
          self.widgets[i].interactSource();
        }
        if (!self.isSolitary) {
          self.groupTarget.interactTarget();
        }
        // Activate first source widget.
        widgets[0].activate();
        // Populate secondary sources immediately,
        // if auto_populate setting says so.
        if (!self.isSolitary) {
          le = self.nSources;
          for (i = 1; i < le; i++) {
            if (widgets[i].conf.autoPopulate > 1) {
              widgets[i].populateCommence();
            }
          }
        }
        setTimeout(function() {
          try {
            // Add initialization done element for Drupal behaviors to discover.
            $(widgets[widgets.length - 1].selectors.field).append(
              '<tabular_group_ready tabular-group-id="' + self.groupId + '"' +
              ' tabular-entity-kind="' + self.entityKind + '" hidden aria-hidden="true"></tabular_group_ready>'
            );
            // Trigger behaviors listening to us.
            Drupal.attachBehaviors(
              $('tabular_group_ready[tabular-group-id="' + self.groupId + '"]').get(0),
              Drupal.settings
            );
          }
          catch (ignore) {
            // Try catch because setTimeout(). If this fails something else has
            // failed fatally and then this is irrelevant.
          }
        }, 50);
      }
      catch (er) {
        if (er.abortSilently && er.hasOwnProperty('abortSilently')) {
          return;
        }
        errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_group');
        inspect.traceLog(
          er,
          {
            type: self.module,
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        Tabular.alertDialog(
          Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode}),
          'error'
        );
      }
    },
    /**
     * @param {string} [widgetName]
     *   Optional, and ignored, if solitary widget.
     * @returns {Tabular.widget.base|undefined}
     */
    getWidget: function(widgetName) {
      var self = this, widgets = self.widgets, le, i;
      if (self.isSolitary) {
        return widgets[0];
      }
      le = widgets.length;
      for (i = 0; i < le; i++) {
        if (widgets[i].widgetName === widgetName) {
          return widgets[i];
        }
      }
    },
    /**
     * Get the widget the user last interacted with.
     * @return {Tabular.widget.base|undefined}
     */
    getLastUserActionWidget: function() {
      var lst = this.lastUserActionWidget;
      if (lst) {
        return this.getWidget(lst);
      }
    },
    /**
     * Get currently active source widget, if any active.
     */
    getActiveSourceWidget: function() {
      var self = this, widgets = self.widgets, le, i;
      if (self.isSolitary) {
        return widgets[0];
      }
      le = self.nSources;
      for (i = 0; i < le; i++) {
        if (widgets[i].active) {
          return widgets[i];
        }
      }
    },
    /**
     * Set options change event handler.
     * Triggers when a source widget rebuilds or clears it's options view.
     * If no arg args or args[0] is undefined or null, the function will be
     * called with a synthetic event object as first arg.
     *
     *  Buckets in synthetic event object:
     *  - (str) type: tabular_option_change_set
     *    or tabular_option_change_truncate
     *  - (str) tabular_group_id: this group's ID.
     *  - (str) tabular_widget_name: the widget origin of the event.
     *
     * Beware that these event type may trigger very frequently for a searchable
     * source widget.
     *
     * @see Tabular.widget.source#setOptionChangeSubscription
     * @param {string} name
     * @param {object|null|undefined} obj
     *   If falsy, the function will be apply'ed to window.
     * @param {function} func
     * @param {Array} [args]
     */
    setOptionChangeSubscription: function(name, obj, func, args) {
      var self = this, le = self.nSources, i;
      for (i = 0; i < le; i++) {
        self.widgets[i].setOptionChangeSubscription(name, obj, func, args);
      }
    },
    /**
     * Remove options change event handler.
     * @see Tabular.widget.source#removeOptionChangeSubscription
     * @param {string} name
     * @returns {Array|undefined}
     *   Array: [function] or [function, object].
     */
    removeOptionChangeSubscription: function(name) {
      var self = this, le = self.nSources, i;
      for (i = 0; i < le; i++) {
        self.widgets[i].removeOptionChangeSubscription(name);
      }
    },
    /**
     * Set value change event handler.
     * If no arg args or args[0] is undefined or null, the function will be
     * called with a synthetic event object as first arg.
     *
     *  Buckets in synthetic event object:
     *  - (str) type: tabular_value_change_none,
     *    tabular_value_change_set,
     *    tabular_value_change_remove,
     *    or tabular_value_change_truncate
     *  - (obj) tabular_value
     *  - (str) tabular_group_id: this group's ID.
     *  - (str) tabular_widget_name: the widget origin of the event.
     *
     * NB: Do not set event handlers directly on HTML value elements (such are
     * reserved for the widget group self).
     *
     * @see Tabular.widget.value#setValueChangeSubscription
     * @example
     * var subscriberFunc = function(evt, arg0, arg1) {
     *   inspect([evt, arg0, arg1], 'subscriberFunc');
     * };
     * Drupal.Tabular.getGroup('1O5s3AJcsk1bOSKr').setValueChangeSubscription('some:subscription', null, subscriberFunc, [null, 'what', 'ever']);
     * @param {string} name
     * @param {object|null|undefined} obj
     *   If falsy, the function will be apply'ed to window.
     * @param {function} func
     * @param {Array} [args]
     */
    setValueChangeSubscription: function(name, obj, func, args) {
      var self = this;
      (self.isSolitary ? self.widgets[0] : self.groupTarget).setValueChangeSubscription(name, obj, func, args);
    },
    /**
     * Remove value change event handler.
     * @see Tabular.widget.value#removeValueChangeSubscription
     * @param {string} name
     * @returns {Array|undefined}
     *   Array: [function] or [function, object].
     */
    removeValueChangeSubscription: function(name) {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).removeValueChangeSubscription(name);
    },
    /**
     * @see Tabular.widget.value#anyValue
     * @returns {boolean}
     *   Return value of widget.value's anyValue() method.
     */
    anyValue: function() {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).anyValue();
    },
    /**
     * Set a row on the widget group.
     * Triggers change event on the group's value element on success.
     * @see Tabular.widget.value#setValue
     * @param {object} row
     * @returns {boolean}
     *   Return value of widget.value's setValue() method.
     */
    setValue: function(row) {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).setValue(row);
    },
    /**
     * Get primary keys of all selected rows.
     * @see Tabular.widget.value#getPrimaryKeys
     * @param {boolean} [asString]
     *   Primary keys may be numbers, but the HTML attribute data-primary-key
     *   is string.
     * @returns {Array|undefined}
     */
    getPrimaryKeys: function(asString) {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).getPrimaryKeys(asString);
    },
    /**
     * Get a row, or all rows, set on the widget group.
     * @see Tabular.widget.value#getValue
     * @param {string|undefined} [primaryKeyValue]
     *   Undefined (no arg at all): get all values.
     *   Must not be underscore-prefixed key in a data list; has to be
     *   the actual value of the primary key column.
     * @returns {object|undefined}
     *   Return value of widget.value's getValue() method.
     */
    getValue: function(primaryKeyValue) {
      var self = this;
      if (arguments.length) {
        return (self.isSolitary ? self.widgets[0] : self.groupTarget).getValue(primaryKeyValue);
      }
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).getValue();
    },
    /**
     * Remove a row from the widget group.
     * Triggers no change event if that row wasn't set.
     * @see Tabular.widget.value#removeValue
     * @param {string|number} primaryKeyValue
     *   Must not be underscore-prefixed key in a data list; has to be
     *   the actual value of the primary key column.
     * @returns {object|undefined}
     */
    removeValue: function(primaryKeyValue) {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).removeValue(primaryKeyValue);
    },
    /**
     * Remove all rows from the widget group.
     * Triggers no change event if there were no rows.
     * @see Tabular.widget.value#truncateValue
     * @returns {object|undefined}
     */
    truncateValue: function() {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).truncateValue();
    },
    /**
     * Remove a row from the widget group, if user confirms.
     * Returns nothing, because removal may be executed in a context where
     * returning value isn't possible.
     * @see Tabular.widget.value#confirmRemoveValue
     * @param {string|number} primaryKeyValue
     *   Must not be underscore-prefixed key in a data list; has to be
     *   the actual value of the primary key column.
     */
    confirmRemoveValue: function(primaryKeyValue) {
      var self = this;
      return (self.isSolitary ? self.widgets[0] : self.groupTarget).confirmRemoveValue(primaryKeyValue);
    },
    /**
     * Activate source widget.
     * @param {string|number} widget
     *   String: widget name.
     *   Integer: source widget index.
     */
    activateWidget: function(widget) {
      var self = this, errorCode, index = widget, i;
      try {
        if (typeof widget === 'string') {
          if (!self.nameToIndex.hasOwnProperty(widget)) {
            errorCode = Tabular.errorCode('widget_name_nonexist');
            throw new Error('Source widget by name[' + widget + '] doesnt exist.');
          }
          index = self.nameToIndex[widget];
        }
        if (!index && index !== 0) {
          errorCode = Tabular.errorCode('widget_index_nonexist');
          throw new Error('Source widget by non-integer index type[' + inspect.typeOF(index) + '] doesnt exist.');
        }
        if (index < 0 || index > self.nSources - 1) {
          errorCode = Tabular.errorCode('widget_index_nonexist');
          throw new Error('Source widget by index[' + index + '] doesnt exist.');
        }
        // Hide all tabular root elements and deactivate all tabs.
        if (!self.isSolitary) {
          for (i = 0; i < self.nSources; i++) {
            if (i !== index) {
              self.widgets[i].deActivate();
            }
          }
        }
        self.widgets[index].activate();
        return true;
      }
      catch (er) {
        if (!errorCode) {
          errorCode = (typeof er.customCode === 'function' ? er.customCode() : 0) || Tabular.errorCode('frontend_init_group');
        }
        inspect.traceLog(
          er,
          {
            type: self.module,
            message: er.message,
            code: errorCode,
            severity: 'error'
          }
        );
        Tabular.alertDialog(
          Tabular.message('generic', 'fatal_frontend_error', {'@errorcode': errorCode}),
          'error'
        );
      }
      return false;
    },
    /**
     * Set which widget the user last interacted with.
     * For internal use only, don't mess with it.
     * @param {string} widgetName
     */
    setLastUserActionWidget: function(widgetName) {
      this.previousUserActionWidget = this.lastUserActionWidget;
      this.lastUserActionWidget = widgetName;
    }
  };
  Tabular.extend(Tabular.widgetGroup, 'Tabular.widgetGroup', Tabular.base);

}(jQuery));
