<?php
/**
 * @file
 * Tabular (tabular) field hook implementations which modules
 * declaring similar field type (not simply implementing a tabular entity kind)
 * must implement (customized) too.
 */

/**
 * Declares field types.
 *
 * Implements hook_field_info().
 *
 * @return array
 */
function tabular_field_info() {
  return array(
    'tabular' => array(
      'label' => t('Tabular'),
      'description' => t('Renders a list of selectable rows with configurable columns.'),
      // Default field settings.
      'settings' => array(
        // Custom settings.
        'entity_kind' => '',
      ),
      // Default instance settings.
      'instance_settings' => array(
        // Custom settings.
        'primary_key' => '',
        // Not called 'columns' to prevent Form API collision/confusion.
        'view_columns' => array(),
        'buttons' => array(),
        'title_column' => '',
        'mock_cardinality' => 1,
        'provider' => array(
          'name' => '',
          'caching' => 'form',
          'fixed_args' => '',
          'searchable' => FALSE,
        ),
        'attach_js' => array(
          'module' => '',
          'file' => '',
        ),
      ),
      // Field in form.
      'default_widget' => 'tabular_source_table',
      // Field as value.
      'default_formatter' => 'tabular_view_table',
      // Entity API integration, required for Rules and more
      'property_type' => 'text',
    )
  );
}

/**
 * Edit settings that apply to all instances of the field.
 *
 * Implements hook_field_settings_form().
 *
 * @param array $field
 * @param array $instance
 * @param boolean $has_data
 *
 * @return array
 */
function tabular_field_settings_form($field, $instance, $has_data) {
  $fld_set = $field['settings'];
  $form = array();
  switch ($field['type']) {
    case 'tabular':
      $entity_kinds = tabular_entity_kind();
      $entity_kind_default = !empty($fld_set['entity_kind']) ? $fld_set['entity_kind'] : '';
      $entity_kind_options = array();
      if (!$entity_kind_default || !isset($entity_kinds[$entity_kind_default])) {
        $entity_kind_options[''] = t('- Select -');
        if ($entity_kind_default) {
          drupal_set_message(
            t('The entity kind \'@entity_kind\' doesn\'t exist in this installation (anymore?).<br>Please enable a module that declares that entity kind, or select another kind.',
              array('@entity_kind' => $entity_kind_default)),
            'warning'
          );
        }
      }
      foreach ($entity_kinds as $name => $info) {
        $entity_kind_options[$name] = t(
          '!name - !class (!module)',
          array('!name' => $name, '!module' => $info['module'], '!class' => $info['class'])
        );
      }
      $form['entity_kind'] = array(
        '#type' => 'select',
        '#title' => t('Tabular entity kind'),
        '#description' => t('
An entity kind is an entity type with a loose schema.
<br>Unlike a Drupal API entity type it\'s schema is not tied to a database table schema - or at least it doesn\'t have to be.
<br>Instead, the schema is defined by it\'s defining PHP class\' schema() method.
<br>And that schema is only used for configuring which columns should be displayed (and how) in tabular widgets.'),
        '#options' => $entity_kind_options,
        '#default_value' => $entity_kind_default,
        '#required' => TRUE,
        // Might actually be changed, even if it has data, because the instance
        // settings form checks for new and obsolete instance view_columns.
        // But let's just stick to the general rule.
        '#disabled' => $has_data,
        // Firefix caches previous selections too aggressively;
        // even ignores selected="selected".
        '#attributes' => array(
          'autocomplete' => 'off',
        ),
      );
      break;
  }

  return $form;
}

/**
 * Edit settings of a particular instance of a field.
 *
 * Implements hook_field_instance_settings_form().
 *
 * @param array $field
 * @param array $instance
 *
 * @return array
 */
function tabular_field_instance_settings_form($field, $instance) {
  // Module declaring similar field type (not simply implementing
  // a tabular entity kind) should set this property.
  $field_type = 'tabular';

  $fld_set = $field['settings'];

  $entity_kind = $fld_set['entity_kind'];
  $entity_kind_specs = tabular_entity_kind($entity_kind);
  //$module = $entity_kind_specs['module'];
  $class = $entity_kind_specs['class'];
  // Include entity kind defining class (and ancestors).
  foreach ($entity_kind_specs['includes'] as $cls => $include) {
    module_load_include($include['ext'], $include['module'], $include['path_file']);
  }
  unset($entity_kind_specs, $cls, $include);

  $inst_set = $instance['settings'];
  $form = array();
  $module_path = drupal_get_path('module', 'tabular');
  drupal_add_js(
    $module_path . '/admin/tabular.field_ui.js',
    array('type' => 'file', 'group' => JS_DEFAULT, 'preprocess' => FALSE, 'every_page' => FALSE)
  );
  drupal_add_css(
    $module_path . '/admin/tabular.field_ui.css',
    array('type' => 'file', 'group' => CSS_DEFAULT, 'preprocess' => FALSE, 'every_page' => FALSE)
  );
  $weight = 100;
  switch ($field['type']) {
    case $field_type:
      $schema = call_user_func($class . '::schema');
      
      $form['translation_help'] = array(
        '#type' => 'markup',
        '#markup' => '<p>'
          . t('
<p><b>Translation of custom form elements</b>
<br>Form API, Features - and Localize Fields - don\'t support translation of custom elements like tabular column labels.
<br>Instead, all translation <code>t()</code> expressions get filed to the private:// dir.
<wbr>Please copy those <code>t()</code>s to a file in a (Features) module.
</p>
')
          . '</p>',
        '#weight' => ++$weight,
      );

      $widget_role = $instance['widget']['settings']['widget_role'];
      $widget_solitary = !empty($instance['widget']['settings']['widget_solitary']);

      // Primary key. ----------------------------------------------------------
      $primary_key = $schema['primary_key'];
      $form['primary_key'] = array(
        '#type' => 'hidden',
        '#default_value' => $primary_key,
      );
      $form['display_primary_key'] = array(
        '#type' => 'markup',
        '#markup' => '<div class="form-item"><label>' . t('Primary key') . '</label>' . $primary_key  . '</div>',
        '#weight' => ++$weight,
      );

      // Columns. --------------------------------------------------------------
      $columns_default =& $schema['view_columns'];
      $columns = $inst_set['view_columns'];
      // Remove columns that have disappeared from the entity kind class.
      $columns = array_intersect_key($columns, $columns_default);
      // Add columns defined in the entity kind class since last instance save.
      $new_columns = array_diff_key($columns_default, $columns);
      if ($new_columns) {
        $columns += $new_columns;
      }
      unset($new_columns);
      // Sort.
      // The primary key column is not necessarily to be displayed.
      $columns_sorted = !isset($columns[$primary_key]) ? array() : array(
        $primary_key => $columns[$primary_key],
      );
      unset($columns[$primary_key]);
      $columns_hidden = array();
      foreach ($columns as $name => $props) {
        if (!empty($props['weight'])) {
          $columns_sorted[$name] = $props;
        }
        else {
          $columns_hidden[$name] = $props;
        }
      }
      uasort($columns_sorted, 'drupal_sort_weight');
      $columns = $columns_sorted + $columns_hidden;
      $form['view_columns'] = array(
        '#type' => 'fieldset',
        '#title' => $widget_role == 'source' ? t('Columns in option rows') :
          t('Columns in value rows'),
        '#collapsible' => TRUE,
        '#collapsed' => TRUE,
        'help' => array(
          '#type' => 'markup',
          '#markup' => t('
<p>View columns are defined by !class::schema(). They are a subset of the columns available in data rows.
<br><b>Grouped fields</b> do not have to display the same columns, and the configuration of the columns may also differ.
</p>
          ',
          array('!class' => $class)),
        ),
        'title_column' => array(
          '#type' => 'markup',
          '#markup' => '<b>' . t('Title column: !title_column', array('!title_column' => $schema['title_column'])) . '</b>',
        ),
        '#weight' => ++$weight,
      );
      $form['title_column'] = array(
        '#type' => 'hidden',
        '#value' => $schema['title_column'],
        '#weight' => ++$weight,
      );
      $n_columns = 0;
      $descriptions_displayed = FALSE;
      $weight_options = drupal_map_assoc(range(-10, -1))
        + array(0 => t('Exclude'))
        + drupal_map_assoc(range(1, 10));
      foreach ($columns as $name => $props) {
        $form['view_columns'][$name] = array(
          '#type' => 'fieldset',
          '#title' => $name,
          '#collapsible' => FALSE,
          '#attributes' => array(
            'class' => array('tabular-column'),
          ),
          'weight' => array(
            '#type' => 'select',
            '#title' => t('Weight in row'),
            '#options' => $weight_options,
            '#default_value' => isset($props['weight']) ? $props['weight'] : 0,
          ),
          'label' => array(
            '#type' => 'textfield',
            '#title' => t('Label'),
            '#default_value' => isset($props['label']) ? $props['label'] : '',
            '#maxlength' => 255,
            '#size' => 32,
            '#states' => array(
              'invisible' => array(
                ':input[name="instance[settings][view_columns][' . $name . '][weight]"]' => array('value' => '0'),
              ),
              'optional' => array(
                ':input[name="instance[settings][view_columns][' . $name . '][weight]"]' => array('value' => '0'),
              ),
            ),
            '#attributes' => array(
              'class' => array('tabular-column-label'),
              // Trim value.
              'onchange' => array(
                'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
              ),
            ),
          ),
          'hover_title' => array(
            '#type' => 'textfield',
            '#title' => t('Hover HTML title'),
            '#default_value' => isset($props['hover_title']) ? $props['hover_title'] : '',
            '#maxlength' => 255,
            '#size' => 32,
            '#states' => array(
              'invisible' => array(
                ':input[name="instance[settings][view_columns][' . $name . '][weight]"]' => array('value' => '0'),
              ),
            ),
            '#attributes' => array(
              'class' => array('tabular-column-hover-title'),
              // Trim value.
              'onchange' => array(
                'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
              ),
            ),
          ),
          'sort' => array(
            '#type' => 'select',
            '#title' => t('Sort by in view'),
            '#options' => array(
              '' => t('No'),
              'asc' => t('Ascending'),
              'desc' => t('Descending'),
            ),
            '#default_value' => !empty($props['sort']) ? $props['sort'] : '',
            '#states' => array(
              'invisible' => array(
                ':input[name="instance[settings][view_columns][' . $name . '][weight]"]' => array('value' => '0'),
              ),
            ),
          ),
          'convert' => array(
            '#type' => 'textfield',
            '#title' => t('Javascript converter'),
            '#default_value' => !empty($props['convert']) ? $props['convert'] : '',
            '#maxlength' => 255,
            '#size' => 64,
            '#states' => array(
              'invisible' => array(
                ':input[name="instance[settings][view_columns][' . $name . '][weight]"]' => array('value' => '0'),
              ),
            ),
            '#attributes' => array(
              // Trim value.
              'onchange' => array(
                'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
              ),
            ),
          ),
        );
        if ((++$n_columns) == 1 || !$descriptions_displayed) {
          if (!empty($form['view_columns'][$name]['weight']['#default_value'])) {
            $descriptions_displayed = TRUE;
          }
          $form['view_columns'][$name]['hover_title']['#description'] = t(
            'Token _VALUE_ for column converted/raw value, _RAW_<br>for raw value. Otherwise: t() column name replacers<br>supported, like \'Whatever !!primary_key\'. Or empty.',
            array('!primary_key' => $primary_key)
          );
          $form['view_columns'][$name]['convert']['#description'] =
            t('Name of a globally accessible JS function.<br>Gets called with the whole row as second argument: <code>colValue = window.convertFunction(colValue, row);</code><br>- because that enables the convert method to alter the value based on the values of other columns.');
        }
      }

      // Buttons. --------------------------------------------------------------
      $buttons_default =& $schema['buttons'];
      $buttons = !empty($inst_set['buttons']) ? $inst_set['buttons'] : array();
      // Remove buttons that have disappeared from the entity kind class.
      $buttons = array_intersect_key($buttons, $buttons_default);
      // Add buttons defined in the entity kind class since last instance save.
      $new_buttons = array_diff_key($buttons_default, $buttons);
      if ($new_buttons) {
        $buttons += $new_buttons;
      }
      unset($new_buttons);
      if ($buttons) {
        foreach ($buttons as $name => &$props) {
          if (empty($props['weight'])) {
            $props['weight'] = 0;
          }
        }
        unset($props);
        uasort($buttons, 'drupal_sort_weight');
      }
      $form['buttons'] = array(
        '#type' => 'fieldset',
        '#title' => $widget_role == 'source' ? t('Buttons in option rows') :
          t('Buttons in value rows'),
        '#collapsible' => TRUE,
        '#collapsed' => TRUE,
        '#weight' => ++$weight,
      );
      if (!$buttons) {
        $form['buttons']['no_buttons'] = array(
          '#type' => 'markup',
          '#markup' => t('No buttons defined.'),
        );
      }
      else {
        $n_buttons = 0;
        $descriptions_displayed = FALSE;
        $weight_options = drupal_map_assoc(range(-10, -1))
          + array(0 => t('Exclude'))
          + drupal_map_assoc(range(1, 10));
        foreach ($buttons as $name => $props) {
          $form['buttons'][$name] = array(
            '#type' => 'fieldset',
            '#title' => $name,
            '#collapsible' => FALSE,
            '#attributes' => array(
              'class' => array('tabular-button'),
            ),
            'weight' => array(
              '#type' => 'select',
              '#title' => t('Weight in row'),
              '#description' => t('Negative is prepend,<br>positive append.'),
              '#options' => $weight_options,
              '#default_value' => isset($props['weight']) ? $props['weight'] : 0,
            ),
            'label' => array(
              '#type' => 'textfield',
              '#title' => t('Label'),
              '#default_value' => isset($props['label']) ? $props['label'] : '',
              '#maxlength' => 255,
              '#size' => 32,
              '#states' => array(
                'invisible' => array(
                  ':input[name="instance[settings][buttons][' . $name . '][weight]"]' => array('value' => '0'),
                ),
                'optional' => array(
                  ':input[name="instance[settings][buttons][' . $name . '][weight]"]' => array('value' => '0'),
                ),
              ),
              '#attributes' => array(
                'class' => array('tabular-button-label'),
                // Trim value.
                'onchange' => array(
                  'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
                ),
              ),
            ),
            'hover_title' => array(
              '#type' => 'textfield',
              '#title' => t('Hover HTML title'),
              '#default_value' => isset($props['hover_title']) ? $props['hover_title'] : '',
              '#maxlength' => 255,
              '#size' => 32,
              '#states' => array(
                'invisible' => array(
                  ':input[name="instance[settings][buttons][' . $name . '][weight]"]' => array('value' => '0'),
                ),
              ),
              '#attributes' => array(
                'class' => array('tabular-button-hover-title'),
                // Trim value.
                'onchange' => array(
                  'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
                ),
              ),
            ),
            'click' => array(
              '#type' => 'textfield',
              '#title' => t('Click event handler'),
              '#default_value' => !empty($props['click']) ? $props['click'] : '',
              '#maxlength' => 255,
              '#size' => 100,
              '#states' => array(
                'invisible' => array(
                  ':input[name="instance[settings][buttons][' . $name . '][weight]"]' => array('value' => '0'),
                ),
              ),
              '#attributes' => array(
                // Trim value.
                'onchange' => array(
                  'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
                ),
              ),
            ),
          );
          if ((++$n_buttons) == 1 || !$descriptions_displayed) {
            if (!empty($form['buttons'][$name]['weight']['#default_value'])) {
              $descriptions_displayed = TRUE;
            }
            $form['buttons'][$name]['hover_title']['#description'] =
              t('t() column name replacers supported,<br>like \'Remove !!primary_key\'', array('!primary_key' => $primary_key));
            $form['buttons'][$name]['click']['#description'] = $widget_role != 'target' ? t('
<b>Supported magic expressions</b>:
<br>_GET_: Get the row value as Javascript object.
<br>_SET_: Add the row to selection (is default).
<br>_INSPECT_: Dump the row value to browser console.
<br><b>Supported tokens</b>: _ENTITY_KIND_, _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
') : t('
<b>Supported magic expressions</b>:
<br>_GET_: Get the row value as Javascript object.
<br>_REMOVE_CONFIRMED_: Remove the row from selection, if user confirms.
<br>_REMOVE_: Remove the row from selection (is default).
<br>_INSPECT_: Dump the row value to browser console.
<br><b>Supported tokens</b>: _ENTITY_KIND_, _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
');
          }
        }
      }

      if ($widget_solitary || $widget_role == 'target') {
        $form['mock_cardinality'] = array(
          '#type' => 'select',
          '#title' => t('Mock cardinality (number of values)'),
          '#description' => t('
Non-empty field value is always a (JSON-encoded) object list, even when only one value (row).
<br>Standard Form API \'!fapi_card\' is not usable and must always (must) be 1.</b>
', array('!fapi_card' => t('Number of values'))),
          '#options' => array(FIELD_CARDINALITY_UNLIMITED => t('Unlimited')) + drupal_map_assoc(range(1, 100)),
          '#default_value' => isset($inst_set['mock_cardinality']) ? $inst_set['mock_cardinality'] : 1,
          '#weight' => ++$weight,
        );
      }
      // Mock cardinality is irrelevant for an in-group source, because we never
      // save value on a field with that widget.
      else {
        $form['mock_cardinality'] = array(
          '#type' => 'hidden',
          '#default_value' => '',
        );
      }

      // Hide Localize Fields' 'Add row' translation field.
      drupal_add_css('div.form-item.form-item-field-settings-add-row-localization-source {display:none}', array('type' => 'inline'));

      // Data provider. --------------------------------------------------------
      if ($widget_role == 'source') {
        $providers = tabular_entity_kind_provider($fld_set['entity_kind']);
        $default_provider = !empty($inst_set['provider']['name']) && isset($providers[$inst_set['provider']['name']])
          ? $inst_set['provider']['name'] : '';
        $form['provider'] = array(
          '#type' => 'fieldset',
          '#title' => t('Data provider'),
          '#collapsible' => TRUE,
          '#collapsed' => TRUE,
          '#weight' => ++$weight,
          'name' => array(
            '#type' => 'select',
            '#title' => t('Name'),
            '#default_value' => $default_provider,
            '#required' => TRUE,
            '#attributes' => array(
              // Set the 'searchable' hidden field on change of provider name.
              'onchange' => 'jQuery(\'input[type=\"hidden\"][name=\"instance[settings][provider][searchable]\"]\').val(jQuery(\'input[type=\"hidden\"][name=\"instance[settings][provider][specs__\' + jQuery(this).val() + \'][searchable]\"]\').val()).trigger(\'change\');',
            ),
          ),
          'searchable' => array(
            '#type' => 'hidden',
            '#default_value' => !$default_provider ? '0' : (!empty($providers[$default_provider]['searchable']) ? '1' : '0'),
          ),
        );
        $provider_options = array();
        foreach ($providers as $name => $info) {
          $provider_options[$name] = $name;
          $specs_legend = array('module:', 'route:', 'callback:', 'success_data_message_response:', 'searchable:',);
          $specs_values = array(
            $info['module'], $info['route'], $info['callback'],
            !empty($info['success_data_message_response']) ? 'true' : 'false',
            !empty($info['searchable']) ? 'true' : 'false',
          );
          $form['provider']['specs__' . $name] = array(
            '#type' => 'container',
            'searchable' => array(
              '#type' => 'hidden',
              '#default_value' => !empty($info['searchable']) ? '1' : '0',
            ),
            'content' => array(
              '#type' => 'markup',
              // Table-like ding.
              '#markup' => '<div>' . join('<br>', $specs_legend) . '</div><div>' . join('<br>', $specs_values) . '</div>',
            ),
            '#states' => array(
              'visible' => array(
                ':input[name="instance[settings][provider][name]"]' => array('value' => $name),
              ),
            ),
            '#attributes' => array(
              'class' => array('tabular-option', 'description'),
            ),
          );
          if (isset($info['fixed_args'])) {
            if (!$info['fixed_args']) {
              $form['provider']['fixed_args__' . $name] = array(
                '#type' => 'textfield',
                '#title' => t('Fixed path args'),
                '#description' => t('After form build ID or wildcard, and before search phrase (if any).<br>Slash-separated list of non-URL-encoded parameters - in essence a path.'),
                '#default_value' => !empty($inst_set['provider']['fixed_args']) ? $inst_set['provider']['fixed_args'] : '',
                '#states' => array(
                  'visible' => array(
                    ':input[name="instance[settings][provider][name]"]' => array('value' => $name),
                  ),
                ),
                '#attributes' => array(
                  'onchange' => 'jQuery(\'input[type=\"hidden\"][name=\"instance[settings][provider][fixed_args]\"]\').val(jQuery.trim(this.value));',
                ),
              );
            }
            else {
              $fixed_args_options = $info['fixed_args'];
              $fixed_args_default = !empty($inst_set['provider']['fixed_args']) ? $inst_set['provider']['fixed_args'] : '';
              if (!array_key_exists($fixed_args_default, $fixed_args_options)) {
                $fixed_args_default = '';
                $fixed_args_options = array('' => t('- Select -')) + $fixed_args_options;
              }
              $form['provider']['fixed_args__' . $name] = array(
                '#type' => 'select',
                '#title' => t('Fixed path args'),
                '#description' => t('After form build ID or wildcard, and before search phrase (if any).<br>Slash-separated list of non-URL-encoded parameters - in essence a path.'),
                '#options' => $fixed_args_options,
                '#default_value' => $fixed_args_default,
                '#states' => array(
                  'visible' => array(
                    ':input[name="instance[settings][provider][name]"]' => array('value' => $name),
                  ),
                  'required' => array(
                    ':input[name="instance[settings][provider][name]"]' => array('value' => $name),
                  ),
                ),
                '#attributes' => array(
                  'onchange' => 'jQuery(\'input[type=\"hidden\"][name=\"instance[settings][provider][fixed_args]\"]\').val(jQuery(this).val());',
                ),
              );
            }
          }
        }
        unset($specs_legend, $specs_values);
        $form['provider']['fixed_args'] = array(
          '#type' => 'hidden',
          '#default_value' => !empty($inst_set['provider']['fixed_args']) ? $inst_set['provider']['fixed_args'] : '',
        );
        $form['provider']['name']['#options'] = $provider_options;
        $form['provider']['caching'] = array(
          '#type' => 'radios',
          '#title' => t('Browser caching'),
          '#options' => array(
            'none' => t('None'),
            'form' => t('Current form'),
            'any' => t('Any form'),
          ),
          '#default_value' => !empty($inst_set['provider']['caching']) ? $inst_set['provider']['caching'] : 'form',
        );
        $form['provider']['get_caching_explain'] = array(
          '#type' => 'container',
          '#attributes' => array(
            'class' => array('tabular-option', 'description'),
          ),
          'none' => array(
            '#type' => 'container',
            'content' => array(
              '#type' => 'markup',
              '#markup' => t('First path arg is unique form build ID, and <code>jQuery.ajax({cache: false}}</code><br>The AJAX callback better use backend Drupal caching if data set retrieval is expensive.'),
            ),
            '#states' => array(
              'visible' => array(
                ':input[name="instance[settings][provider][caching]"]' => array('value' => 'none'),
              ),
            ),
          ),
          'form' => array(
            '#type' => 'container',
            'content' => array(
              '#type' => 'markup',
              '#markup' => t('First path arg is unique form build ID, and <code>jQuery.ajax({cache: true}}</code>'),
            ),
            '#states' => array(
              'visible' => array(
                ':input[name="instance[settings][provider][caching]"]' => array('value' => 'form'),
              ),
            ),
          ),
          'any' => array(
            '#type' => 'container',
            'content' => array(
              '#type' => 'markup',
              '#markup' => t('First path arg is wildcard *, and <code>jQuery.ajax({cache: true}}</code>'),
            ),
            '#states' => array(
              'visible' => array(
                ':input[name="instance[settings][provider][caching]"]' => array('value' => 'any'),
              ),
            ),
          ),
        );
      }
      $module_options = module_list();
      $module_options[''] = '- None -';
      ksort($module_options);
      $form['attach_js'] = array(
        '#type' => 'fieldset',
        '#title' => t('Attach custom Javascript behaviors script'),
        '#collapsible' => TRUE,
        '#collapsed' => TRUE,
        '#weight' => ++$weight,
        'attach_js_explain' => array(
          '#type' => 'markup',
          '#markup' => t('
<p><b>Subscribe to options or value(s) changes - or whatever you like</b>
<br>See example in the example module\'s js folder: <code>tabular_example.instance_customization.js</code>
<br>If the widgets are grouped then you only have to add the script to one of the fields/widgets of the group.
</p>'),
        ),
        'module' => array(
          '#type' => 'select',
          '#title' => t('Module'),
          '#options' => $module_options,
          '#default_value' => !empty($inst_set['attach_js']['module']) ? $inst_set['attach_js']['module'] : '',
        ),
        'file' => array(
          '#type' => 'textfield',
          '#title' => t('Path and filename relative to module path'),
          '#default_value' => !empty($inst_set['attach_js']['file']) ? $inst_set['attach_js']['file'] : '',
          '#attributes' => array(
            'placeholder' => 'js/my_module.entity_kind.instance_customization.js',
          )
        ),
      );

      return $form;
  }

  return $form;
}

/**
 * Validates for emptyness, and fixes that only the actual hidden value element
 * must count as an item.
 *
 * Implements hook_field_is_empty().
 *
 * @see field_default_validate()
 * @see _field_filter_items()
 *
 * @param array $item
 * @param array $field
 *
 * @return boolean
 */
function tabular_field_is_empty($item, $field) {
  // If array it's not the actual value element (but populate_search field or
  // other) and by returning true we avoid the default field_default_validate();
  // which would see the field as having more than one value (sic!).
  if (is_array($item)
    || empty($item) || $item == '{}'
  ) {
    return TRUE;
  }
  return FALSE;
}

/**
 * Checks required instance for emptyness;
 * Form API doesn't act on instance:required otherwise.
 *
 * Implements hook_field_validate().
 *
 * @param string $entity_type
 * @param string $entity
 * @param array $field
 * @param array $instance
 * @param string $langcode
 * @param array $items
 * @param array &$errors
 * @return void
 */
function tabular_field_validate($entity_type, $entity, $field, $instance, $langcode, $items, &$errors) {
  if ($instance['required'] && (empty($items['value']) || $items['value'] === '{}')) {
    // Despite no delta - because of widget info
    // behaviors.multiple values:FIELD_BEHAVIOR_CUSTOM -
    // the error must be put in a delta zero sub bucket.
    $errors[$field['field_name']][$langcode][0][] = array(
      'error' => 'required',
      'message' => t(
        'The field %field is required.',
        array('%field' => $instance['label'])
      ),
    );
  }
}

/**
 * Implements hook_field_widget_error().
 *
 * @param array $element
 * @param array $error
 * @param array $form
 * @param array &$form_state
 */
function tabular_field_widget_error($element, $error, $form, &$form_state) {
  form_error($element, $error['message']);
}

/**
 * Element value callback for the tabular field type.
 *
 * @see _form_builder_handle_input_element()
 * @see tabular_field_widget_form()
 *
 * @param array $element
 * @param string|boolean $input
 *   False: during form building.
 *   Array: upon form submission.
 * @param array $form_state
 *
 * @return string
 */
function tabular_tabular_value(&$element, $input, $form_state) {
  if ($input === FALSE) {
    return isset($element['#default_value']) ? $element['#default_value'] : '';
  }
  return $input;
}

/**
 * Fixes that a tabular field's value isn't placed correctly
 * right before saving to database.
 *
 * Presumably an effect of widget info
 * behaviours.multiple values:FIELD_BEHAVIOR_CUSTOM.
 *
 * @param $entity_type
 * @param $entity
 * @param $field
 * @param $instance
 * @param $langcode
 * @param array &$items
 */
function tabular_field_presave($entity_type, $entity, $field, $instance, $langcode, &$items) {
  if ($field['type'] == 'tabular') {
    // Move value from [delta] to [delta][value].
    if (isset($items[0])) {
      $items[0] = array(
        'value' => $items[0],
      );
    }
  }
}
