<?php
/**
 * @file
 * Tabular (tabular) field widget hook implementations
 * which modules declaring similar field type (not simply implementing
 * a tabular entity kind) needn't implement too.
 */

/**
 * Defines widgets and which field types of this module may use the widgets.
 *
 * Field types of other modules must implement hook_field_widget_info_alter()
 * to use these widgets.
 *
 * Implements hook_field_widget_info().
 *
 * @return array
 */
function tabular_field_widget_info() {
  $weight = 0;
  return array(
    // Solitary tabular with no display of selected row(s) is really just
    // a variant of the next in-group source tabular.
    'tabular_solitary_table' => array(
      'label' => t('Tabular solitary source with no display of selected row(s)'),
      'field types' => array(
        'tabular',
      ),
      'behaviors' => array(
        // One widget accepts multiple values (rows).
        // NB: like list - [und][value] not [und][0][value].
        'multiple values' => FIELD_BEHAVIOR_CUSTOM,
        // Skip default value GUI in backend field instance settings.
        'default value' => FIELD_BEHAVIOR_NONE,
      ),
      // Default widget settings.
      'settings' => array(
        // Standard settings.
        // - None. -

        // Custom settings.
        // Widget display type (hardcoded).
        'view_type' => 'table',
        // Cross widget type flags, which are needed by instance settings form.
        'widget_role' => 'source',
        'widget_solitary' => 1,

        // Solitary tabular gets new random group ID (by frontend) every time
        // the field gets rendered; rather discovered by fronted.
        // In case widget type is changed from a group type to solitary.
        'widget_group_id' => '',
        'widget_name' => '',

        'field_label' => 2,
        // Tab above tabular (source).
        'tab_label' => '',
        // Population. Solitary default: on page load.
        'auto_populate' => 2,
        // Searchable provider.
        'populate_search' => array(),
        // Row click event handler.
        'row_click' => '_SET_',

        // Show value change in status list.
        // 0: no | 1: on remove | 2: on set | 3: on both.
        'show_value_change' => 3,

        // Render legend in view; like head in table.
        'view_legend' => 1,

        'visible_rows' => 5,

        'text_clear_button' => 1,
      ),
      'weight' => (++$weight),
    ),
    'tabular_source_table' => array(
      'label' => t('Tabular in-group source table'),
      'field types' => array(
        'tabular',
      ),
      'behaviors' => array(
        // One widget accepts multiple values (rows).
        // NB: like list - [und][value] not [und][0][value].
        'multiple values' => FIELD_BEHAVIOR_CUSTOM,
        // Skip default value GUI in backend field instance settings.
        'default value' => FIELD_BEHAVIOR_NONE,
      ),
      // Default widget settings.
      'settings' => array(
        // Standard settings.
        // - None. -

        // Custom settings.
        // Widget display type (hardcoded).
        'view_type' => 'table',
        // Cross widget type flags, which are needed by instance settings form.
        'widget_role' => 'source',

        'widget_group_id' => '',
        'widget_name' => '',

        'field_label' => 2,
        // Tab above tabular (source).
        'tab_label' => '',
        // Population. Grouped source default: on activation.
        'auto_populate' => 1,
        // Searchable provider.
        'populate_search' => array(),
        // Row click event handler.
        'row_click' => '_SET_',

        // Show value change in status list.
        // 0: no | 1: on remove | 2: on set | 3: on both.
        'show_value_change' => 3,

        // Render legend in view; like head in table.
        'view_legend' => 1,

        'visible_rows' => 5,

        'text_clear_button' => 1,
      ),
      'weight' => (++$weight),
    ),
    'tabular_target_table' => array(
      'label' => t('Tabular in group target with optional display of selected row(s)'),
      'field types' => array(
        // Field types known (declared) by this module which use the widget.
        'tabular',
      ),
      'behaviors' => array(
        // One widget accepts multiple values (rows).
        // NB: like list - [und][value] not [und][0][value].
        'multiple values' => FIELD_BEHAVIOR_CUSTOM,
        // Skip default value GUI in backend field instance settings.
        'default value' => FIELD_BEHAVIOR_NONE,
      ),
      'settings' => array(
        // Standard settings.
        // - None. -

        // Custom settings.
        // Custom settings.
        // Widget display type (hardcoded).
        'view_type' => 'table',
        // Cross widget type flags, which are needed by instance settings form.
        'widget_role' => 'target',

        'discovery_mode' => NULL,

        'widget_group_id' => '',
        'widget_name' => '',
        // Number of source fields in the group.
        'widget_group_n_sources' => '',

        // Display selected row(s).
        'display_selected' => 0,
        // Display empty list when none selected.
        'display_empty' => 0,

        // Default to display when value(s).
        'field_label' => 1,

        // Row click event handler.
        'row_click' => '_REMOVE_',
        // Row removal confirm message.
        'remove_confirm' => '',

        // Show value change in status list.
        // 0: no | 1: on remove (2: on set | 3: on both - target can only remove
        'show_value_change' => 1,

        // Render legend in view; like head in table.
        'view_legend' => 1,

        'visible_rows' => -1,

        'text_clear_button' => 1,
      ),
      'weight' => (++$weight),
    ),
  );
}

/**
 * Implements hook_field_widget_settings_form().
 *
 * @param $field
 * @param $instance
 *
 * @return array
 */
function tabular_field_widget_settings_form($field, $instance) {
  $widg_set = $instance['widget']['settings'];

  $form = array(
    '#after_build' => array('tabular_field_widget_settings_form_after_build'),
    '#attributes' => array(
      'autocomplete' => 'off',
    ),
  );
  $module_path = drupal_get_path('module', 'tabular');
  drupal_add_js(
    $module_path . '/admin/tabular.field_ui.js',
    array('type' => 'file', 'group' => JS_DEFAULT, 'preprocess' => FALSE, 'every_page' => FALSE)
  );
  drupal_add_css(
    $module_path . '/admin/tabular.field_ui.css',
    array('type' => 'file', 'group' => CSS_DEFAULT, 'preprocess' => FALSE, 'every_page' => FALSE)
  );
  $weight = 200;

  $form['widget_settings_header'] = array(
    '#type' => 'markup',
    '#markup' => '<hr/><h4>' . t('Widget settings') . '</h4>',
    '#weight' => ++$weight,
  );

  $widget_type = $instance['widget']['type'];
  $widget_role = $widg_set['widget_role'];
  $form['widget_role'] = array(
    '#type' => 'hidden',
    '#default_value' => $widg_set['widget_role'],
    '#weight' => ++$weight,
  );
  $widget_solitary = !empty($widg_set['widget_solitary']);

  // Grouped fields - group ID.
  if ($widget_solitary) {
    $form['widget_solitary'] = array(
      '#type' => 'hidden',
      '#default_value' => 1,
      '#weight' => ++$weight,
    );
    // In case widget type is changed from a group type to solitary.
    $form['widget_group_id'] = array(
      '#type' => 'hidden',
      '#default_value' => '',
    );
  }
  else {
    $form['group_requirements'] = array(
      '#type' => 'markup',
      '#markup' => '<div class="description">'
        . t('All members of the group must have the same entity kind.<br>The group must consist of a single <em>target</em> and at least one <em>source</em>.')
        . '</div>',
      '#weight' => ++$weight,
    );
    // Group source.
    if ($widget_role == 'source') {
      if (!empty($widg_set['widget_group_id'])) {
        $widget_group_id = $widg_set['widget_group_id'];
      }
      else {
        // Create short random [a-zA-Z0-9] string.
        $c = explode('.', uniqid('', TRUE));
        $widget_group_id = Inspect::baseConvert($c[0], 16, 62) . Inspect::baseConvert($c[1], 16, 62);
      }
      $form['widget_group_id'] = array(
        '#type' => 'textfield',
        '#title' => t('Widget group ID'),
        '#description' => t('Obsolete if the <em>target</em> uses the \'Siblings\' discovery mode.'),
        '#default_value' => $widget_group_id,
        '#maxlength' => 32,
        '#size' => 32,
        '#attributes' => array(
          // Enforce name'ish value.
          'onkeyup' => array(
            'Drupal.tabular.field_ui.eventHandlers.convertLaxName(0, this)'
          ),
        ),
        '#weight' => ++$weight,
      );
    }
    // Group target.
    else {
      $form['discovery_mode'] = array(
        '#type' => 'radios',
        '#title' => t('Source widget discovery mode'),
        '#description' => t('How to figure out which source widget(s) belong to this target.<br><b>If \'Siblings\', this target <em>must</em> have an alphanumerically \'later\' field name than the sources.</b>'),
        '#options' => array(
          'group_id' => t('Common widget group ID'),
          'sibling' => t('Siblings - all source widgets within parent (field collection or form)'),
        ),
        '#default_value' => isset($widg_set['discovery_mode']) ? $widg_set['discovery_mode'] : NULL,
        '#required' => TRUE,
        '#weight' => ++$weight,
      );
      $form['identical_groups_warning'] = array(
        '#type' => 'fieldset',
        '#collapsible'=> TRUE,
        '#collapsed'=> TRUE,
        '#title' => t('Multiple instances warning'),
        'content' => array(
          '#type' => 'markup',
          '#markup' => t('
<h5>Grouped fields - N <em>source</em> tabulars + 1 <em>target</em></h5>
<p><b>Widget group IDs must be unique</b>
<br>The widget group ID is used to group one or more <em>source</em> widgets with a <em>target</em> widget.
<br>If two identical groups appear in the DOM, then the first group will cease to work immediately after the second has appeared.
<br>So that will only work if <em>the user isn\'t supposed to - and actively prevented from - interacting with the first group</em> after the latter appear.
</p>
<p><b>Groups in field collections</b>
<br>When fields of a group are placed in a field collection, then that field collection is an instance by itself.
<br>Things may break if that field collection instance appears twice in a node (or other root entity) - unless <em>the user isn\'t supposed to...</em>.
</p>
<p><b>How can a field collection instance appear twice?</b>
<br>The field collection allows multiple values.
<br>Or the field collection is child of another field collecition, and <em>that</em> allows multiple values.
<br>Or is added directly to the node plus added as child of another field collection.
</p>
<p><b>Allow multiple instances of groups</b>
<br>See Tabular admin page; set var <code>tabular_allow_group_overwrite</code> to truthy.
</p>
<hr>
<h5>Solitary widgets - 1 <em>source</em> tabular with built-in <em>target</em></h5>
<p><b>Widget ID uniqueness is not an issue</b>
<br>Because every instance gets it\'s own random unique ID.
<br>The same field (instance) can appear multiple times across node (root entity) and field collections, because the instances are not identical.
</p>
'),
        ),
        '#states' => array(
          'visible' => array(
            ':input[name="instance[widget][settings][discovery_mode]"]' => array('value' => 'group_id'),
          ),
        ),
        '#weight' => ++$weight,
      );
      if (!empty($widg_set['widget_group_id'])) {
        $widget_group_id = $widg_set['widget_group_id'];
      }
      else {
        // Create short random [a-zA-Z0-9] string.
        $c = explode('.', uniqid('', TRUE));
        $widget_group_id = Inspect::baseConvert($c[0], 16, 62) . Inspect::baseConvert($c[1], 16, 62);
      }
      $form['widget_group_id'] = array(
        '#type' => 'textfield',
        '#title' => t('Widget group ID - all <em>sources</em> and the <em>target</em> must have the same ID'),
        '#description' => t('Used to identify group members, and set interrelated behaviours.<br><b>The ID must be copied manually across all members of the group.</b>'),
        '#default_value' => $widget_group_id,
        '#maxlength' => 32,
        '#size' => 32,
        '#attributes' => array(
          // Enforce name'ish value.
          'onkeyup' => array(
            'Drupal.tabular.field_ui.eventHandlers.convertLaxName(0, this)'
          ),
        ),
        '#states' => array(
          'visible' => array(
            ':input[name="instance[widget][settings][discovery_mode]"]' => array('value' => 'group_id'),
          ),
          'required' => array(
            ':input[name="instance[widget][settings][discovery_mode]"]' => array('value' => 'group_id'),
          ),
        ),
        '#weight' => ++$weight,
      );
      // A target widget is always last in a group.
      $form['widget_group_n_sources'] = array(
        '#type' => 'select',
        '#title' => t('Number of in-group <em>source</em> fields'),
        '#description' => t('Needed by the Javascript layer to determine whether all in-group sources have been added to the DOM yet.'),
        // '_none' doesn't work here (as empty), but empty key apparanly does.
        '#options' => array('' => t('- Select -'))
          + drupal_map_assoc(range(1, 10)),
        '#default_value' => isset($widg_set['widget_group_n_sources']) ? $widg_set['widget_group_n_sources'] : '',
        '#states' => array(
          'visible' => array(
            ':input[name="instance[widget][settings][discovery_mode]"]' => array('value' => 'group_id'),
          ),
          'required' => array(
            ':input[name="instance[widget][settings][discovery_mode]"]' => array('value' => 'group_id'),
          ),
        ),
        '#weight' => ++$weight,
      );
    }
  }
  // All widget types.
  $form['widget_name'] = array(
    '#type' => 'textfield',
    '#title' => t('Widget name'),
    '#description' => t('For grouped widgets this name must be unique across the group.'),
    '#default_value' => !empty($widg_set['widget_name']) ? $widg_set['widget_name'] : substr($field['field_name'], 6),
    '#required' => TRUE,
    '#maxlength' => 64,
    '#size' => 32,
    '#attributes' => array(
      // Enforce name'ish value.
      'onkeyup' => array(
        'Drupal.tabular.field_ui.eventHandlers.convertLaxName(0, this)'
      ),
    ),
    '#weight' => ++$weight,
  );
  // Group target.
  if ($widget_role == 'target') {
    $form['display_selected'] = array(
      '#type' => 'checkbox',
      '#title' => t('Display selected row(s)'),
      '#description' => t('Otherwise they will only appear in hidden value element.'),
      '#default_value' => !empty($widg_set['display_selected']),
      '#weight' => ++$weight,
    );
    $form['display_empty'] = array(
      '#type' => 'checkbox',
      '#title' => t('Display empty list when none selected'),
      '#description' => t('If <em>on</em>: \'Show field label\' should probably be \'Show\' ;-)'),
      '#default_value' => !empty($widg_set['display_empty']),
      '#states' => array(
        'visible' => array(
          ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
        ),
      ),
      '#weight' => ++$weight,
    );
  }

  // All widget types.
  $form['field_label'] = array(
    '#type' => 'radios',
    '#title' => t('Show field label'),
    '#options' => array(
      2 => t('Show'),
      1 => t('Show when value(s)'),
      0 => t('Hide'),
    ),
    '#default_value' => isset($widg_set['field_label']) ? $widg_set['field_label'] :
      ($widget_role == 'source' ? 2 : 1),
    '#weight' => ++$weight,
  );
  if ($widget_role == 'source') {
    unset($form['field_label']['#options'][1]);
  }
  if (!$widget_solitary) {
    if ($widget_role == 'source') {
      $form['field_label']['#description'] = t('Visible field label rarely makes sense for a <em>source</em> which isn\'t the first in the group,<br>because the label gets rendered below tabs.');
    }
    else {
      $form['field_label']['#description'] = t('Field label never gets rendered unless \'Display selected row(s)\'.');
      $form['field_label']['#states'] = array(
        'visible' => array(
          ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
        ),
      );
    }
  }

  if ($widget_role == 'source') {
    // Tab.
    $form['tab_label'] = array(
      '#type' => 'textfield',
      '#title' => t('Tab label'),
      '#description' => t('Set/leave empty to hide.'),
      '#default_value' => isset($widg_set['tab_label']) ? $widg_set['tab_label'] :
        // Solitary probably shan't have visible tab.
        ($widget_solitary ? '' : $instance['label']),
      '#maxlength' => 255,
      '#attributes' => array(
        // Trim value.
        'onchange' => array(
          'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
        ),
      ),
      '#weight' => ++$weight,
    );
    $form['auto_populate'] = array(
      '#type' => 'radios',
      '#title' => t('Auto-populate - retrieve and render options'),
      '#options' => array(
        2 => t('On page load'),
        1 => t('First time activated (click on tab, visible or not)'),
        0 => t('No'),
      ),
      '#default_value' => isset($widg_set['auto_populate']) ? $widg_set['auto_populate'] : ($widget_solitary ? 2 : 1),
      '#states' => array(
        // Hide if provider is searchable.
        'visible' => array(
          ':input[name="instance[settings][provider][searchable]"]' => array('value' => '0'),
        ),
      ),
      '#weight' => ++$weight,
    );

    // If provider is searchable.
    $form['populate_search'] = array(
      '#type' => 'fieldset',
      '#title' => t('Populate search'),
      '#collapsible' => TRUE,
      '#collapsed' => TRUE,
      '#states' => array(
        // Hide unless provider is searchable.
        'visible' => array(
          ':input[name="instance[settings][provider][searchable]"]' => array('value' => '1'),
        ),
      ),
      'mode' => array(
        '#type' => 'radios',
        '#title' => t('Mode'),
        '#options' => array(
          'auto' => t('Auto-complete field'),
          'button' => t('Field with button'),
        ),
        '#default_value' => !empty($widg_set['populate_search']['mode']) ? $widg_set['populate_search']['mode'] : 'auto',
      ),
      'label' => array(
        '#type' => 'textfield',
        '#title' => t('Search field label'),
        '#default_value' => isset($widg_set['populate_search']['label']) ? $widg_set['populate_search']['label'] : '',
        '#size' => 20,
        '#prefix' => '<div class="tabular-populate-search"><hr>'
          . '<div>' . t('Leave any field empty to avoid rendering the element (or text).') . '</div>',
      ),
      'placeholder' => array(
        '#type' => 'textfield',
        '#title' => t('Search field placeholder text'),
        '#default_value' => isset($widg_set['populate_search']['placeholder']) ? $widg_set['populate_search']['placeholder'] : '',
        '#size' => 40,
      ),
      'button' => array(
        '#type' => 'textfield',
        '#title' => t('Search button label'),
        '#default_value' => isset($widg_set['populate_search']['button']) ? $widg_set['populate_search']['button'] : '',
        '#size' => 20,
        '#states' => array(
          'visible' => array(
            ':input[name="instance[widget][settings][populate_search][mode]"]' => array('value' => 'button'),
          ),
        ),
        '#suffix' => '</div>',
      ),
      '#weight' => ++$weight,
    );

    $form['row_click'] = array(
      '#type' => 'textfield',
      '#title' => t('Row click handler'),
      '#description' => t('
<b>Supported magic expressions</b>:
<br>_GET_: Get the row value as Javascript object.
<br>_SET_: Add the row to selection (is default).
<br>_INSPECT_: Dump the row value to browser console.
<br><b>Supported tokens</b>: _ENTITY_KIND_, _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
'),
      '#default_value' => isset($widg_set['row_click']) ? $widg_set['row_click'] : '_SET_',
      '#maxlength' => 255,
      '#size' => 100,
      '#attributes' => array(
        // Trim value.
        'onchange' => array(
          'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
        ),
      ),
      '#weight' => ++$weight,
    );
    // Source can also be 'on remove' because it may contain removal buttons.
    $form['show_value_change'] = array(
      '#type' => 'radios',
      '#title' => t('Show value change in status list'),
      '#options' => array(
        0 => t('No'),
        1 => t('On removal'),
        2 => t('On set'),
        3 => t('On both'),
      ),
      '#default_value' => $widg_set['show_value_change'],
      '#weight' => ++$weight,
    );
  }
  // Role:target.
  else {
    $form['row_click'] = array(
      '#type' => 'textfield',
      '#title' => t('Row click handler'),
      '#description' => t('
<b>Supported magic expressions</b>:
<br>_GET_: Get the row value as Javascript object.
<br>_REMOVE_CONFIRMED_: Remove the row from selection, if user confirms.
<br>_REMOVE_: Remove the row from selection (is default).
<br>_INSPECT_: Dump the row value to browser console.
<br><b>Supported tokens</b>: _ENTITY_KIND_, _GROUP_ID_, _WIDGET_NAME_, _PRIMARY_KEY_.
'),
      '#default_value' => isset($widg_set['row_click']) ? $widg_set['row_click'] : '_REMOVE_',
      '#maxlength' => 255,
      '#size' => 100,
      '#states' => array(
        'visible' => array(
          ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
        ),
      ),
      '#attributes' => array(
        // Trim value.
        'onchange' => array(
          'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
        ),
      ),
      '#weight' => ++$weight,
    );
    $form['remove_confirm'] = array(
      '#type' => 'textfield',
      '#title' => t('Confirm removal message'),
      '#description' => t('t() column name replacers supported, like \'Remove !name_of_primary_key_column?\'.<br>However value conversion is not applicable here (unlike column/button hover titles).'),
      '#default_value' => isset($widg_set['remove_confirm']) ? $widg_set['remove_confirm'] : '',
      '#maxlength' => 255,
      '#states' => array(
        'visible' => array(
          ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
        ),
        'required' => array(
          ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
        ),
      ),
      '#attributes' => array(
        // Trim value.
        'onchange' => array(
          'Drupal.tabular.field_ui.eventHandlers.trim(0, this)'
        ),
      ),
      '#weight' => ++$weight,
    );
    // Target can't be 'on set' because you'll never be able to set from target.
    $form['show_value_change'] = array(
      '#type' => 'radios',
      '#title' => t('Show value removal change in status list'),
      '#options' => array(
        0 => t('No'),
        1 => t('Yes'),
      ),
      '#default_value' => $widg_set['show_value_change'],
      '#weight' => ++$weight,
    );
  }

  // All types.
  $form['visible_rows'] = array(
    '#type' => 'select',
    '#title' => t('Number of visible rows'),
    '#options' => array(-1 => t('Unlimited')) + drupal_map_assoc(range(1, 100)),
    '#default_value' => isset($widg_set['visible_rows']) ? $widg_set['visible_rows'] : 10,
    '#weight' => ++$weight,
  );
  $form['view_type'] = array(
    '#type' => 'hidden',
    // There is no other view_type defined.
    '#default_value' => 'table',
    '#weight' => ++$weight,
  );

  // The 'layout' setting only exists in this form; doesn't get saved.
  // 'dropdown' and 'dynamic_table' only exist as boolean true or as array
  // having a truthy sub setting.
  $layout_dropdown = $layout_dynamic_table = FALSE;
  if (empty($widg_set['dropdown'])) {
    $widg_set['dropdown'] = array();
  }
  else {
    $layout_dropdown = TRUE;
    if (!is_array($widg_set['dropdown'])) {
      $widg_set['dropdown'] = array();
    }
  }
  if (empty($widg_set['dynamic_table'])) {
    $widg_set['dynamic_table'] = array();
  }
  else {
    $layout_dynamic_table = TRUE;
    if (!is_array($widg_set['dynamic_table'])) {
      $widg_set['dynamic_table'] = array();
    }
  }
  $form['layout'] = array(
    '#type' => 'radios',
    '#title' => t('View layout'),
    '#description' => t('Dropdown is not applicable for non-searchable provider combined with mock cardinality more/unlimited.'),
    '#options' => array(
      'plain' => t('Plain table'),
      'dropdown' => t('Dropdown'),
      'dynamic_table' => t('Dynamic table'),
    ),
    '#default_value' => $layout_dropdown ? 'dropdown' : ($layout_dynamic_table ? 'dynamic_table' : 'plain'),
    '#required' => TRUE,
    '#weight' => ++$weight,
  );
  if ($widget_role == 'target') {
    unset($form['layout']['#options']['dropdown']);
  }
  else {
    $form['dropdown'] = array(
      '#type' => 'fieldset',
      '#title' => t('Dropdown'),
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#states' => array(
        'visible' => array(
          ':input[name="instance[widget][settings][layout]"]' => array('value' => 'dropdown'),
        ),
        'expanded' => array(
          ':input[name="instance[widget][settings][layout]"]' => array('value' => 'dropdown'),
        ),
      ),
      '#weight' => ++$weight,
      // Non-searchable provider only.
      'none_option' => array(
        '#type' => 'checkbox',
        '#title' => t('Prepend \'none selected\' option'),
        '#default_value' => !empty($widg_set['dropdown']['none_option']),
        '#states' => array(
          // Only applicable when non-searchable provider.
          'visible' => array(
            ':input[name="instance[settings][provider][searchable]"]' => array('value' => '0'),
          ),
        ),
      ),
      // Irrelevant when non-searchable provider, because then we always do it.
      'pass_title' => array(
        '#type' => 'checkbox',
        '#title' => t('Upon selection, pass the <em>title</em> column value to the search field'),
        '#description' => t('Only applicable when mock cardinality is 1 (one). Ignored if otherwise.'),
        '#default_value' => isset($widg_set['dropdown']['pass_title']) ? $widg_set['dropdown']['pass_title'] : 1,
        '#states' => array(
          // Only relevant when searchable provider.
          'visible' => array(
            ':input[name="instance[settings][provider][searchable]"]' => array('value' => '1'),
          ),
          // Ignored if mock cardinality != 1.
          // But we can't require 1, because a grouped source's mock_cardinality
          // is empty string (~ undefined; gets overridden by target's ditto).
          'enabled' => array(
            ':input[name="instance[settings][mock_cardinality]"]' => array(
              array('value' => '1'),
              array('value' => ''),
            ),
          ),
        ),
        '#weight' => ++$weight,
      ),
      // Truncate upon user clearing search field.
      'clear_truncate' => array(
        '#type' => 'checkbox',
        '#title' => t('Truncate if/when user clears the search field'),
        '#description' => t('Only applicable when mock cardinality is 1 (one). Ignored if otherwise.'),
        '#default_value' => isset($widg_set['dropdown']['clear_truncate']) ? $widg_set['dropdown']['clear_truncate'] : 1,
        '#states' => array(
          // Only relevant when searchable provider.
          'visible' => array(
            ':input[name="instance[settings][provider][searchable]"]' => array('value' => '1'),
          ),
          // Ignored if mock cardinality != 1.
          // But we can't require 1, because a grouped source's mock_cardinality
          // is empty string (~ undefined; gets overridden by target's ditto).
          'enabled' => array(
            ':input[name="instance[settings][mock_cardinality]"]' => array(
              array('value' => '1'),
              array('value' => ''),
            ),
          ),
        ),
        '#weight' => ++$weight,
      ),
    );
  }
  $form['dynamic_table'] = array(
    '#type' => 'fieldset',
    '#title' => t('Dynamic table'),
    '#collapsible' => TRUE,
    '#collapsed' => FALSE,
    '#states' => array(
      'visible' => array(
        ':input[name="instance[widget][settings][layout]"]' => array('value' => 'dynamic_table'),
      ),
      'expanded' => array(
        ':input[name="instance[widget][settings][layout]"]' => array('value' => 'dynamic_table'),
      ),
    ),
    'dynamic_table' => array(
      '#type' => 'markup',
      '#markup' => '<p>' . t(
          'Based on !link (since later major versions aren\'t compatible with Drupal\'s bundled jQuery 1.4.4).',
          array('!link' => l(t('jQuery.dataTable, version 1.9.4'), 'https://legacy.datatables.net/ref'))
        ) . '</p>',
    ),
    'result_search' => array(
      '#type' => 'checkbox',
      '#title' => t('Search within table (retrieved data set)'),
      '#default_value' => !empty($widg_set['dynamic_table']['result_search']),
    ),
    'result_search_label' => array(
      '#type' => 'textfield',
      '#title' => t('Search label'),
      '#description' => t('Empty for none.'),
      '#default_value' => !isset($widg_set['dynamic_table']['result_search_label']) ? '' : $widg_set['dynamic_table']['result_search_label'],
      '#states' => array(
        'visible' => array(
          ':input[name="instance[widget][settings][dynamic_table][result_search]"]' => array('checked' => TRUE),
        ),
      ),
    ),
    'result_search_placeholder' => array(
      '#type' => 'textfield',
      '#title' => t('Search placeholder'),
      '#description' => t('Empty for none.'),
      '#default_value' => !isset($widg_set['dynamic_table']['result_search_placeholder']) ? '' : $widg_set['dynamic_table']['result_search_placeholder'],
      '#states' => array(
        'visible' => array(
          ':input[name="instance[widget][settings][dynamic_table][result_search]"]' => array('checked' => TRUE),
        ),
      ),
    ),
    '#weight' => ++$weight,
  );
  if ($widget_role == 'source') {
    $form['dynamic_table']['result_search']['#description'] =
      t('Not great usability-wise if the provider is searchable, because then there\'ll be <em>two</em> search fields.');
  }
  else {
    $conditionals = array('show_value_change', 'view_legend', 'visible_rows', 'layout');
    foreach ($conditionals as $name) {
      $form[$name]['#states'] = array(
        'visible' => array(
          ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
        ),
      );
    }
    unset($conditionals);
    // And.
    $form['dynamic_table']['#states']['visible'][] = array(
      ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
    );
  }
  $form['text_clear_button'] = array(
    '#type' => 'checkbox',
    '#title' => t('Place clear button inside text fields (like IE)'),
    '#default_value' => !empty($widg_set['text_clear_button']),
    '#states' => array(
      'visible' => array(
        // OR, requiring at least one of either.
        array(
          ':input[name="instance[settings][provider][searchable]"]' => array('value' => '1'),
        ),
        'or',
        array(
          ':input[name="instance[widget][settings][dynamic_table][result_search]"]' => array('checked' => TRUE),
        ),
      ),
    ),
    '#weight' => ++$weight,
  );
  if ($widget_role == 'target') {
    $form['text_clear_button']['#states']['visible'] = array(
      // AND, requiring both.
      ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
      ':input[name="instance[widget][settings][dynamic_table][result_search]"]' => array('checked' => TRUE),
    );
  }
  $form['view_legend'] = array(
    '#type' => 'checkbox',
    '#title' => t('Create view legend'),
    '#description' => t('For a table view that would be head. <em>Looks stupid if Dropdown</em>.'),
    '#default_value' => !empty($widg_set['view_legend']),
    '#weight' => ++$weight,
  );
  if ($widget_role == 'target') {
    $form['view_legend']['#states']['visible'][] = array(
      ':input[name="instance[widget][settings][display_selected]"]' => array('checked' => TRUE),
    );
  }

  // Hide some irrelevant or misleading fields.
  $hide_fields = array(
    // We don't use standard cardinality at all.
    'div.form-item.form-type-select.form-item-field-cardinality {display:none}',
  );
  // Group source.
  if (!$widget_solitary && $widget_role == 'source') {
    // Hide 'Required' field; only the group's target has 'requiredness'.
    $hide_fields[] = 'div.form-item.form-type-checkbox.form-item-instance-required {display:none}';
    // Hide mock_cardinality field; only the group's target has cardinality.
    $hide_fields[] = 'div.form-item.form-type-select.form-item-instance-settings-mock-cardinality {display:none}';
  }
  drupal_add_css(join(' ', $hide_fields), array('type' => 'inline'));

  // We need to work some on the input. But setting a submit handler doesn't
  // work, so we have to do it an validate function.
  $form['#element_validate'][] = 'tabular_field_widget_settings_form_validate';

  return $form;
}

/**
 * Add some #states.
 *
 * @param array $form
 * @return array
 */
function tabular_field_widget_settings_form_after_build($form) {

  // Add #states to 'layout' radio options.
  // Only relevant when source; target widget cannot be dropdown.
  if (isset($form['layout']['dropdown'])) {
    $form['layout']['dropdown']['#states'] = array(
      // Disable dropdown?
      'enabled' => array(
        array(
          ':input[name="instance[settings][provider][searchable]"]' => array('value' => '1'),
        ),
        'or',
        array(
          // Non-searchable dropdown only when mock_cardinality:1.
          // But we can't require 1, because a grouped source's mock_cardinality
          // is empty string (~ undefined; gets overridden by target's ditto).
          ':input[name="instance[settings][mock_cardinality]"]' => array(
            array('value' => '1'),
            array('value' => ''),
          ),
        ),
      ),
    );
  }

  return $form;
}

/**
 * Validator for tabular_field_widget_settings_form() and in effect also
 * tabular_field_instance_settings_form().
 *
 * Also used to tamper with values because setting submit handler has no effect.
 *
 * @param array &$form
 * @param array &$form_state
 */
function tabular_field_widget_settings_form_validate(&$form, &$form_state) {
  $failed = FALSE;

  $values_instance =& $form_state['values']['instance']['settings'];
  $values_widget =& $form_state['values']['instance']['widget']['settings'];

  $bundle = $form_state['values']['instance']['bundle'];
  $field_name = $form_state['values']['instance']['field_name'];

  $regex = array(
    'lax_name' => '[a-zA-Z\d_\-]+',
    'js' => '[a-zA-Z\d_]+(\.[a-zA-Z\d_]+)*',
  );
  $regex_display = array();
  foreach ($regex as $name => $rgx) {
    $regex_display[$name] = str_replace(array('\\d', '\\'), array('0-9', ''), $rgx);
  }

  // Instance settings ---------------------------------------------------------

  $view_columns =& $values_instance['view_columns'];
  foreach ($view_columns as $name => &$props) {
    // Convert.
    if (isset($props['convert'])) {
      if ($props['convert'] !== '') {
        $props['convert'] = trim($props['convert']);
      }
      if ($props['convert'] !== '' && !preg_match('/^' . $regex['js'] . '$/', $props['convert'])) {
        $failed = TRUE;
        form_set_error(
          'instance][settings][view_columns][' . $name . '][convert',
          t('!label is not a valid function name !regex',
            array('!regex' => $regex_display['js'], '!label' => t('Javascript converter'))
          )
        );
      }
    }
    // Hover title cannot contain double quote, because it's an HTML attribute.
    if (!empty($props['hover_title'])) {
      $props['hover_title'] = str_replace('"', "'", $props['hover_title']);
    }
  }
  // Iteration ref.
  unset($props);

  if (empty($values_instance['buttons'])) {
    $buttons = array();
  }
  else {
    $buttons =& $values_instance['buttons'];
    $n_buttons_prepend = 0;
    if ($buttons) {
      foreach ($buttons as $name => &$props) {
        // (on)click.
        if (isset($props['click'])) {
          $v = $props['click'];
          if ($v !== '') {
            $v = trim($v);
            if ($v !== '') {
              // Escape double quote - will be set a onclick HTML attribute.
              $v = str_replace('"', '\\"', $v);
            }
          }
          if ($v !== '') {
            $props['click'] = $v;
          }
          else {
            unset($props['click']);
          }
          if ($props['weight'] < 0) {
            ++$n_buttons_prepend;
            if ($n_buttons_prepend > 3) {
              $failed = TRUE;
              form_set_error(
                'instance][settings][buttons][' . $name . '][weight',
                t('Maximally 3 buttons can be prepended. That is: max 3 buttons can have negative weight.')
              );
            }
          }
        }
        // Hover title cannot contain double quote, because it's an HTML attribute.
        if (!empty($props['hover_title'])) {
          $props['hover_title'] = str_replace('"', "'", $props['hover_title']);
        }
      }
      // Iteration ref.
      unset($props);
    }
  }

  $widget_role = $values_widget['widget_role'];
  $widget_solitary = !empty($values_widget['widget_solitary']);

  // Provider fixed path args.
  if ($widget_role == 'source') {
    if ($values_instance['provider']['fixed_args'] !== '') {
      $v = trim($values_instance['provider']['fixed_args']);
      if ($v !== '') {
        // Remove leading and trailing slash.
        if ($v{0} === '/') {
          $v = substr($v, 1);
          if ($v !== '' && $v{strlen($v) - 1} === '/') {
            $v = substr($v, 0, strlen($v) - 1);
          }
        }
      }
      $values_instance['provider']['fixed_args'] = $v;
    }
    if (!$failed) {
      $values_instance['provider']['searchable'] = (int) $values_instance['provider']['searchable'];
      // Clean up, remove specs__... and fixed_args placeholders fields.
      $provider_field_keys = array_keys($values_instance['provider']);
      $redundant = array();
      foreach ($provider_field_keys as $k) {
        if (strpos($k, 'specs__') === 0 || strpos($k, 'fixed_args__') === 0) {
          $redundant[] = $k;
        }
      }
      if ($redundant) {
        foreach ($redundant as $k) {
          unset($values_instance['provider'][$k]);
        }
      }
      unset($provider_field_keys);
    }
  }

  if (!$failed) {
    if (!empty($values_instance['attach_js']['module'])) {
      if (!empty($values_instance['attach_js']['file'])) {
        $values_instance['attach_js']['file'] = trim($values_instance['attach_js']['file']);
      }
      if (empty($values_instance['attach_js']['file'])) {
        form_set_error(
          'instance][settings][attach_js][file',
          t('!file cannot be empty when !module is non-empty.',
            array('!module' => t('Attach Javascript - module'), '!file' => t('Attach Javascript - file'))
          )
        );
      }
    }
    elseif (!empty($values_instance['attach_js']['file'])) {
      form_set_error(
        'instance][settings][attach_js][module',
        t('!module cannot be empty when !file is non-empty.',
          array('!module' => t('Attach Javascript - module'), '!file' => t('Attach Javascript - file'))
        )
      );
    }
  }


  // Widget settings. ----------------------------------------------------------

  // Trim some.
  $fields = array('widget_group_id', 'widget_name', 'tab_label');
  foreach ($fields as $name) {
    if (isset($values_widget[$name])) {
      $values_widget[$name] = trim($values_widget[$name]);
    }
  }

  // Widget name.
  if (isset($values_widget['widget_name']) && drupal_strlen($values_widget['widget_name']) <= 64) {
    if (!preg_match('/^' . $regex['lax_name'] . '$/', $values_widget['widget_name'])) {
      $failed = TRUE;
      form_set_error(
        'instance][widget][settings][widget_name',
        t('!label is not a valid name !regex',
          array('!regex' => $regex_display['lax_name'], '!label' => t('Widget name'))
        )
      );
    }
  }

  // Solitary widget shan't have hardcoded widget group ID.
  // Validate widget_group_id.
  if (!$widget_solitary && isset($values_widget['widget_group_id']) && $values_widget['widget_group_id'] !== '') {
    if (!preg_match('/^' . $regex['lax_name'] . '$/', $values_widget['widget_group_id'])) {
      $failed = TRUE;
      form_set_error(
        'instance][widget][settings][widget_group_id',
        t('!label is not a valid ID !regex',
          array('!regex' => $regex_display['lax_name'], '!label' => t('Widget group ID')))
      );
    }
  }
  // Remove widget_group_id.
  if (!$failed && $widget_solitary) {
    unset($values_widget['widget_group_id']);
  }

  // Non-displaying target.
  if ($widget_role == 'target' && !$values_widget['display_selected']) {
    // Hide field label if target shan't display selected row(s).
    $values_widget['display_empty'] = 0;
    $values_widget['field_label'] = 0;
  }
  // Source or displaying target.
  else {
    // Layout is required.
    if (empty($values_widget['layout'])) {
      $failed = TRUE;
      form_set_error(
        'instance][widget][settings][layout',
        t('!name field is required.',
          array('!name' => t('View layout')))
      );
    }
    elseif ($widget_role == 'source'
      && $values_widget['layout'] == 'dropdown'
      && !$values_instance['provider']['searchable']
      && ($values_instance['mock_cardinality'] != 1 && $values_instance['mock_cardinality'] !== '')
    ) {
      $failed = TRUE;
      form_set_error(
        'instance][widget][settings][layout',
        t('!name cannot be \'Dropdown\' when non-searchable provider combined with mock cardinality more/unlimited.',
          array('!name' => t('View layout')))
      );
    }
  }

  // Populate search.
  if (!$failed && $widget_role != 'target') {
    // Static provider.
    if (!$values_instance['provider']['searchable']) {
      // No populate_search.
      $values_widget['populate_search'] = array();
    }
    // Searchable provider.
    else {
      if (!empty($values_widget['populate_search'])) {
        // No auto_populate.
        $values_widget['auto_populate'] = 0;
        // Almost all buckets are strings.
        foreach ($values_widget['populate_search'] as &$val) {
          $val = trim($val);
        }
        // Iteration ref.
        unset($val);
      }
    }
  }

  // From now on work for submission, not for validation.-----------------------
  if ($failed) {
    return;
  }

  // Layout.
  $layout_dynamic_table = FALSE;
  if ($widget_role != 'target' || $values_widget['display_selected']) {
    switch ($values_widget['layout']) {
      case 'dropdown':
        unset($values_widget['dynamic_table']);
        if (!$values_instance['provider']['searchable']) {
          // Non-searchable dropdown has no pass_title.
          unset($values_widget['dropdown']['pass_title']);
          unset($values_widget['dropdown']['clear_truncate']);
        }
        // Searchable provider.
        else {
          // Searchable dropdown has no none_option.
          unset($values_widget['dropdown']['none_option']);
          // Searchable dropdown shan't pass title unless cardinality:1.
          if ($values_instance['mock_cardinality'] != 1 && $values_instance['mock_cardinality'] !== '') {
            unset($values_widget['dropdown']['pass_title']);
            unset($values_widget['dropdown']['clear_truncate']);
          }
        }
        // Dropdown with no options is boolean true, not array.
        if (!$values_widget['dropdown']) {
          $values_widget['dropdown'] = TRUE;
        }
        break;
      case 'dynamic_table':
        $layout_dynamic_table = TRUE;
        unset($values_widget['dropdown']);
        if (!empty($values_widget['dynamic_table']['result_search'])) {
          $values_widget['dynamic_table']['result_search_label'] = trim($values_widget['dynamic_table']['result_search_label']);
          $values_widget['dynamic_table']['result_search_placeholder'] = trim($values_widget['dynamic_table']['result_search_placeholder']);
        }
        else {
          unset($values_widget['dynamic_table']['result_search']);
          unset($values_widget['dynamic_table']['result_search_label']);
          unset($values_widget['dynamic_table']['result_search_placeholder']);
        }
        if (!$values_widget['dynamic_table']) {
          $values_widget['dynamic_table'] = TRUE;
        }
        break;
      default:
        unset($values_widget['dropdown']);
        unset($values_widget['dynamic_table']);
    }
  }
  // Non-displaying target.
  else {
    unset($values_widget['dynamic_table']);
    // It has no dropdown settings, so we needn't clear those.
  }


  // Don't save redundant properties to db:field_config/field_config_instance.--
  // Columns.
  $redundant = array();
  foreach ($view_columns as $name => $props) {
    if (empty($props['weight'])) {
      $redundant[$name][] = 'weight';
    }
    // Cannot skip these if original schema has values; would be too annoying.
    //if ($props['label'] == '') {
    //  $redundant[$name][] = 'label';
    //}
    //if ($props['hover_title'] == '') {
    //  $redundant[$name][] = 'hover_title';
    //}
    //if (empty($props['convert'])) {
    //  $redundant[$name][] = 'convert';
    //}
  }
  if ($redundant) {
    foreach ($redundant as $name => $props) {
      foreach ($props as $k) {
        unset($view_columns[$name][$k]);
      }
    }
  }
  // Buttons.
  if ($buttons) {
    $redundant = array();
    foreach ($buttons as $name => $props) {
      if (empty($props['weight'])) {
        $redundant[$name][] = 'weight';
      }
      // Cannot skip these if original schema has values; would be too annoying.
      //if ($props['label'] == '') {
      //  $redundant[$name][] = 'label';
      //}
      //if ($props['hover_title'] == '') {
      //  $redundant[$name][] = 'hover_title';
      //}
      //if (empty($props['click'])) {
      //  $redundant[$name][] = 'click';
      //}
    }
    if ($redundant) {
      foreach ($redundant as $name => $props) {
        foreach ($props as $k) {
          unset($buttons[$name][$k]);
        }
      }
    }
    unset($redundant);
  }
  // Obsolete widget settings.
  unset($values_widget['layout']);


  // Translation: file t() calls for use in a Features module ------------------
  if (!variable_get('file_private_path')) {
    drupal_set_message(
      t('Can\'t file translatables because the private files dir isn\'t set (check Media > File system).'),
      'warning'
    );
    return;
  }
  $translation_context = 'field_instance:' . $bundle . '-' . $field_name;
  $translatables = array();
  // Columns.
  foreach ($view_columns as $name => $props) {
    if (!empty($props['weight'])) {
      if (isset($props['label']) && $props['label'] !== '') {
        $translatables[] = array(
          $props['label'],
          $translation_context . ':column:' . $name . ':label',
        );
      }
      if (isset($props['hover_title']) && $props['hover_title'] !== ''
        && $props['hover_title'] !== '_RAW_' && $props['hover_title'] !== '_VALUE_'
      ) {
        $translatables[] = array(
          $props['hover_title'],
          $translation_context . ':column:' . $name . ':hover_title',
        );
      }
    }
  }
  // Buttons.
  if ($buttons) {
    foreach ($buttons as $name => $props) {
      if (!empty($props['weight'])) {
        if (isset($props['label']) && $props['label'] !== '') {
          $translatables[] = array(
            $props['label'],
            $translation_context . ':button:' . $name . ':label',
          );
        }
        if (isset($props['hover_title']) && $props['hover_title'] !== '') {
          $translatables[] = array(
            $props['hover_title'],
            $translation_context . ':button:' . $name . ':hover_title',
          );
        }
      }
    }
  }
  // Translatable tab_label.
  if (isset($values_widget['tab_label']) && $values_widget['tab_label'] !== '') {
    $translatables[] = array(
      $values_widget['tab_label'],
      $translation_context . ':tab_label',
    );
  }
  // Searchable provider.
  if (!empty($values_widget['populate_search'])) {
    $populate_search =& $values_widget['populate_search'];
    $names = array('label', 'placeholder', 'button');
    foreach ($names as $name) {
      if (isset($populate_search[$name]) && $populate_search[$name] !== '') {
        $translatables[] = array(
          $populate_search[$name],
          $translation_context . ':populate_search_' . $name,
        );
      }
    }
    unset($names);
  }

  // View type: table.
  if ($layout_dynamic_table && is_array($values_widget['dynamic_table'])) {
    $dynamic_table =& $values_widget['dynamic_table'];
    // result_search.
    if (!empty($dynamic_table['result_search'])) {
      $names = array('label', 'placeholder');
      foreach ($names as $name) {
        if ($dynamic_table['result_search_' . $name] !== '') {
          $translatables[] = array(
            $dynamic_table['result_search_' . $name],
            $translation_context . ':result_search_' . $name,
          );
        }
      }
      unset($names);
    }
  }

  // Confirm remove message (target only).
  if (isset($values_widget['remove_confirm']) && $values_widget['remove_confirm'] !== '') {
    $translatables[] = array(
      $values_widget['remove_confirm'],
      $translation_context . ':remove_confirm',
    );
  }

  // Write to file in private files dir.
  if ($translatables) {
    $s = '<' . '?' . 'php' . "\n\n"
      . '// Translatables of bundle:' . $bundle
      . ', field name:' . $field_name . '.';
    foreach ($translatables as $t) {
      $s .= "\n  t('" . str_replace('\'', "''", $t[0]) . "', array(), array('context' => '" . $t[1] . "'));";
    }
    $s .= "\n";

    $filed = FALSE;
    $dir = 'private://tabular-translatables';
    $filename = $dir . '/' . $bundle . '-' . $field_name . '.custom_translations.php';
    if (file_prepare_directory($dir, FILE_CREATE_DIRECTORY)) {
      if (file_unmanaged_save_data($s, $filename, FILE_EXISTS_REPLACE)) {
        $filed = TRUE;
      }
    }
    if ($filed) {
      drupal_set_message(
        t('Filed translatables to !filename', array('!filename' => $filename))
      );
    }
    else {
      drupal_set_message(
        t('Failed to save translatables to !filename', array('!filename' => $filename)),
        'warning'
      );
    }
  }
}

/**
 * Defines field rendering in a form.
 *
 * Implements hook_field_widget_form().
 *
 * @param array &$form
 * @param array &$form_state
 * @param array $field
 * @param array $instance
 * @param string $langcode
 * @param array $items
 * @param integer $delta
 * @param array $element
 *
 * @return array
 */
function tabular_field_widget_form(&$form, &$form_state, $field, $instance, $langcode, $items, $delta, $element) {
  static $attached = array('generic' => FALSE);

  $element['#delta'] = $delta;

  $entity_kind = $field['settings']['entity_kind'];

  $entity_kind_specs = tabular_entity_kind($entity_kind);
  if (!$entity_kind_specs) {
    watchdog(
      'tabular',
      'No module declares entity kind \'@entity_kind\', thus can\'t build tabular field \'@field_name\' on bundle \'@bundle\'.',
      array('@entity_kind' => $entity_kind, '@field_name' => $field['field_name'], '@bundle' => $instance['bundle']),
      WATCHDOG_ERROR
    );

    return array();
  }
  $module = $entity_kind_specs['module'];

  $attach_messages = NULL;
  if (empty($attached[$entity_kind])) {
    // Include class handling this tabular entity kind.
    $class = $entity_kind_specs['class'];
    // Include entity kind defining class (and ancestors).
    foreach ($entity_kind_specs['includes'] as $cls => $include) {
      module_load_include($include['ext'], $include['module'], $include['path_file']);
    }
    // Get messages.
    $attach_messages = call_user_func_array($class . '::message', array($entity_kind));
  }
  unset($entity_kind_specs, $cls, $include);

  // Performance: attach generic attachments to form (not element)
  // and do it only once.
  if (!$attached['generic']) {
    $attached['generic'] = TRUE;

    $module_path = drupal_get_path('module', 'tabular');
    // Include our Javascript.
    // Judy Javascript Utilities.
    $form['#attached']['library'][] = array('judy', 'judy');
    // Generic settings.
    $form['#attached']['js'][] = array(
      'data' => array(
        'tabular' => array(
          'errorCodes' => Tabular::$errorCodes,
          'allow_group_overwrite' => variable_get('tabular_allow_group_overwrite', FALSE),
          // JS Drupal.settings has no language prop.
          'language' => $lang = $GLOBALS['language']->language,
          // Translated messages (generic).
          'message' => array(
            'generic' => Tabular::message('generic'),
          ),
          'converter_format' => variable_get('tabular_converter_format_' . $lang, tabular_converter_format_default()),
        ),
      ),
      'type' => 'setting',
    );
    // Controller etc.
    // Can't use array_merge() nor union operator (mixed keys?).
    $attached =& $form['#attached']['js'];
    // Base has to go first, and behaviors last.
    $attached[] = $module_path . '/js/tabular.base.js';
    $attached[] = $module_path . '/js/tabular.status_list.js';
    $attached[] = $module_path . '/js/tabular.widget.0.base.js';
    $attached[] = $module_path . '/js/tabular.widget.1.value.js';
    $attached[] = $module_path . '/js/tabular.widget.2.source.js';
    $attached[] = $module_path . '/js/tabular.widget.3.solitary.js';
    $attached[] = $module_path . '/js/tabular.widget.4.target.js';
    // widget.table because this is widgets using HTML table as view.
    $attached[] = $module_path . '/js/tabular.widget.table.js';
    $attached[] = $module_path . '/js/tabular.widget_group.js';
    $attached[] = $module_path . '/js/tabular.behaviors.js';
    // Clear ref.
    unset($attached);

    // Styling.
    if (variable_get('tabular_default_css', 1)) {
      //$form['#attached']['css'][] = $module_path . '/css/tabular.suggestions.css';
      $form['#attached']['css'][$module_path . '/css/tabular.suggestions.css'] = array(
        'group' => CSS_THEME,
        'weight' => 100,
      );
      if ($GLOBALS['theme'] == 'bartik') {
        // Relative+absolute positioning 'top' apparantly not mendable.
        $form['#attached']['css'][$module_path . '/css/tabular.suggestions.bartik.css'] = array(
          'group' => CSS_THEME,
          'weight' => 100,
        );
      }
    }
  }
  // Attach tabular entity kind specifics.
  if ($attach_messages) {
    $attached[$entity_kind] = TRUE;
    $form['#attached']['js'][] = array(
      'data' => array(
        'tabular' => array(
          // Translated messages.
          'message' => array(
            $entity_kind => $attach_messages,
          ),
        ),
      ),
      'type' => 'setting',
    );
    unset($attach_messages);
  }

  $inst_set = $instance['settings'];
  $widg_set = $instance['widget']['settings'];

  // Attach custom JS.
  if (!empty($inst_set['attach_js']['module'])) {
    $js_file = $inst_set['attach_js']['file'];
    $form['#attached']['js'][] = drupal_get_path('module', $inst_set['attach_js']['module'])
      . ($js_file{0} === '/' ? '' : '/') . $js_file;
  }

  $widget_type = $instance['widget']['type'];
  $widget_role = $widg_set['widget_role'];
  $widget_solitary = !empty($widg_set['widget_solitary']);
  $view_type = $widg_set['view_type'];

  // Attach libraries required by the view type.
  if ($view_type == 'table' && !empty($widg_set['dynamic_table'])) {
    // Attach jQuery.dataTable via Library API.
    if (($datatable = variable_get('tabular_library_datatable', ''))) {
      $datatable = explode('|', $datatable);
      $form['#attached']['library'][] = array($datatable[0], $datatable[1]);
    }
    unset($datatable);
  }

  $translation_context = 'field_instance:' . $instance['bundle'] . '-' . $instance['field_name'];

  // Widget (and perhaps instance) settings will be empty during field creation;
  // because the form renders default value field (which we hide; useless).
  $view_columns = $buttons = array();
  if (!empty($inst_set['view_columns'])) {
    $view_columns =& $inst_set['view_columns'];
  }
  if (!empty($inst_set['buttons'])) {
    $buttons =& $inst_set['buttons'];
  }
  if ($view_columns) {
    // Don't work on columns that aren't to be used.
    $excluded = array();
    foreach ($view_columns as $name => &$props) {
      if (!$props || empty($props['weight'])) {
        $excluded[] = $name;
      }
      else {
        $props['label'] = t($props['label'], array(), array('context' => $translation_context . ':column:' . $name . ':label'));
        if (!empty($props['hover_title'])) {
          $props['hover_title'] = t($props['hover_title'], array(), array('context' => $translation_context . ':column:' . $name . ':hover_title'));
        }
        // Remove some empties.
        if (empty($props['convert'])) {
          unset($props['convert']);
        }
        // Prevent that empty array PHP:json_encode()s to array, not object.
        if (!$props) {
          $excluded[] = $name;
        }
      }
    }
    // Iteration ref.
    unset($props);
    if ($excluded) {
      foreach ($excluded as $name) {
        unset($view_columns[$name]);
      }
    }
    unset($excluded);
  }
  uasort($view_columns, 'drupal_sort_weight');

  if ($buttons) {
    // Don't work on buttons that aren't to be used.
    $excluded = array();
    foreach ($buttons as $name => &$props) {
      if (!$props || empty($props['weight'])) {
        $excluded[] = $name;
      }
      else {
        $props['label'] = t($props['label'], array(), array('context' => $translation_context . ':button:' . $name . ':label'));
        if (!empty($props['hover_title'])) {
          $props['hover_title'] = t($props['hover_title'], array(), array('context' => $translation_context . ':button:' . $name . ':hover_title'));
        }
        // Remove some empties.
        if (empty($props['click'])) {
          unset($props['click']);
        }
        // Prevent that empty array PHP:json_encode()s to array, not object.
        if (!$props) {
          $excluded[] = $name;
        }
      }
    }
    // Iteration ref.
    unset($props);
    if ($excluded) {
      foreach ($excluded as $name) {
        unset($buttons[$name]);
      }
    }
    unset($excluded);
  }
  $buttons_before = $buttons_after = array();
  if ($buttons) {
    uasort($buttons, 'drupal_sort_weight');
    foreach ($buttons as $name => $props) {
      if ($props['weight'] < 0) {
        $buttons_before[$name] = $props;
      }
      else {
        $buttons_after[$name] = $props;
      }
    }
    unset($buttons);
  }

  // Field label.
  $attributes = array(
    'class' => array(
      'tabular-field-label',
    ),
  );
  if ($widg_set['field_label'] < 2) {
    $attributes['hidden'] = '1';
    $attributes['aria-hidden'] = 'true';
  }
  $element['tabular_label'] = array(
    '#type' => 'html_tag',
    '#tag' => 'label',
    // Localize Fields support.
    // And append space to field title, because core Form API does that.
    '#value' => (($translated = t($element['#title'], array(), array('context' => $translation_context . ':label'))) != $element['#title'] ?
      $translated : t($element['#title'])) . ' ',
    '#attributes' => $attributes,
    // Standard Form API item wrapper, except we don't set '...und-0' class
    // because can't figure out here (and don't care).
    '#prefix' => '<div class="form-item form-type-tabular">',
  );
  unset($attributes);
  if ($instance['required']) {
    // Form API #states will not work :-(
    // Space prefixed because Form API leaves one.
    $element['tabular_label']['#value'] .= ' <span class="form-required">*</span>';
  }

  // Render nothing but value element in instance/widget settings form during
  // field creation. Columns will be empty at that point.
  if ($view_columns) {
    // Root element with container(s).
    $frontend_widget_type = '';
    $provider = $dropdown_display_container = $populate_search_container = array();
    // 'dropdown' and 'dynamic_table' only exist as boolean true or as array
    // having a truthy sub setting.
    $layout_dropdown = $layout_dynamic_table = FALSE;
    if (!empty($widg_set['dropdown'])) {
      $layout_dropdown = TRUE;
      $dropdown = $widg_set['dropdown'];
      if (!is_array($dropdown)) {
        $dropdown = array();
      }
    }
    if (!empty($widg_set['dynamic_table'])) {
      $layout_dynamic_table = TRUE;
      $dynamic_table = $widg_set['dynamic_table'];
      if (!is_array($dynamic_table)) {
        $dynamic_table = array();
      }
    }

    if ($widget_role == 'source') {
      // Tab; placed above root element.
      $attributes = array(
        'tabular-widget-name' => $widg_set['widget_name'],
        'class' => array(
          'tabular-tab',
        ),
      );
      $label = $widg_set['tab_label'];
      if ($label === '') {
        $attributes['hidden'] = '1';
        $attributes['aria-hidden'] = 'true';
      }
      else {
        $label = t($label, array(), array('context' => $translation_context . ':tab_label'));
      }
      $element['tabular_tab_container'] = array(
        // Not type:container, to avoid HTML ID and 'form-wrapper' class.
        '#theme' => 'tabular_nested_container',
        '#attributes' => array(
          'hidden' => '1',
          'aria-hidden' => 'true',
          'class' => array(
            'tabular-tab-container',
          ),
        ),
        'tabular_tab' => array(
          // Not type:container, to avoid HTML ID and 'form-wrapper' class.
          '#theme' => 'tabular_nested_container',
          '#attributes' => $attributes,
          'label' => array(
            '#type' => 'html_tag',
            '#tag' => 'label',
            '#value' => $label,
          ),
        ),
      );

      // Provider, for frontend settings.
      $provider =& $inst_set['provider'];
      $prvdr = tabular_entity_kind_provider($entity_kind, $inst_set['provider']['name']);
      $provider['route'] = $prvdr['route'];
      $provider['successDataMessageResponse'] = !empty($prvdr['success_data_message_response']);
      if (!empty($prvdr['js_populate_ajax_path_args'])) {
        $provider['populateAjaxPathArgs'] = $prvdr['js_populate_ajax_path_args'][0];
        $form['#attached']['js'][] = $prvdr['js_populate_ajax_path_args'][1];
      }
      if (!empty($prvdr['js_titulate'])) {
        $provider['titulate'] = $prvdr['js_titulate'][0];
        $form['#attached']['js'][] = $prvdr['js_titulate'][1];
      }
      if (!empty($prvdr['searchable'])) {
        $provider['searchable'] = TRUE;
        if (!empty($prvdr['js_populate_search_min'])) {
          $provider['populateSearchMin'] = (int) $prvdr['js_populate_search_min'];
        }
        if (!empty($prvdr['js_populate_search_change'])) {
          $provider['populateSearchChange'] = $prvdr['js_populate_search_change'][0];
          $form['#attached']['js'][] = $prvdr['js_populate_search_change'][1];
        }
        // Searchable provider search field.
        // The search field is not shared among in-group sources.
        // If you want to see the same search phrase for all sources, then
        // the search phrase will have to copied around.
        $populate_search = isset($widg_set['populate_search']) ? $widg_set['populate_search'] : array();
        $populate_search_mode = !empty($populate_search['mode']) ? $populate_search['mode'] : 'button';

        $populate_search_field = array(
          '#type' => 'textfield',
          '#default_value' => '',
          '#name' => '',
          '#attributes' => array(
            'autocomplete' => 'off',
            'placeholder' => empty($populate_search['placeholder']) ? '' :
              t($populate_search['placeholder'], array(), array('context' => $translation_context . ':populate_search_placeholder')),
            'class' => array(
              'tabular-populate-search-field',
            ),
          ),
        );
        if (empty($populate_search['label'])) {
          $populate_search_has_label = FALSE;
        }
        else {
          $populate_search_has_label = TRUE;
          $populate_search_field['#title'] =
            t($populate_search['label'], array(), array('context' => $translation_context . ':populate_search_label'));
        }
        if ($populate_search_mode == 'auto') {
          $populate_search_button = array();
        }
        else {
          $attributes = array(
            'class' => array(
              'form-item',
              'form-type-button',
            ),
          );
          if ($populate_search_has_label) {
            $attributes['tabular-populate-search-field-has-label'] = '1';
          }
          $populate_search_button = array(
            // Not type:container, avoid HTML ID and 'form-wrapper' class.
            '#theme' => 'tabular_nested_container',
            '#attributes' => $attributes,
            'populate_search_button' => array(
              '#type' => 'button',
              '#value' => empty($populate_search['button']) ? '' :
                t($populate_search['button'], array(), array('context' => $translation_context . ':populate_search_button')),
              '#attributes' => array(
                // Clear useless 'op' name.
                'name' => '',
                'button-not-submit' => '1',
              ),
              // Convert type:submit to button immediately; don't wait for
              // Drupal.behaviors. Drupal Form API doesn't support type:button.
              '#suffix' => '<script type="text/javascript">
jQuery(\'input[type="submit"][button-not-submit="1"]\').each(function(){this.removeAttribute(\'button-not-submit\');this.type = \'button\';});
</script>',
            )
          );
        }
        $populate_search_container = array(
          // Not type:container, to avoid HTML ID and 'form-wrapper' class.
          '#theme' => 'tabular_nested_container',
          '#attributes' => array(
            'tabular-populate-search-mode' => $populate_search_mode,
            'class' => array(
              'tabular-populate-search-container',
            ),
          ),
          'populate_search_field' => $populate_search_field,
          'populate_search_button' => $populate_search_button,
        );
        unset($populate_search_field, $populate_search_button);
      }
      else {
        $provider['searchable'] = FALSE;
        if ($layout_dropdown) {
          $dropdown_display_container = array(
            // Not type:container, to avoid HTML ID and 'form-wrapper' class.
            '#theme' => 'tabular_nested_container',
            '#attributes' => array(
              'class' => array(
                'tabular-dropdown-display-container',
              ),
            ),
            'dropdown_display' => array(
              '#type' => 'textfield',
              '#default_value' => '',
              '#name' => '',
              '#attributes' => array(
                'autocomplete' => 'off',
                'readonly' => 'readonly',
                'placeholder' => !$instance['required'] ? t('- None -') : t('- Select -'),
                'class' => array(
                  'tabular-dropdown-display',
                ),
              ),
            ),
          );
        }
      }
      unset($prvdr);
      // Class: [widget type]-container.
      // The solitary type - tabular_solitary_table - is just a variant
      // of an in-group tabular_source_table.
      $frontend_widget_type = 'tabular_source_table';
    }
    else {
      $frontend_widget_type = 'tabular_target_table';
    }

    $element['tabular_root'] = array(
      // Not type:container, to avoid HTML ID and 'form-wrapper' class.
      '#theme' => 'tabular_nested_container',
      '#attributes' => array(
        'tabular-widget-name' => $widg_set['widget_name'],
        'tabular-entity-kind' => $entity_kind,
        'tabular-widget-role' => $widg_set['widget_role'],
        // Make it focusable.
        'tabindex' => '0',
        'class' => array(
          'tabular-root',
        ),
      ),
      'messages_error' => array(
        '#type' => 'html_tag',
        '#tag' => 'div',
        '#value' => '',
        '#attributes' => array(
          'hidden' => '1',
          'aria-hidden' => 'true',
          'class' => array(
            'messages',
            'error',
            'tabular-messages-error',
          ),
        ),
      ),
      'status_container' => array(
        // Not type:container, to avoid HTML ID and 'form-wrapper' class.
        '#theme' => 'tabular_nested_container',
        '#attributes' => array(
          'class' => array(
            'tabular-status-container',
          ),
        ),
        'status_positioning' => array(
          // Not type:container, to avoid HTML ID and 'form-wrapper' class.
          '#theme' => 'tabular_nested_container',
          'status_list' => array(
            '#type' => 'html_tag',
            '#tag' => 'div',
            '#value' => '',
            '#attributes' => array(
              'class' => array(
                'tabular-status-list',
              ),
            ),
          ),
        ),
      ),
      'dropdown_display_container' => $dropdown_display_container,
      'populate_search_container' => $populate_search_container,
      'view_container' => array(
        '#type' => 'html_tag',
        '#tag' => 'div',
        '#value' => '',
        '#attributes' => array(
          'class' => array(
            'tabular-view-container',
            $frontend_widget_type . '-container',
          ),
        ),
      ),
    );
    if ($widget_role == 'source' && $layout_dropdown) {
      $view_container_attributes =& $element['tabular_root']['view_container']['#attributes'];
      $view_container_attributes['tabular-view-dropdown'] = '1';
      if ($dropdown_display_container) {
        $view_container_attributes['tabular-view-dropdown-collapsed'] = '1';
      }
      // Clear ref.
      unset($view_container_attributes);
    }

    // Field description.
    if ($element['#description']) {
      $element['tabular_root']['tabular_description'] = array(
        '#type' => 'html_tag',
        '#tag' => 'div',
        // Localize Fields support.
        '#value' => ($translated = t($element['#description'], array(), array('context' => $translation_context . ':description'))) != $element['#description'] ?
          $translated : t($element['#description']),
        '#attributes' => array(
          'class' => array(
            'description',
            'tabular-field-description',
          ),
        ),
      );
    }

    // Group discovery element.
    $group_tag = '';
    $attributes = array(
      'hidden' => '1',
      'aria-hidden' => 'true',
      'tabular-group-id' => $widg_set['widget_group_id'],
      'tabular-widget-name' => $widg_set['widget_name'],
      'tabular-module' => $module,
      'tabular-entity-kind' => $entity_kind,
      'tabular-widget-type' => $frontend_widget_type,
      'tabular-view-type' => $view_type,
    );
    $js_widget_ctlr_conf = array(
      'schema' => array(
        'primaryKey' => $inst_set['primary_key'],
      ),
    );
    $js_widget_ctlr_conf['schema']['viewColumns'] =& $view_columns;
    if ($buttons_before) {
      $js_widget_ctlr_conf['schema']['buttonsBefore'] =& $buttons_before;
    }
    if ($buttons_after) {
      $js_widget_ctlr_conf['schema']['buttonsAfter'] =& $buttons_after;
    }
    $js_widget_ctlr_conf['schema']['titleColumn'] = $inst_set['title_column'];

    if ($widget_role == 'source') {
      $js_widget_ctlr_conf['autoPopulate'] = (int) $widg_set['auto_populate'];
      $js_widget_ctlr_conf['provider'] =& $provider;
      if ($widget_solitary) {
        $group_tag = 'tabular_group_solitary';
        $js_widget_ctlr_conf['mockCardinality'] = (int) $inst_set['mock_cardinality'];
      }
      else {
        $group_tag = 'tabular_group_source';
        // Use ordinary Form API for determining weight of in-group sources.
        // Weight is important for stuff like tabs.
        $attributes['tabular-group-weight'] = $instance['widget']['weight'];
      }
    }
    // Role:target.
    else {
      $group_tag = 'tabular_group_target';
      $attributes['tabular-discovery-mode'] = $widg_set['discovery_mode'];
      if ($widg_set['discovery_mode'] == 'group_id') {
        $attributes['tabular-group-nsources'] = $widg_set['widget_group_n_sources'];
      }
      $js_widget_ctlr_conf['mockCardinality'] = (int) $inst_set['mock_cardinality'];
      $js_widget_ctlr_conf['displaySelected'] = !empty($widg_set['display_selected']);
      $js_widget_ctlr_conf['displayEmpty'] = !empty($widg_set['display_empty']);
      $js_widget_ctlr_conf['displayLabelWhenValue'] = $widg_set['field_label'] == 1;
      if (!empty($widg_set['remove_confirm'])) {
        $js_widget_ctlr_conf['removeConfirm'] = $widg_set['remove_confirm'];
      }
    }

    $js_widget_ctlr_conf['rowClick'] = $widg_set['row_click'];
    $js_widget_ctlr_conf['showValueChange'] = isset($widg_set['show_value_change']) ? (int) $widg_set['show_value_change'] : 0;
    $js_widget_ctlr_conf['viewLegend'] = !empty($widg_set['view_legend']);
    $js_widget_ctlr_conf['visibleRows'] = empty($widg_set['visible_rows']) ? -1 : (int) $widg_set['visible_rows'];
    if ($widget_role == 'source') {
      if (!$layout_dropdown) {
        $js_widget_ctlr_conf['dropdown'] = FALSE;
      }
      elseif (!$provider['searchable']) {
        $js_widget_ctlr_conf['dropdown'] = array(
          'passTitle' => TRUE,
          'noneOption' => !empty($dropdown['none_option']),
        );
      }
      else {
        $js_widget_ctlr_conf['dropdown'] = array(
          'passTitle' => !empty($dropdown['pass_title']),
          'clearTruncate' => !empty($dropdown['clear_truncate']),
        );
      }
    }
    $js_widget_ctlr_conf['textClearButton'] = !empty($widg_set['text_clear_button']);
    // Widget view type specific.
    // Not if target that shan't display selected value(s).
    if ($widget_role == 'source' || $widg_set['display_selected']) {
      $viewTypeSpecs = new stdClass();
      if ($view_type == 'table') {
        if (!$layout_dynamic_table) {
          $viewTypeSpecs->dynamicTable = null;
        }
        else {
          if (!empty($dynamic_table['result_search'])) {
            $dynamic_table['result_search'] = 1;
            if ($dynamic_table['result_search_label'] !== '') {
              $dynamic_table['result_search_label'] = t($dynamic_table['result_search_label'], array(),
                array('context' => $translation_context . ':result_search_label')
              );
            }
            if ($dynamic_table['result_search_placeholder'] !== '') {
              $dynamic_table['result_search_placeholder'] = t($dynamic_table['result_search_placeholder'], array(),
                array('context' => $translation_context . ':result_search_placeholder')
              );
            }
          }
          else {
            // Keep a sub setting (even falsy) to force JSON object, not array.
            $dynamic_table['result_search'] = 0;
          }
          $viewTypeSpecs->dynamicTable =& $dynamic_table;
        }
      }
      $js_widget_ctlr_conf['viewTypeSpecs'] = $viewTypeSpecs;
    }

    $attributes['tabular-widget-ctlr-conf'] = drupal_json_encode($js_widget_ctlr_conf);
    $element['tabular_group'] = array(
      '#type' => 'html_tag',
      '#tag' => $group_tag,
      '#value' => '',
      '#attributes' => $attributes,
    );
  }

  if ($widget_solitary || $widget_role == 'target') {
    // Hidden field for storing data about the selected rows, JSON-encoded.
    $element['value'] = array(
      '#type' => 'hidden',
      '#value_callback' => 'tabular_tabular_value',
      '#default_value' => !empty($items[$delta]['value']) ? $items[$delta]['value'] : '',
      '#attributes' => array(
        'tabular-widget-name' => $widg_set['widget_name'],
        'tabular-entity-kind' => $entity_kind,
        // Firefox tends to autocomplete hidden fields.
        'autocomplete' => 'off',
      ),
      // Form item wrapper.
      '#suffix' => '</div>',
    );
  }
  else {
    // A group source field cannot have value. However we do set the HTML
    // element to conform with Form API model, to avoid trouble. But disable
    // it to avoid processing.
    $element['value'] = array(
      '#type' => 'hidden',
      '#default_value' => '',
      '#attributes' => array(
        'autocomplete' => 'off',
        'disabled' => 'disabled',
      ),
      // Form item wrapper.
      '#suffix' => '</div>',
    );
  }

  return $element;
}
